Package: RNAmf
Type: Package
Title: Recursive Non-Additive Emulator for Multi-Fidelity Data
Version: 1.1.4
Authors@R: c(
	person(given="Junoh", family="Heo", role=c("aut", "cre"), email="heojunoh@msu.edu"), 	     	person(given="Chih-Li", family="Sung", role="aut", email="sungchih@msu.edu")
	)
Maintainer: Junoh Heo <heojunoh@msu.edu>
Description: Performs RNA emulation and active learning proposed by Heo and Sung (2025) <doi:10.1080/00401706.2024.2376173> for multi-fidelity computer experiments. The RNA emulator is particularly useful when the simulations with different fidelity level are nonlinearly correlated. The hyperparameters in the model are estimated by maximum likelihood estimation. 
License: GPL-3
Encoding: UTF-8
Imports: plgp, stats, methods, lhs, doParallel, foreach, doRNG, fields,
        mvtnorm, ggplot2, ggpubr, scales
Suggests: knitr, rmarkdown
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-05-05 02:51:06 UTC; junoh
Author: Junoh Heo [aut, cre],
  Chih-Li Sung [aut]
Repository: CRAN
Date/Publication: 2026-05-05 05:20:02 UTC
Built: R 4.6.0; ; 2026-05-05 05:50:22 UTC; unix

Package: RNAmf
Type: Package
Title: Recursive Non-Additive Emulator for Multi-Fidelity Data
Version: 1.1.4
Authors@R: c(
	person(given="Junoh", family="Heo", role=c("aut", "cre"), email="heojunoh@msu.edu"), 	     	person(given="Chih-Li", family="Sung", role="aut", email="sungchih@msu.edu")
	)
Maintainer: Junoh Heo <heojunoh@msu.edu>
Description: Performs RNA emulation and active learning proposed by Heo and Sung (2025) <doi:10.1080/00401706.2024.2376173> for multi-fidelity computer experiments. The RNA emulator is particularly useful when the simulations with different fidelity level are nonlinearly correlated. The hyperparameters in the model are estimated by maximum likelihood estimation. 
License: GPL-3
Encoding: UTF-8
Imports: plgp, stats, methods, lhs, doParallel, foreach, doRNG, fields,
        mvtnorm, ggplot2, ggpubr, scales
Suggests: knitr, rmarkdown
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-05-05 02:51:06 UTC; junoh
Author: Junoh Heo [aut, cre],
  Chih-Li Sung [aut]
Repository: CRAN
Date/Publication: 2026-05-05 05:20:02 UTC
Built: R 4.4.3; ; 2026-05-05 07:29:43 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(coef,RNAmf)
S3method(fitted,RNAmf)
S3method(plot,RNAmf)
S3method(predict,RNAmf)
S3method(print,RNAmf)
S3method(print,summary.RNAmf)
S3method(residuals,RNAmf)
S3method(summary,RNAmf)
export(AL_RNAmf)
export(NestedX)
export(RNAmf)
export(imputer_RNA)
import(ggplot2)
importFrom(doParallel,registerDoParallel)
importFrom(doParallel,stopImplicitCluster)
importFrom(doRNG,"%dorng%")
importFrom(fields,rdist)
importFrom(foreach,foreach)
importFrom(ggpubr,ggarrange)
importFrom(lhs,maximinLHS)
importFrom(lhs,randomLHS)
importFrom(methods,is)
importFrom(plgp,covar.sep)
importFrom(plgp,distance)
importFrom(scales,hue_pal)
importFrom(stats,fitted)
importFrom(stats,optim)
importFrom(stats,pnorm)
importFrom(stats,predict)
importFrom(stats,quantile)
importFrom(stats,rnorm)
importFrom(stats,setNames)
importFrom(stats,uniroot)
importFrom(stats,var)

#' scale inputs before fit the model
#'
#' @param X vector or matrix of input locations to be scaled.
#' @param back logical indicating for scale back to the original vector or matrix (back=TRUE) or scale the original vector or matrix (back=FALSE). Default is FALSE.
#'
#' @return A scaled X or original X.
#'
#' @noRd
#'

scale_inputs <- function(X, center = NULL, scale = NULL, back = FALSE) {
  if (back) {
    if (is.null(center) | is.null(scale)) stop("center and scale are required to scale back")
    X <- t(t(X) * scale + center)
  } else {
    if (is.null(center)) center <- attr(scale(X), "scaled:center")
    if (is.null(scale)) scale <- attr(scale(X), "scaled:scale")

    X <- t((t(X) - center) / scale)
    attr(X, "scaled:center") <- center
    attr(X, "scaled:scale") <- scale
  }
  return(X)
}

#' Check the design is nested
#'
#' @param XX1 vector or matrix of input locations at lower fidelity
#' @param XX2 vector or matrix of input locations at higher fidelity
#'
#' @return A logical indicating if XX2 is nested or not.
#'
#' @noRd
#'

checknested <- function(XX1, XX2) {
  if (is.null(dim(XX1))) XX1 <- matrix(XX1, ncol = 1)
  if (is.null(dim(XX2))) XX2 <- matrix(XX2, ncol = 1)
  checknest <- c()
  for (i in 1:nrow(XX2)) {
    checknest <- c(checknest, suppressWarnings(any(apply(XX1, 1, function(xx) {
      all.equal(XX2[i, ], xx, tolerance = sqrt(.Machine$double.eps))
    }))))
  }
  checknest[is.na(checknest)] <- FALSE
  all(checknest)
}

#' @title Construct nested design sets
#'
#' @description The function constructs nested designs for multi-fidelity models as
#' \eqn{\mathcal{X}^*_L = \mathcal{X}_L},
#' \eqn{\mathcal{X}^*_l = \mathcal{X}_l \cup \mathcal{X}^*_{l+1}} for \eqn{l = 1, \dots, L-1},
#' and pseudo inputs \eqn{\widetilde{\mathcal{X}}_l := \mathcal{X}^*_l \setminus \mathcal{X}_l}.
#'
#' @param X_list A list of design sets for all fidelity levels.
#'
#' @return A list containing:
#' \itemize{
#'   \item \code{X_star}: A pseudo-complete nested inputs \eqn{\mathcal{X}^*_l}.
#'   \item \code{X_list}: An original inputs \eqn{\mathcal{X}_l}.
#'   \item \code{X_tilde}: A pseudo inputs \eqn{\widetilde{\mathcal{X}}_l}.
#' }
#'
#' @noRd
#'

makenested <- function(X_list){
  # Initialize the list to store nested matrices
  L <- length(X_list)
  X_star <- vector("list",L)

  # Set X^*_L = XL
  X_star[[L]] <- X_list[[L]]

  # Recursively build X^*_l for l = L down to 1
  for (l in (L-1):1) {
    X_star[[l]] <- unique(rbind(X_list[[l]], X_star[[l+1]])) # intersections are on X_list
  }

  # Compute X_tilde where X_tilde[[l]] = X_star[[l]] \ X_list[[l]]
  X_tilde <- vector("list", L)
  for (l in 1:(L-1)) {
    combined <- rbind(X_list[[l]], X_star[[l]])
    dup <- duplicated(combined)
    indices <- (nrow(X_list[[l]]) + 1) : nrow(combined)
    X_tilde[[l]] <- X_star[[l]][!dup[indices], , drop = FALSE]
  }

  return(list(X_star=X_star, X_list=X_list, X_tilde = X_tilde))
}

#' @title Find matching indices in nested designs
#'
#' @description The function identifies matching indices
#' between nested design based on a numerical distance tolerance.
#'
#' @param XX1 A matrix of original design locations.
#' @param XX2 A matrix of nested design locations.
#'
#' @return A vector of matching row indices.
#' @importFrom fields rdist
#'
#' @noRd
#'

checkindices <- function(XX1, XX2) {
  dist_matrix <- fields::rdist(XX2, XX1)
  matches <- which(dist_matrix < sqrt(.Machine$double.eps), arr.ind = TRUE)
  return(matches[, 2])
}

#' Match rows between two nested designs
#'
#' @param X_sup Matrix of the design at the lower fidelity level (the superset).
#' @param X_sub Matrix of the design at the higher fidelity level (the subset).
#' @return An integer vector of indices such that each element corresponds
#'   to the row position in \code{X_sup} matching the respective row in \code{X_sub}.
#' @details
#' This function is used to align the higher-level design points with their
#' corresponding points in the lower-level design when the designs are nested.
#' Matching is done by converting each row to a unique key string after rounding
#' to 12 decimal places, to avoid mismatches caused by floating-point precision.
#' @noRd

.match_rows <- function(X_sup, X_sub) {
  as_key <- function(M) {
    if (!is.matrix(M)) M <- as.matrix(M)
    # Optional: set a fixed rounding to avoid floating point mismatch
    M <- round(M, 12)  # fixed rounding, not user-specified
    do.call(paste, c(as.data.frame(M), sep = "\r"))
  }
  sup_key <- as_key(X_sup)
  sub_key <- as_key(X_sub)
  match(sub_key, sup_key)
}

#' computing xi component in the closed form posterior mean of matern kernel.
#'
#' @param w numerical value of \eqn{y_i^{[l-1]}}.
#' @param m numerical value of \eqn{\mu^*_{l-1}(x)}.
#' @param s numerical value of \eqn{\sigma^{*2}_{l-1}(x)}.
#' @param nu numerical value of smoothness hyperparameter. It should be 1.5 or 2.5.
#' @param theta numerical value of lengthscale hyperparameter \eqn{\theta_{ly}}.
#'
#' @return calculated value of xi component.
#'
#' @importFrom stats pnorm
#' @noRd
#'

xifun <- function(w, m, s, theta, nu) {
  L <- length(w)
  sig <- sqrt(s)
  t2 <- theta^2

  if (nu == 1.5) {
    c3 <- sqrt(3)
    mua <- m - c3 * s / theta
    mub <- m + c3 * s / theta

    e1_1 <- (theta - c3 * w) / theta
    e1_2 <- c3 / theta
    e2_1 <- (theta + c3 * w) / theta

    lam11_2 <- mua
    lam21_2 <- -mub

    term1 <- exp((3*s + 2*c3*theta*(w - m)) / (2*t2)) *
      ( (e1_1 + e1_2 * lam11_2) * pnorm((mua - w)/sig) +
          e1_2 * sig / sqrt(2 * pi) * exp(-(w - mua)^2/(2*s)) )
    term2 <- exp((3*s - 2*c3*theta*(w - m)) / (2*t2)) *
      ( (e2_1 + e1_2 * lam21_2) * pnorm((w - mub)/sig) +
          e1_2 * sig / sqrt(2 * pi) * exp(-(w - mub)^2/(2*s)) )
  } else if (nu == 2.5) {
    c5 <- sqrt(5)
    mua <- m - c5 * s / theta
    mub <- m + c5 * s / theta

    e1_1 <- 1 - c5*w/theta + 5*w^2/(theta^2*3)
    e1_2 <- c5/theta - 10*w/(theta^2*3)
    e1_3 <- 5/(theta^2*3)
    e2_1 <- 1 + c5*w/theta + 5*w^2/(theta^2*3)
    e2_2 <- c5/theta + 10*w/(theta^2*3)

    lam11_2 <- mua
    lam11_3 <- mua^2 + s
    lam12_3 <- mua + w
    lam21_2 <- -mub
    lam21_3 <- mub^2 + s
    lam22_3 <- -(mub + w)

    term1 <- exp((5*s + 2*c5*theta*(w - m)) / (2*t2)) *
      ( (e1_1 + e1_2 * lam11_2 + e1_3 * lam11_3) * pnorm((mua - w)/sig) +
          (e1_2 + e1_3 * lam12_3) * sig / sqrt(2 * pi) * exp(-(w - mua)^2/(2*s)) )
    term2 <- exp((5*s - 2*c5*theta*(w - m)) / (2*t2)) *
      ( (e2_1 + e2_2 * lam21_2 + e1_3 * lam21_3) * pnorm((w - mub)/sig) +
          (e2_2 + e1_3 * lam22_3) * sig / sqrt(2 * pi) * exp(-(w - mub)^2/(2*s)) )
  }else {
    stop("xifun: 'nu' must be 1.5 or 2.5")
  }
  return(term1 + term2)
}

#' @method summary RNAmf
#' @export
summary.RNAmf <- function(object,...){
  # Create a shallow copy to avoid modifying the original object in place
  ans <- object
  # Prepend the summary class (safer than overwriting all classes)
  class(ans) <- c("summary.RNAmf", class(object))
  ans
}

#' @export
print.summary.RNAmf <- function(x, ...){
  cat("----------------------------------------------------\n")
  cat(" Recursive Non-Additive Multi-Fidelity (RNAmf) Model\n")
  cat("----------------------------------------------------\n")
  cat(sprintf("  * Levels:   %d\n", x$level))
  cat(sprintf("  * Kernel:   %s\n", x$kernel))
  cat(sprintf("  * Constant: %s\n", x$constant))
  cat(sprintf("  * nested:   %s\n", x$nested))

  # Calculate sample sizes for all levels
  sizes <- vapply(x$fits, function(f) NROW(f$X), integer(1))
  size_str <- paste(sprintf("Level %d: %d", seq_along(sizes), sizes), collapse = ", ")
  cat(sprintf("  * Samples:  %s\n", size_str))

  cat(sprintf("  * Time:     %.3f sec\n", x$time))

  cat("----------------------------------------------------\n")
  invisible(x)
}

#' @method print RNAmf
#' @export
print.RNAmf <- function(x, ...) {
  print(summary(x, ...)) # Delegate printing to the summary method
  invisible(x)           # CRITICAL: Return the original object 'x'
}

#' @method coef RNAmf
#' @export
coef.RNAmf <- function(object, ...) {
  # Return a structured list of parameters for each level
  lapply(seq_len(object$level), function(l) {
    fit <- object$fits[[l]]
    list(
      theta = fit$theta,
      mu_hat = fit$mu.hat,
      tau2_hat = fit$tau2hat
    )
  })
}

#' @method fitted RNAmf
#' @importFrom stats predict
#' @export
fitted.RNAmf <- function(object, ...) {
  # 1. Determine the input dimension 'd' from the first level (which is never augmented)
  d <- ncol(object$fits[[1]]$X)

  # 2. Iterate through each level 'l'
  lapply(seq_along(object$fits), function(l) {

    # Extract ONLY the original input columns (1:d)
    # We drop the augmented columns (previous fidelity outputs) used during training
    X_train_raw <- object$fits[[l]]$X[, 1:d, drop = FALSE]

    # Predict using the model on these training points
    # We extract 'mu[[l]]' because we want the prediction for this specific fidelity level
    predict(object, X_train_raw, ...)$mu[[l]]
  })
}

#' @method residuals RNAmf
#' @importFrom stats fitted
#' @export
residuals.RNAmf <- function(object, ...) {
  # 1. Get Fitted values (returns list of length L)
  y_hat <- fitted(object)

  # 2. Get Observed values
  # In RNAmf, the response 'y' is stored inside each fit object in the 'fits' list
  y_obs <- lapply(object$fits, function(f) f$y)

  # 3. Calculate Residuals (y_obs - y_hat)
  # Map safely subtracts corresponding elements of the two lists
  residuals_list <- Map("-", y_obs, y_hat)

  # Optional: Assign names for clarity
  names(residuals_list) <- paste0("Level", seq_along(residuals_list))

  return(residuals_list)
}

#' @method plot RNAmf
#' @import ggplot2
#' @importFrom ggpubr ggarrange
#' @importFrom scales hue_pal
#' @importFrom stats setNames predict
#' @export
plot.RNAmf <- function(x, true_funcs = NULL, cols = NULL, ylim = NULL, ...) {
  y <- lower <- upper <- mu <- NULL
  L <- x$level
  if (ncol(x$fits[[1]]$X) > 1) {
    stop("The plot method currently only supports 1-dimensional inputs.")
  }
  
  # 2. Setup colors
  if (is.null(cols)) {
    cols <- c("red", "black", "blue", "darkgreen", "purple")
  }
  if (L > length(cols)) cols <- rep(cols, ceiling(L/length(cols)))
  
  pred_col <- hue_pal()(3)[3]
  
  # 3. Generate prediction grid
  xg <- matrix(seq(0, 1, length.out = 101), ncol = 1)
  preds <- predict(x, xg)
  df_xg <- data.frame(x = as.numeric(xg))
  
  # 4. Handle ylim
  if (is.null(ylim)) {
    y_all <- numeric(0)
    for (l in 1:L) {
      y_all <- c(y_all, x$fits[[l]]$y, 
                 preds$mu[[l]] - 1.96 * sqrt(preds$sig2[[l]]), 
                 preds$mu[[l]] + 1.96 * sqrt(preds$sig2[[l]])) 
    }
    if (!is.null(true_funcs)) {
      for (l in 1:L) y_all <- c(y_all, true_funcs[[l]](xg)) 
    }
    global_ylim <- range(y_all, na.rm = TRUE)
  } else {
    global_ylim <- ylim
  }
  
  user_theme <- theme_minimal() +
    theme(
      axis.title.x = element_text(size = 15, margin = margin(t = 10), hjust = 0.5),
      axis.title.y = element_blank(),
      panel.border = element_blank(), 
      plot.title = element_text(hjust = 0.5, size = 16),
      legend.title = element_blank(), 
      legend.position = "bottom"
    )
  
  level_names <- paste("Level", 1:L)
  all_legend_names <- c(level_names, "RNA emulator")
  
  global_colors <- setNames(c(cols[1:L], pred_col), all_legend_names)
  global_shapes <- setNames(c(1:L, NA), all_legend_names)
  global_linetypes <- setNames(c(rep("dashed", L), "solid"), all_legend_names)
  
  global_scales <- list(
    scale_color_manual(name = "", values = global_colors, drop = FALSE, limits = all_legend_names),
    scale_shape_manual(name = "", values = global_shapes, drop = FALSE, limits = all_legend_names),
    scale_linetype_manual(name = "", values = global_linetypes, drop = FALSE, limits = all_legend_names)
  )
  
  # Inject invisible data so the master legend ALWAYS generates flawlessly
  dummy_df <- data.frame(x = NA, y = NA)
  dummy_layers <- list(
    geom_line(data = dummy_df, aes(x = x, y = y, color = "RNA emulator", linetype = "RNA emulator"), linewidth = 0.75, na.rm = TRUE)
  )
  for (l in 1:L) {
    lvl <- level_names[l]
    dummy_layers <- append(dummy_layers, list(
      geom_line(data = dummy_df, aes(x = x, y = y, color = !!lvl, linetype = !!lvl), linewidth = 0.75, na.rm = TRUE),
      geom_point(data = dummy_df, aes(x = x, y = y, color = !!lvl, shape = !!lvl), size = 3, na.rm = TRUE)
    ))
  }
  
  plot_list <- list()
  
  # 5. Model Fits (True functions are now overlaid here if provided)
  for (l in 1:L) {
    df_fit <- data.frame(
      x = df_xg$x,
      mu = preds$mu[[l]],
      lower = preds$mu[[l]] - 1.96 * sqrt(preds$sig2[[l]]),
      upper = preds$mu[[l]] + 1.96 * sqrt(preds$sig2[[l]])
    )
    df_train <- data.frame(x = x$fits[[l]]$X[, 1], y = x$fits[[l]]$y)
    lvl_name <- level_names[l]
    
    p_fit <- ggplot() +
      geom_ribbon(data = df_fit, aes(x = x, ymin = lower, ymax = upper), fill = "grey", alpha = 0.6) +
      geom_line(data = df_fit, aes(x = x, y = mu, color = "RNA emulator", linetype = "RNA emulator"), linewidth = 0.75) +
      geom_point(data = df_train, aes(x = x, y = y, color = !!lvl_name, shape = !!lvl_name), size = 3)
    
    # Overlay the true function for this specific level if it exists
    if (!is.null(true_funcs)) {
      df_true <- data.frame(x = df_xg$x, y = true_funcs[[l]](xg))
      p_fit <- p_fit +
        geom_line(data = df_true, aes(x = x, y = y, color = !!lvl_name, linetype = !!lvl_name), linewidth = 0.75) 
    }
    
    # Add labels, scales, and dummy layers
    p_fit <- p_fit +
      labs(title = paste("Level", l, "Fit"), x = "x", y = "y") +
      coord_cartesian(ylim = global_ylim) +
      user_theme + global_scales + dummy_layers 
    
    plot_list[[paste0("Fit_", l)]] <- p_fit
  }
  
  # 6. Arrange plots
  final_plot <- suppressWarnings(ggarrange(plotlist = plot_list, nrow = 1, ncol = length(plot_list), common.legend = TRUE, legend = "bottom"))
  
  print(final_plot)
  invisible(final_plot) 
}

#' @title Fitting the Recursive Non-Additive model with multiple fidelity levels
#'
#' @description The function fits RNA models with designs of multiple fidelity levels.
#' The estimation method is based on MLE.
#' Available kernel choices include the squared exponential kernel, and the Matern kernel with smoothness parameter 1.5 and 2.5.
#' The function returns the fitted model at each level \eqn{1, \ldots, l}, the number of fidelity levels \eqn{l}, the kernel choice, whether constant mean or not, and the computation time.
#'
#' @seealso \code{\link{predict.RNAmf}} for prediction.
#'
#' @details Consider the model
#' \eqn{\begin{cases}
#' & f_1(\bm{x}) = W_1(\bm{x}),\\
#' & f_l(\bm{x}) = W_l(\bm{x}, f_{l-1}(\bm{x})) \quad\text{for}\quad l \geq 2,
#' \end{cases}}
#' where \eqn{f_l} is the simulation code at fidelity level \eqn{l}, and
#' \eqn{W_l(\bm{x}) \sim GP(\alpha_l, \tau_l^2 K_l(\bm{x}, \bm{x}'))} is GP model.
#' Hyperparameters \eqn{(\alpha_l, \tau_l^2, \bm{\theta_l})} are estimated by
#' maximizing the log-likelihood via an optimization algorithm "L-BFGS-B".
#' For \code{constant=FALSE}, \eqn{\alpha_l=0}.
#'
#' Covariance kernel is defined as:
#' \eqn{K_l(\bm{x}, \bm{x}')=\prod^d_{j=1}\phi(x_j,x'_j;\theta_{lj})} with
#' \eqn{\phi(x, x';\theta) = \exp \left( -\frac{ \left( x - x' \right)^2}{\theta}  \right)}
#' for squared exponential kernel; \code{kernel="sqex"},
#' \eqn{\phi(x,x';\theta) =\left( 1+\frac{\sqrt{3}|x- x'|}{\theta} \right) \exp \left( -\frac{\sqrt{3}|x- x'|}{\theta} \right)}
#' for Matern kernel with the smoothness parameter of 1.5; \code{kernel="matern1.5"} and
#' \eqn{\phi(x, x';\theta) = \left( 1+\frac{\sqrt{5}|x-x'|}{\theta} +\frac{5(x-x')^2}{3\theta^2} \right) \exp \left( -\frac{\sqrt{5}|x-x'|}{\theta} \right)}
#' for Matern kernel with the smoothness parameter of 2.5; \code{kernel="matern2.5"}.
#'
#' For further details, see Heo and Sung (2025, <\doi{https://doi.org/10.1080/00401706.2024.2376173}>).
#'
#' @param X_list A list of the matrices of input locations for all fidelity levels.
#' @param y_list A list of the vectors or matrices of response values for all fidelity levels.
#' @param kernel A character specifying the kernel type to be used. Choices are \code{"sqex"}(squared exponential), \code{"matern1.5"}, or \code{"matern2.5"}. Default is \code{"sqex"}.
#' @param constant A logical indicating for constant mean of GP (\code{constant=TRUE}) or zero mean (\code{constant=FALSE}). Default is \code{TRUE}.
#' @param init A list of the vectors of initial parameter values for optimization. Default is \code{NULL}.
#' @param n.iter Number of iterations for the stochastic EM algorithm for non-nested designs. Default is 50.
#' @param burn.ratio Fraction of iterations to discard as burn-in. Default is 0.75.
#' @param trace A logical indicating to print progress of iterations if \code{TRUE}, or not if \code{FALSE}. Default is \code{TRUE}.
#' @param ... Additional arguments for compatibility with \code{optim}.
#'
#' @return A list of class \code{RNAmf} with:
#' \itemize{
#'   \item \code{fits}: A list of fitted Gaussian process models \eqn{f_l(x)} at each level \eqn{1, \ldots, l}. Each element contains:
#'   \eqn{\begin{cases} & f_1 \text{ for } (X_1, y_1),\\ & f_l \text{ for } ((X_l, f_{l-1}(X_l)), y_l), \end{cases}}.
#'   \item \code{level}: The number of fidelity levels \eqn{l}.
#'   \item \code{kernel}: A copy of \code{kernel}.
#'   \item \code{constant}: A copy of \code{constant}.
#'   \item \code{nested}: A logical indicating whether the design is nested.
#'   \item \code{time}: A scalar indicating the computation time.
#' }
#' @usage RNAmf(X_list, y_list, kernel = "sqex", constant = TRUE,
#' init = NULL, n.iter = 50, burn.ratio = 0.75, trace = TRUE, ...)
#' @export
#' @examples
#' \donttest{
#' ### two levels example ###
#' 
#' ### Perdikaris function ###
#' f1 <- function(x) {
#'   sin(8 * pi * x)
#' }
#'
#' f2 <- function(x) {
#'   (x - sqrt(2)) * (sin(8 * pi * x))^2
#' }
#'
#' ### training data ###
#' n1 <- 13
#' n2 <- 8
#'
#' ### fix seed to reproduce the result ###
#' set.seed(1)
#'
#' ### generate initial nested design ###
#' X <- NestedX(c(n1, n2), 1)
#' X1 <- X[[1]]
#' X2 <- X[[2]]
#'
#' y1 <- f1(X1)
#' y2 <- f2(X2)
#'
#' ### fit an RNAmf ###
#' fit.RNAmf <- RNAmf(list(X1, X2), list(y1, y2), kernel = "sqex", constant=TRUE)
#'
#'
#' ### three levels example ###
#'
#' ### Branin function ###
#' branin <- function(xx, l){
#'   x1 <- xx[1]
#'   x2 <- xx[2]
#'   if(l == 1){
#'     10*sqrt((-1.275*(1.2*x1+0.4)^2/pi^2+5*(1.2*x1+0.4)/pi+(1.2*x2+0.4)-6)^2 +
#'               (10-5/(4*pi))*cos((1.2*x1+0.4))+ 10) +
#'               2*(1.2*x1+1.9) - 3*(3*(1.2*x2+2.4)-1) - 1 - 3*x2 + 1
#'   }else if(l == 2){
#'     10*sqrt((-1.275*(x1+2)^2/pi^2+5*(x1+2)/pi+(x2+2)-6)^2 +
#'               (10-5/(4*pi))*cos((x1+2))+ 10) + 2*(x1-0.5) - 3*(3*x2-1) - 1
#'   }else if(l == 3){
#'     (-1.275*x1^2/pi^2+5*x1/pi+x2-6)^2 + (10-5/(4*pi))*cos(x1)+ 10
#'   }
#' }
#'
#' output.branin <- function(x, l){
#'   factor_range <- list("x1" = c(-5, 10), "x2" = c(0, 15))
#'
#'   for(i in 1:length(factor_range)) x[i] <- factor_range[[i]][1] + x[i] * diff(factor_range[[i]])
#'   branin(x[1:2], l)
#' }
#'
#' ### training data ###
#' n1 <- 20; n2 <- 15; n3 <- 10
#'
#' ### fix seed to reproduce the result ###
#' set.seed(1)
#'
#' ### generate initial nested design ###
#' X <- NestedX(c(n1, n2, n3), 2)
#' X1 <- X[[1]]
#' X2 <- X[[2]]
#' X3 <- X[[3]]
#'
#' y1 <- apply(X1,1,output.branin, l=1)
#' y2 <- apply(X2,1,output.branin, l=2)
#' y3 <- apply(X3,1,output.branin, l=3)
#'
#' ### fit an RNAmf ###
#' fit.RNAmf <- RNAmf(list(X1, X2, X3), list(y1, y2, y3), kernel = "sqex", constant=TRUE)
#'}
#'
#'

RNAmf <- function(X_list, y_list, kernel = "sqex", constant = TRUE, init = NULL, n.iter = 50, burn.ratio = 0.75, trace = TRUE, ...) {
  t0 <- proc.time()[3]
  if (!is.list(X_list)) stop("RNAmf: 'X_list' must be a list of input matrices/vectors.")
  if (!is.list(y_list)) stop("RNAmf: 'y_list' must be a list of output matrices/vectors.")
  if(kernel != "sqex") nu <- ifelse(kernel == "matern1.5", 1.5, 2.5)

  L <- length(X_list)
  d <- NCOL(X_list[[1]])
  if (length(y_list) != L) stop("RNAmf: Length of 'X_list' and 'y_list' must match.")
  if (L < 2) stop("RNAmf: Requires at least two fidelity levels (L >= 2).")
  if (!is.character(kernel) || length(kernel) != 1 || !kernel %in% c("sqex", "matern1.5", "matern2.5")) {
    stop(paste("RNAmf: 'kernel' must be one of:", paste(c("sqex", "matern1.5", "matern2.5"), collapse = ", ")))
  }
  if (!is.logical(constant)) stop("RNAmf: 'constant' must be TRUE or FALSE.")
  if (!is.numeric(n.iter) || length(n.iter) != 1 || n.iter < 1) stop("RNAmf: 'n.iter' must be a positive integer.")
  if (!is.numeric(burn.ratio) || length(burn.ratio) != 1 || burn.ratio <= 0 || burn.ratio >= 1) stop("RNAmf: 'burn.ratio' must be a numeric value strictly between 0 and 1.")
  if (!is.logical(trace)) stop("RNAmf: 'trace' must be TRUE or FALSE.")

  fits <- vector("list", L)
  fit_fun <- switch(kernel,
                    sqex = function(X, y, init=init, ...) GP(X, y, constant = constant, init=init, ...),
                    matern1.5 = function(X, y, init=init, ...) matGP(X, y, nu = nu, constant = constant, init=init, ...),
                    matern2.5 = function(X, y, init=init, ...) matGP(X, y, nu = nu, constant = constant, init=init, ...) )
  pred_fun <- switch(kernel,
                     sqex = function(fit, Xnew, cov.out=FALSE) pred.GP(fit, Xnew, cov.out=cov.out),
                     matern1.5 = function(fit, Xnew, cov.out=FALSE) pred.matGP(fit, Xnew, cov.out=cov.out),
                     matern2.5 = function(fit, Xnew, cov.out=FALSE) pred.matGP(fit, Xnew, cov.out=cov.out) )
  # Check nested
  nested <- all(mapply(checknested, X_list[-L], X_list[-1]))

  fit_core <- function(X_star, y_star, init_list) {
    fits <- vector("list", L)
    # level 1
    fits[[1]] <- fit_fun(X_star[[1]], y_star[[1]], init_list[[1]], ...)
    # levels 2..L (augmenting with prior predicted means)
    for (lvl in 2:L) {
      X_curr <- X_star[[lvl]]
      # preallocate augmented design to avoid repeated cbind
      X_aug <- X_curr
      # fill augmentation columns sequentially (no reallocation)
      for (j in 1:(lvl - 1)) {
        pj <- pred_fun(fits[[j]], X_aug[, seq_len(ncol(X_curr) + (j - 1)), drop = FALSE])
        X_aug <- cbind(X_curr, pj$mu)
      }
      fits[[lvl]] <- fit_fun(X_aug, y_star[[lvl]], init_list[[lvl]])
    }
    fits
  }

  if(nested){
    fits <- fit_core(X_star = X_list, y_star = y_list, init_list = init)
    structure(list(fits = fits, level = L, kernel = kernel, constant = constant, nested = nested, time = proc.time()[3] - t0), class = "RNAmf" )
  } else { # Non-nested
    XX <- makenested(c(X_list[1], X_list[-1]))
    y_all <- lapply(y_list, matrix)
    yy <- list(y_star  = vector("list", L), y_list  = y_all, y_tilde = vector("list", L))

    ### Draw initial y tilde ###
    ### sample initial y1_tilde from the imputer ###
    fit1  <- fit_fun(XX$X_list[[1]], y_all[[1]], init = NULL)
    pred1 <- pred_fun(fit1, XX$X_tilde[[1]], cov.out = TRUE)
    yy$y_tilde[[1]] <- pred1$mu
    yy$y_star[[1]]  <- rbind(yy$y_list[[1]], yy$y_tilde[[1]])

    ### sample initial y_2_tilde using individual GPs ###
    if (L > 2) {
      for (l in 2:(L - 1)) {
        fit_next  <- fit_fun(XX$X_list[[l]], yy$y_list[[l]], init = NULL)
        pred_next <- pred_fun(fit_next, XX$X_tilde[[l]])
        yy$y_tilde[[l]] <- pred_next$mu
        yy$y_star[[l]]  <- rbind(yy$y_list[[l]], yy$y_tilde[[l]])
      }
    }
    yy$y_star[[L]] <- yy$y_list[[L]]

    n.burnin <- ceiling(n.iter * burn.ratio) # number of burn in
    n.param <- n.iter - n.burnin # number of collected estimates
    param_mat <- rep(list(matrix(NA, nrow=n.param, ncol=d+3)), L) # theta_x, theta_y, mu.hat, tau.hat
    param_mat[[1]] <- param_mat[[1]][,-(d+3)]

    ### initial estimates
    fits <- fit_core(X_star = XX$X_star, y_star = yy$y_star, init_list = rep(list(NULL), L))

    if (trace) {
      cat(paste(vapply(seq_len(L), function(l) {
        vals <- c(fits[[l]]$theta, fits[[l]]$mu.hat, fits[[l]]$tau2hat)
        sprintf("initial (l=%d): %s", l, paste(sprintf("%.4e", vals), collapse = " "))
      }, FUN.VALUE = character(1L)), collapse = "\n"), "\n")
    }

    for (j in 1:n.iter) { # Imputation and Maximization
      # Imputation step; impute y tilde using ESS
      yy <- imputer_RNA(XX, yy, kernel=kernel, pred1, fits)

      param.init <- lapply(fits, `[[`, "theta")
      # Maximization step; optimize parameters n.iter times
      fits <- fit_core(X_star = XX$X_star, y_star = yy$y_star, init_list = param.init)

      if(j > n.burnin){
        for (l in 1:L) {
          param_mat[[l]][j-n.burnin,] <- c(fits[[l]]$theta, fits[[l]]$mu.hat, fits[[l]]$tau2hat)
        }
      }
      if (trace) {
        cat(paste(vapply(seq_len(L), function(l) {
          vals <- c(fits[[l]]$theta, fits[[l]]$mu.hat, fits[[l]]$tau2hat)
          sprintf("iter %2d/%d (l=%d): %s", j, n.iter, l, paste(sprintf("%.4e", vals), collapse = " "))
        }, FUN.VALUE = character(1L)), collapse = "\n"), "\n")
      }
    } # end of j for loop

    # average with 75% burn-in
    colnames(param_mat[[1]]) <- c(paste0("theta_", seq_len(d)), "mu_hat", "tau2_hat")
    param_mat[-1] <- lapply(param_mat[-1], function(m) { colnames(m) <- c(paste0("theta_", seq_len(d)), "theta_y", "mu_hat", "tau2_hat"); m })

    final_params <- lapply(param_mat,colMeans)

    fits[[1]]$theta <- final_params[[1]][1:d]
    fits[[1]]$mu.hat <- final_params[[1]][d+1]
    fits[[1]]$tau2hat <- final_params[[1]][d+2]
    for (l in 2:L) {
      fits[[l]]$theta <- final_params[[l]][1:(d+1)]
      fits[[l]]$mu.hat <- final_params[[l]][d+2]
      fits[[l]]$tau2hat <- final_params[[l]][d+3]
    }

    structure(list(fits = fits, level = L, kernel = kernel, constant = constant, XX = XX, yy = yy, nested = nested, pred1 = pred1, time = proc.time()[3] - t0), class = "RNAmf" )
  }
}

#' @title Imputation step in stochastic EM for the non-nested RNA Model
#'
#' @description The function performs the imputation step of the stochastic EM algorithm for the RNA model when the design is not nested.
#' The function generates pseudo outputs \eqn{\widetilde{\mathbf{y}}_l} at pseudo inputs \eqn{\widetilde{\mathcal{X}}_l}.
#'
#' @details The \code{imputer_RNA} function then imputes the corresponding pseudo outputs
#' \eqn{\widetilde{\mathbf{y}}_l = f_l(\widetilde{\mathcal{X}}_l)}
#' by drawing samples from the conditional normal distribution,
#' given fixed parameter estimates and previous-level outputs \eqn{Y_{l}^{*(m-1)}} for each \eqn{l},
#' at the \eqn{m}-th iteration of the EM algorithm.
#'
#' @param XX A list of design sets for all fidelity levels, containing \code{X_star}, \code{X_list}, and \code{X_tilde}.
#' @param yy A list of current observed and pseudo-responses, containing \code{y_star}, \code{y_list}, and \code{y_tilde}.
#' @param kernel A character specifying the kernel type to be used. Choices are \code{"sqex"}(squared exponential), \code{"matern1.5"}, or \code{"matern2.5"}.
#' @param pred1 Predictive results for the lowest fidelity level \eqn{f_1}. It should include \code{cov} obtained by setting \code{cov.out=TRUE}.
#' @param fits A fitted GP object from \code{RNAmf}.
#'
#' @return An updated \code{yy} list containing:
#' \itemize{
#'   \item \code{y_star}: An updated pseudo-complete outputs \eqn{\mathbf{y}^*_l}.
#'   \item \code{y_list}: An original outputs \eqn{\mathbf{y}_l}.
#'   \item \code{y_tilde}: A newly imputed pseudo outputs \eqn{\widetilde{\mathbf{y}}_l}.
#' }
#' @usage imputer_RNA(XX, yy, kernel=kernel, pred1, fits)
#' @export
#'

imputer_RNA <- function(XX, yy, kernel=kernel, pred1, fits){

  if (!is.list(XX) || is.null(XX$X_star)) stop("imputer_RNA: 'XX' must be a list containing 'X_star'.")
  if (!is.list(yy) || is.null(yy$y_list)) stop("imputer_RNA: 'yy' must be a valid output list.")
  if (!is.list(pred1) || is.null(pred1$mu) || is.null(pred1$cov)) stop("imputer_RNA: 'pred1' must contain 'mu' and 'cov' (ensure predict was with cov.out=TRUE).")
  if (!is.list(fits) || length(fits) < 1) stop("imputer_RNA: 'fits' must be a non-empty list from RNAmf.")

  L <- length(XX$X_star)
  n_list <- sapply(yy$y_list, length)

  # Sample y_1
  yy$y_tilde[[1]] <- t(mvtnorm::rmvnorm(1, mean = pred1$mu, sigma = pred1$cov))
  yy$y_star[[1]] <- rbind(yy$y_list[[1]],yy$y_tilde[[1]])

  if(L>2){
    # Sample y_2, ..., y_{L-1}
    for(l in 2:(L-1)){
      fit_l <- fits[[l]]
      if (is.null(fit_l$K) || is.null(fit_l$mu.hat)) stop(paste("imputer_RNA: Fit object at level", l, "is malformed."))

      y <- fit_l$y
      K <- fit_l$K
      g <- fit_l$g
      mu.hat <- fit_l$mu.hat
      tau2hat <- fit_l$tau2hat

      ### sampling from Y_tilde give Y_list
      list_idx <- 1:n_list[l]
      K_list <- K[list_idx, list_idx] # K corresponding to X_list
      K_tilde <- K[-list_idx, -list_idx] # K corresponding to X_tilde
      K_list_tilde <- K[list_idx, -list_idx] # K corresponding to X_list given X_tilde
      y_list <- y[list_idx]

      chol_list <- chol(K_list)
      cond_mean <- mu.hat + crossprod(K_list_tilde, backsolve(chol_list, forwardsolve(t(chol_list), y_list - mu.hat)))
      cond_var <- tau2hat * (K_tilde - crossprod(forwardsolve(t(chol_list), K_list_tilde)))
      cond_var <- (cond_var + t(cond_var)) / 2

      y_prior <- t(mvtnorm::rmvnorm(1, mean = cond_mean, sigma = cond_var))
      yy$y_tilde[[l]] <- y_prior
      yy$y_star[[l]] <- rbind(yy$y_list[[l]], yy$y_tilde[[l]])
    }
  }
  return(yy)
}

#' computing zeta component in the closed form posterior variance of matern kernel.
#'
#' @param w1 numerical value of \eqn{y_i^{[l-1]}}.
#' @param w2 numerical value of \eqn{y_k^{[l-1]}}.
#' @param m numerical value of \eqn{\mu^*_{l-1}(x)}.
#' @param s numerical value of \eqn{\sigma^{*2}_{l-1}(x)}.
#' @param nu numerical value of smoothness hyperparameter. It should be 1.5 or 2.5.
#' @param theta numerical value of lengthscale hyperparameter \eqn{\theta_{ly}}.
#'
#' @return calculated value of zeta component.
#'
#' @importFrom stats pnorm
#' @noRd
#'

zetafun <- function(w1, w2, m, s, nu, theta) {
  L    <- length(w1)
  sig    <- sqrt(s)
  t2     <- theta^2

  wsmall <- pmin(w1, w2)
  wlarge <- pmax(w1, w2)

  if (nu == 1.5) {
    c3    <- sqrt(3)
    muc <- m - 2 * c3 * s/theta
    mud <- m + 2 * c3 * s/theta

    lam31_2 <- muc; lam31_3 <- muc^2+s; lam32_3 <- muc + wlarge
    lam41_2 <- m; lam41_3 <- m^2+s; lam42_3 <- m + wsmall; lam43_3 <- m + wlarge
    lam51_2 <- -mud; lam51_3 <- mud^2+s; lam52_3 <- -(mud + wsmall)

    e3_1 <- 1 + (3*wsmall*wlarge - c3*theta*(wsmall+wlarge))/t2
    e3_2 <- (2*c3*theta - 3*(wsmall+wlarge))/t2
    e3_3 <- 3/t2

    e4_1 <- 1 + ( -3*wsmall*wlarge + c3*theta*(wlarge-wsmall))/t2
    e4_2 <- 3*(wsmall+wlarge)/t2
    e4_3 <- -3/t2

    e5_1 <- 1 + (3*wsmall*wlarge + c3*theta*(wsmall+wlarge))/t2
    e5_2 <- (2*c3*theta + 3*(wsmall+wlarge))/t2
    e5_3 <- 3/t2

    term1 <- exp((6*s + c3*theta*(wsmall+wlarge - 2*m)) / t2) *
      ( (e3_1 + e3_2*lam31_2 + e3_3*lam31_3) * pnorm((muc-wlarge)/sig) +
          (e3_2 + e3_3*lam32_3) * (sig/sqrt(2*pi)) * exp(-(wlarge-muc)^2/(2*s)) )
    term2 <- exp(-(c3*(wlarge-wsmall)/theta)) *
      ( (e4_1 + e4_2*lam41_2 + e4_3*lam41_3) * (pnorm((wlarge-m)/sig) - pnorm((wsmall-m)/sig)) +
          (e4_2 + e4_3*lam42_3) * (sig/sqrt(2*pi)) * exp(-(wsmall-m)^2/(2*s)) -
          (e4_2 + e4_3*lam43_3) * (sig/sqrt(2*pi)) * exp(-(wlarge-m)^2/(2*s)) )
    term3 <- exp((6*s - c3*theta*(wsmall+wlarge - 2*m)) / t2) *
      ( (e5_1 + e5_2*lam51_2 + e5_3*lam51_3) * pnorm((wsmall-mud)/sig) +
          (e5_2 + e5_3*lam52_3) * (sig/sqrt(2*pi)) * exp(-(wsmall-mud)^2/(2*s)) )
  } else if (nu == 2.5) {
    c5  <- sqrt(5)
    muc <- m - 2*c5*s/theta
    mud <- m + 2*c5*s/theta

    lam31_2 <- muc; lam31_3 <- muc^2+s; lam31_4 <- muc^3+3*muc*s; lam31_5 <- muc^4+6*muc^2*s+3*s^2
    lam32_3 <- muc + wlarge; lam32_4 <- muc^2 + 2*s + wlarge^2 + muc*wlarge; lam32_5 <- wlarge^3 + muc^3 + muc^2*wlarge + muc*wlarge^2 + 3*s*wlarge + 5*muc*s
    lam41_2 <- m; lam41_3 <- m^2+s; lam41_4 <- m^3+3*m*s; lam41_5 <- m^4+6*m^2*s+3*s^2
    lam42_3 <- m + wsmall; lam42_4 <- m^2 + 2*s + wsmall^2 + m*wsmall; lam42_5 <- wsmall^3 + wsmall^2*m + wsmall*m^2 + m^3 + 3*s*wsmall + 5*s*m
    lam43_3 <- m + wlarge; lam43_4 <- m^2 + 2*s + wlarge^2 + m*wlarge; lam43_5 <- wlarge^3 + wlarge^2*m + wlarge*m^2 + m^3 + 3*s*wlarge + 5*s*m
    lam51_2 <- -mud; lam51_3 <- mud^2+s; lam51_4 <- -mud^3-3*mud*s; lam51_5 <- mud^4+6*mud^2*s+3*s^2
    lam52_3 <- -(mud + wsmall); lam52_4 <- mud^2 + 2*s + wsmall^2 + mud*wsmall; lam52_5 <- -mud^3 - wsmall^3 - mud^2*wsmall - mud*wsmall^2 - 3*s*wsmall - 5*mud*s

    e3_1 <- 1 + (25*wsmall^2*wlarge^2 - 3*sqrt(5)*(3*theta^3+5*theta*wsmall*wlarge)*(wsmall+wlarge) + 15*theta^2*(wsmall^2+wlarge^2+3*wsmall*wlarge)) / (9*theta^4)
    e3_2 <- (18*sqrt(5)*theta^3 + 15*sqrt(5)*theta*(wsmall^2+wlarge^2) - 75*theta^2*(wsmall+wlarge) - 50*wsmall*wlarge*(wsmall+wlarge) + 60*sqrt(5)*theta*wsmall*wlarge) / (9*theta^4)
    e3_3 <- 5*(5*wsmall^2+5*wlarge^2+15*theta^2 - 9*sqrt(5)*theta*(wsmall+wlarge)+20*wsmall*wlarge) / (9*theta^4)
    e3_4 <- 10*(3*sqrt(5)*theta - 5*(wsmall+wlarge)) / (9*theta^4)
    e3_5 <- 25/(9*theta^4)

    e4_1 <- 1 + (25*wsmall^2*wlarge^2 + 3*sqrt(5)*(3*theta^3-5*theta*wsmall*wlarge)*(wlarge-wsmall) + 15*theta^2*(wsmall^2+wlarge^2-3*wsmall*wlarge)) / (9*theta^4)
    e4_2 <- 5*(3*sqrt(5)*theta*(wlarge^2-wsmall^2) + 3*theta^2*(wsmall+wlarge) - 10*wsmall*wlarge*(wsmall+wlarge)) / (9*theta^4)
    e4_3 <- 5*(5*wsmall^2+5*wlarge^2-3*theta^2 - 3*sqrt(5)*theta*(wlarge-wsmall)+20*wsmall*wlarge) / (9*theta^4)
    e4_4 <- -50*(wsmall+wlarge) / (9*theta^4)
    e4_5 <- 25/(9*theta^4)

    e5_1 <- 1 + (25*wsmall^2*wlarge^2 + 3*sqrt(5)*(3*theta^3+5*theta*wsmall*wlarge)*(wsmall+wlarge) + 15*theta^2*(wsmall^2+wlarge^2+3*wsmall*wlarge)) / (9*theta^4)
    e5_2 <- (18*sqrt(5)*theta^3 + 15*sqrt(5)*theta*(wsmall^2+wlarge^2) + 75*theta^2*(wsmall+wlarge) + 50*wsmall*wlarge*(wsmall+wlarge) + 60*sqrt(5)*theta*wsmall*wlarge) / (9*theta^4)
    e5_3 <- 5*(5*wsmall^2+5*wlarge^2+15*theta^2 + 9*sqrt(5)*theta*(wsmall+wlarge)+20*wsmall*wlarge) / (9*theta^4)
    e5_4 <- 10*(3*sqrt(5)*theta + 5*(wsmall+wlarge)) / (9*theta^4)
    e5_5 <- 25/(9*theta^4)

    term1 <- exp((10*s + sqrt(5)*theta*(wsmall+wlarge-2*m)) / t2) *
      ( (e3_1 + e3_2*lam31_2 + e3_3*lam31_3 + e3_4*lam31_4 + e3_5*lam31_5) * pnorm((muc-wlarge)/sig) +
          (e3_2 + e3_3*lam32_3 + e3_4*lam32_4 + e3_5*lam32_5) * (sig/sqrt(2*pi)) * exp(-(wlarge-muc)^2/(2*s)) )
    term2 <- exp(-sqrt(5)*(wlarge-wsmall)/theta) *
      ( (e4_1 + e4_2*lam41_2 + e4_3*lam41_3 + e4_4*lam41_4 + e4_5*lam41_5) * (pnorm((wlarge-m)/sig) - pnorm((wsmall-m)/sig)) +
          (e4_2 + e4_3*lam42_3 + e4_4*lam42_4 + e4_5*lam42_5) * (sig/sqrt(2*pi)) * exp(-(wsmall-m)^2/(2*s)) -
          (e4_2 + e4_3*lam43_3 + e4_4*lam43_4 + e4_5*lam43_5) * (sig/sqrt(2*pi)) * exp(-(wlarge-m)^2/(2*s)) )
    term3 <- exp((10*s - sqrt(5)*theta*(wsmall+wlarge-2*m)) / t2) *
      ( (e5_1 + e5_2*lam51_2 + e5_3*lam51_3 + e5_4*lam51_4 + e5_5*lam51_5) * pnorm((wsmall-mud)/sig) +
          (e5_2 + e5_3*lam52_3 + e5_4*lam52_4 + e5_5*lam52_5) * (sig/sqrt(2*pi)) * exp(-(wsmall-mud)^2/(2*s)) )
  }else {
    stop("zetafun: 'nu' must be 1.5 or 2.5")
  }
  return(term1 + term2 + term3)
}

#' @title Closed-form prediction for RNAmf model
#'
#' @description The function computes the closed-form posterior mean and variance for the RNAmf model
#' both at the fidelity levels used in model fitting using the chosen kernel.
#'
#' @param fits A fitted GP object from \code{RNAmf}.
#' @param x A vector or matrix of new input locations to predict.
#' @param kernel A character specifying the kernel type to be used. Choices are \code{"sqex"}(squared exponential kernel), \code{"matern1.5"}(Matern kernel with \eqn{\nu=1.5}), or \code{"matern2.5"}(Matern kernel with \eqn{\nu=2.5}). Default is \code{"sqex"}.
#' @param XX A list containing a pseudo-complete inputs \code{X_star}(\eqn{\left\{\mathcal{X}^*_l\right\}_{l=1}^{L}}), an original inputs \code{X_list}(\eqn{\left\{\mathcal{X}_l\right\}_{l=1}^{L}}), and a pseudo inputs \code{X_tilde}(\eqn{\left\{\widetilde{\mathcal{X}}_l\right\}_{l=1}^{L}}) for non-nested design.
#' @param pseudo_yy A list containing a pseudo-complete outputs \code{y_star}(\eqn{\left\{\mathbf{y}^*_l\right\}_{l=1}^{L}}), an original outputs \code{y_list}(\eqn{\left\{\mathbf{y}_l\right\}_{l=1}^{L}}), and a pseudo outputs \code{y_tilde}(\eqn{\left\{\widetilde{\mathbf{y}}_l\right\}_{l=1}^{L}}) imputed by \code{\link{imputer_RNA}}.
#'
#' @return A list of predictive posterior mean and variance for each level containing:
#' \itemize{
#'   \item \code{mu}: A list of predictive posterior mean at each fidelity level.
#'   \item \code{sig2}: A list of predictive posterior variance at each fidelity level.
#' }
#'
closed_form_RNA <- function(fits, x, kernel, XX=NULL, pseudo_yy=NULL) {

  if (!is.list(fits) || length(fits) < 1) stop("closed_form_RNA: 'fits' must be a valid RNA objects.")
  if (!is.numeric(x)) stop("closed_form_RNA: 'x' must be a numeric matrix or vector.")
  if (!is.null(pseudo_yy)) {
    if (is.null(XX)) stop("closed_form_RNA: 'XX' must be provided when 'pseudo_yy' is not NULL.")
    if (!is.list(XX) || is.null(XX$X_star)) stop("closed_form_RNA: 'XX' is malformed.")
    if (!is.list(pseudo_yy) || is.null(pseudo_yy$y_star)) stop("closed_form_RNA: 'pseudo_yy' is malformed.")
  }
  L <- length(fits)
  d <- ncol(fits[[1]]$X)

  mu_list   <- vector("list", L)
  sig2_list <- vector("list", L)

  if(is.null(pseudo_yy)){
    if (kernel == "sqex") {
      pred1 <- pred.GP(fits[[1]], x)
    } else {
      nu <- fits[[1]]$nu
      pred1 <- pred.matGP(fits[[1]], x)
    }
  } else {
    X1 <- XX$X_star[[1]]
    y1 <- pseudo_yy$y_star[[1]]
    n1 <- nrow(X1)

    if(kernel=="sqex"){
      K <- covar.sep(X1, d = fits[[1]]$theta, g = fits[[1]]$g)
    } else {
      nu <- fits[[1]]$nu
      K <- cor.sep(X1, theta = fits[[1]]$theta, nu = nu) + diag(fits[[1]]$g, n1)
    }

    chol_K <- chol(K)
    Ki <- chol2inv(chol_K)
    fits[[1]]$Ki <- Ki
    one.vec <- matrix(1, ncol = 1, nrow = n1)
    fits[[1]]$mu.hat <- drop(crossprod(one.vec, crossprod(Ki, y1)) / crossprod(one.vec, crossprod(Ki, one.vec)))
    fits[[1]]$tau2hat <- drop(crossprod(y1 - fits[[1]]$mu.hat, crossprod(Ki, y1 - fits[[1]]$mu.hat)) / n1)

    if(kernel == "sqex") {
      pred1 <- pred.GP(fits[[1]], x)
    } else {
      pred1 <- pred.matGP(fits[[1]], x)
    }
  }

  mu_list[[1]]   <- pred1$mu
  sig2_list[[1]] <- pred1$sig2

  for (k in 2:L) {
    fit_curr <- fits[[k]]
    X_curr  <- matrix(fit_curr$X[, -(d + 1)], ncol = d)
    theta   <- fit_curr$theta
    if(is.null(pseudo_yy)){
      w_curr  <- fit_curr$X[, d + 1]
      y_curr  <- fit_curr$y
      Ci      <- fit_curr$Ki
      mu_curr <- fit_curr$mu.hat
      tau2hat <- fit_curr$tau2hat
      n <- length(y_curr)
    } else {
      w_curr <- c(pseudo_yy$y_star[[k-1]][checkindices(XX$X_star[[k-1]], XX$X_star[[k]]), , drop = FALSE])
      y_curr <- pseudo_yy$y_star[[k]]
      g <- fit_curr$g
      n <- length(y_curr)

      X_aug <- cbind(X_curr, w_curr)
      if(kernel=="sqex"){
        K <- covar.sep(X_aug, d = theta, g = 0)
      } else {
        K <- cor.sep(X_aug, theta = theta, nu = nu)
      }
      chol_Ci <- chol(K + diag(g, n))
      Ci <- chol2inv(chol_Ci)
      one.vec <- matrix(1, ncol = 1, nrow = n)
      mu_curr <- drop(crossprod(one.vec, crossprod(Ci, y_curr)) / crossprod(one.vec, crossprod(Ci, one.vec)))
      tau2hat <- drop(crossprod(y_curr - mu_curr, crossprod(Ci, y_curr - mu_curr)) / n)
    }
    a <- crossprod(Ci, y_curr - mu_curr)

    x.mu <- mu_list[[k - 1]]
    sig2 <- sig2_list[[k - 1]]

    if (kernel == "sqex") {
      theta_d1 <- theta[d + 1]
      sqrt_theta <- sqrt(theta[-(d + 1)])
      x_s <- sweep(x, 2, sqrt_theta, "/")
      X_curr_s <- sweep(X_curr, 2, sqrt_theta, "/")
      K_x_X <- exp(-distance(x_s, X_curr_s))

      diff_sq <- drop(outer(x.mu, w_curr, "-")^2)
      denom_vec <- theta_d1 + 2 * sig2
      term_exp <- exp(-(diff_sq / denom_vec))
      term_scale <- 1 / sqrt(1 + 2 * sig2 / theta_d1)
      weights <- K_x_X * term_exp * term_scale

      predy <- mu_curr + crossprod(t(weights), a)

      predsig2 <- numeric(nrow(x))
      common_term_w <- exp(-(outer(w_curr, w_curr, "-"))^2 / (2 * theta_d1))

      for (i in 1:nrow(x)) {
        s2_i <- sig2[i]
        k_x <- K_x_X[i, ]
        mat_base <- tcrossprod(k_x)
        w_sum_half <- outer(w_curr, w_curr, "+") / 2
        exp_mat <- exp(-(w_sum_half - x.mu[i])^2 / (theta_d1 / 2 + 2 * s2_i))
        scale_val <- 1 / sqrt(1 + 4 * s2_i / theta_d1)
        mat <- mat_base * scale_val * exp_mat * common_term_w
        quad <- drop(crossprod(a, crossprod(mat, a)))
        tr_term <- sum(Ci * mat)

        predsig2[i] <- pmax(0, tau2hat - (predy[i] - mu_curr)^2 + quad - tau2hat * tr_term)
      }

    } else {
      K_x_X <- cor.sep(x, X_curr, theta[-(d + 1)], nu=nu)
      xi_vec <- xifun(w = rep(w_curr, each = nrow(x)),
                      m = rep(x.mu, n),
                      s = rep(sig2, n),
                      theta = theta[d+1], nu = nu)
      xi_matrix <- matrix(xi_vec, nrow(x), n)

      predy <- mu_curr + rowSums(crossprod(t(K_x_X * xi_matrix), a))

      predsig2 <- numeric(nrow(x))
      for(i in 1:nrow(x)) {
        zeta_mat <- matrix(zetafun(w1 = rep(w_curr, times = n),
                                   w2 = rep(w_curr, each  = n),
                                   m = x.mu[i], s = sig2[i],
                                   nu = nu, theta = theta[d+1]), n, n)
        mat <- tcrossprod(K_x_X[i, ]) * zeta_mat
        quad <- drop(crossprod(a, crossprod(mat, a)))
        tr_term <- sum(Ci * mat)

        predsig2[i] <- pmax(0, tau2hat - (predy[i] - mu_curr)^2 + quad - tau2hat * tr_term)
      }
    }
    mu_list[[k]] <- predy
    sig2_list[[k]] <- predsig2
  }
  return(list(mu = mu_list, sig2 = sig2_list))
}


#' @title prediction of the RNAmf emulator with multiple fidelity levels.
#'
#' @description The function computes the posterior mean and variance of RNA models with multiple fidelity levels
#' by fitted model from \code{\link{RNAmf}}.
#'
#' @seealso \code{\link{RNAmf}} for model fitting.
#'
#' @details The \code{predict.RNAmf} function internally calls \code{\link{closed_form_RNA}}
#' to recursively compute the closed-form posterior mean and variance at each level.
#' 
#' From the fitted model from \code{\link{RNAmf}},
#' the posterior mean and variance are calculated based on the closed-form expression derived by a recursive fashion.
#' The formulas depend on its kernel choices.
#' For further details, see Heo and Sung (2025, <\doi{https://doi.org/10.1080/00401706.2024.2376173}>).
#'
#' @param object An object of class \code{RNAmf} fitted by \code{\link{RNAmf}}.
#' @param x A vector or matrix of new input locations for prediction.
#' @param nimpute Number of imputations for non-nested designs. Default is 50.
#' @param ... Additional arguments for compatibility with generic method \code{predict}.
#'
#' @return
#' \itemize{
#'   \item \code{mu}: A list of vectors containing the predictive posterior mean at each fidelity level.
#'   \item \code{sig2}: A list of vectors containing the predictive posterior variance at each fidelity level.
#'   \item \code{time}: A scalar indicating the computation time.
#' }
#'
#' @importFrom plgp distance
#' @importFrom stats predict
#' @rdname predict
#' @method predict RNAmf
#' @export
#' @examples
#' \donttest{
#' ### two levels example ###
#'
#' ### Perdikaris function ###
#' f1 <- function(x) {
#'   sin(8 * pi * x)
#' }
#'
#' f2 <- function(x) {
#'   (x - sqrt(2)) * (sin(8 * pi * x))^2
#' }
#'
#' ### training data ###
#' n1 <- 13
#' n2 <- 8
#'
#' ### fix seed to reproduce the result ###
#' set.seed(1)
#'
#' ### generate initial nested design ###
#' X <- NestedX(c(n1, n2), 1)
#' X1 <- X[[1]]
#' X2 <- X[[2]]
#'
#' y1 <- f1(X1)
#' y2 <- f2(X2)
#'
#' ### n=100 uniform test data ###
#' x <- seq(0, 1, length.out = 100)
#'
#' ### fit an RNAmf ###
#' fit.RNAmf <- RNAmf(list(X1, X2), list(y1, y2), kernel = "sqex", constant=TRUE)
#'
#' ### predict ###
#' predy <- predict(fit.RNAmf, x)$mu[[2]]
#' predsig2 <- predict(fit.RNAmf, x)$sig2[[2]]
#'
#' ### RMSE ###
#' print(sqrt(mean((predy - f2(x))^2)))
#'
#' ### visualize the emulation performance ###
#' plot(x, predy,
#'   type = "l", lwd = 2, col = 3, # emulator and confidence interval
#'   ylim = c(-2, 1)
#' )
#' lines(x, predy + 1.96 * sqrt(predsig2 * length(y2) / (length(y2) - 2)), col = 3, lty = 2)
#' lines(x, predy - 1.96 * sqrt(predsig2 * length(y2) / (length(y2) - 2)), col = 3, lty = 2)
#'
#' curve(f2(x), add = TRUE, col = 1, lwd = 2, lty = 2) # high fidelity function
#'
#' points(X1, y1, pch = 1, col = "red") # low-fidelity design
#' points(X2, y2, pch = 4, col = "blue") # high-fidelity design
#'
#' ### three levels example ###
#' ### Branin function ###
#' branin <- function(xx, l){
#'   x1 <- xx[1]
#'   x2 <- xx[2]
#'   if(l == 1){
#'     10*sqrt((-1.275*(1.2*x1+0.4)^2/pi^2+5*(1.2*x1+0.4)/pi+(1.2*x2+0.4)-6)^2 +
#'     (10-5/(4*pi))*cos((1.2*x1+0.4))+ 10) + 2*(1.2*x1+1.9) - 3*(3*(1.2*x2+2.4)-1) - 1 - 3*x2 + 1
#'   }else if(l == 2){
#'     10*sqrt((-1.275*(x1+2)^2/pi^2+5*(x1+2)/pi+(x2+2)-6)^2 +
#'     (10-5/(4*pi))*cos((x1+2))+ 10) + 2*(x1-0.5) - 3*(3*x2-1) - 1
#'   }else if(l == 3){
#'     (-1.275*x1^2/pi^2+5*x1/pi+x2-6)^2 + (10-5/(4*pi))*cos(x1)+ 10
#'   }
#' }
#'
#' output.branin <- function(x, l){
#'   factor_range <- list("x1" = c(-5, 10), "x2" = c(0, 15))
#'
#'   for(i in 1:length(factor_range)) x[i] <- factor_range[[i]][1] + x[i] * diff(factor_range[[i]])
#'   branin(x[1:2], l)
#' }
#'
#' ### training data ###
#' n1 <- 20; n2 <- 15; n3 <- 10
#'
#' ### fix seed to reproduce the result ###
#' set.seed(1)
#'
#' ### generate initial nested design ###
#' X <- NestedX(c(n1, n2, n3), 2)
#' X1 <- X[[1]]
#' X2 <- X[[2]]
#' X3 <- X[[3]]
#'
#' y1 <- apply(X1,1,output.branin, l=1)
#' y2 <- apply(X2,1,output.branin, l=2)
#' y3 <- apply(X3,1,output.branin, l=3)
#'
#' ### n=10000 grid test data ###
#' x <- as.matrix(expand.grid(seq(0, 1, length.out = 100),seq(0, 1, length.out = 100)))
#'
#' ### fit an RNAmf ###
#' fit.RNAmf <- RNAmf(list(X1, X2, X3), list(y1, y2, y3), kernel = "sqex", constant=TRUE)
#'
#' ### predict ###
#' pred.RNAmf <- predict(fit.RNAmf, x)
#' predy <- pred.RNAmf$mu[[3]]
#' predsig2 <- pred.RNAmf$sig2[[3]]
#'
#' ### RMSE ###
#' print(sqrt(mean((predy - apply(x,1,output.branin, l=3))^2)))
#'
#' ### visualize the emulation performance ###
#' x1 <- x2 <- seq(0, 1, length.out = 100)
#' oldpar <- par(mfrow=c(1,2))
#' image(x1, x2, matrix(apply(x,1,output.branin, l=3), ncol=100),
#' zlim=c(0,310), main="Branin function")
#' image(x1, x2, matrix(predy, ncol=100),
#' zlim=c(0,310), main="RNAmf prediction")
#' par(oldpar)
#'
#' ### predictive variance ###
#' print(predsig2)}

predict.RNAmf <- function(object, x=NULL, nimpute=50, ...) {
  if (is.null(x)) return(fitted(object))
  t1 <- proc.time()
  if (!inherits(object, "RNAmf")) stop("The object is not of class \"RNAmf\" \n")
  if (!is.numeric(x)) stop("'x' must be numeric.")
  if (!is.numeric(nimpute) || length(nimpute) != 1 || nimpute < 1) stop("'nimpute' must be a positive integer.")

  d <- ncol(object$fits[[1]]$X)
  if (is.vector(x)) {
    if (length(x) %% d != 0) {
      stop(paste0("Input 'x' length must be a multiple of the model dimension (", d, ")."))
    }
    x <- matrix(x, ncol = d, byrow = FALSE)
  } else {
    x <- as.matrix(x)
    if (ncol(x) != d) {
      stop(paste0("Input 'x' must have ", d, " columns (matches model dimension)."))
    }
  }
  L <- object$level
  kernel <- object$kernel
  fits <- object$fits

  if(object$nested){
    pred_result <- closed_form_RNA(fits, x, kernel)
    return(list(mu = pred_result$mu, sig2 = pred_result$sig2, time = (proc.time() - t1)[3]))
  }else{
    XX <- object$XX
    yy <- object$yy
    pred1 <- object$pred1
    n <- nrow(x)

    mean_mu   <- vector("list", L); for (l in 1:L) mean_mu[[l]]   <- numeric(n)
    M2_mu     <- vector("list", L); for (l in 1:L) M2_mu[[l]]     <- numeric(n)
    mean_sig2 <- vector("list", L); for (l in 1:L) mean_sig2[[l]] <- numeric(n)

    for (m in 1:nimpute) {
      # Generate imputed dataset for the m-th imputation
      yy <- imputer_RNA(XX, yy, kernel=kernel, pred1, fits)
      pred <- closed_form_RNA(fits, x, kernel, XX, yy)

      for (l in 1:L) {
        mu_new   <- pred$mu[[l]]
        sig2_new <- pred$sig2[[l]]

        delta        <- mu_new - mean_mu[[l]]
        mean_mu[[l]] <- mean_mu[[l]] + delta / m
        M2_mu[[l]]   <- M2_mu[[l]] + delta * (mu_new - mean_mu[[l]])

        # simple running mean for sig2
        mean_sig2[[l]] <- mean_sig2[[l]] + (sig2_new - mean_sig2[[l]]) / m
      }
    }

    mu_star   <- mean_mu
    sig2_star <- vector("list", L)
    for (l in 1:L) {
      var_mu_pop   <- if (nimpute > 0) M2_mu[[l]] / nimpute else numeric(n)
      sig2_star[[l]] <- mean_sig2[[l]] + var_mu_pop
    }
    names(mu_star)   <- paste0("mu_",   1:L)
    names(sig2_star) <- paste0("sig2_", 1:L)

    return(list(mu = mu_star, sig2 = sig2_star, time = (proc.time() - t1)[3]))
  }
}

#' calculating Matern kernel with corresponding smoothness parameter
#'
#'
#' @param r vector or matrix of input.
#' @param nu numerical value of smoothness hyperparameter. It should be 0.5, 1.5, 2.5, 3.5, or 4.5.
#' @param derivative logical indicating for its first derivative(derivative=1)
#' @noRd
#' @keywords internal
#' @return A value from matern kernel.

matern.kernel <- function(r, nu, derivative = 0) {
  if (nu == 1 / 2) { # nu=0.5
    if (derivative == 0) out <- exp(-r)
    if (derivative == 1) out <- -exp(-r)
  } else if (nu == 3 / 2) { # nu=1.5
    if (derivative == 0) out <- (1 + r * sqrt(3)) * exp(-r * sqrt(3))
    if (derivative == 1) out <- -3 * r * exp(-sqrt(3) * r)
  } else if (nu == 5 / 2) { # nu=2.5
    if (derivative == 0) out <- (1 + r * sqrt(5) + 5 * r^2 / 3) * exp(-r * sqrt(5))
    if (derivative == 1) out <- -(r * (5^(3 / 2) * r + 5) * exp(-sqrt(5) * r)) / 3
  } else if (nu == 7 / 2) { # nu=3.5
    if (derivative == 0) out <- (1 + r * sqrt(7) + 2.8 * r^2 + 7 / 15 * sqrt(7) * r^3) * exp(-r * sqrt(7))
    if (derivative == 1) out <- -(r * (49 * r^2 + 3 * 7^(3 / 2) * r + 21) * exp(-sqrt(7) * r)) / 15
  } else if (nu == 9 / 2) { # nu=4.5
    if (derivative == 0) out <- (1 + r * sqrt(9) + 27 * r^2 / 7 + 18 / 7 * r^3 + 27 / 35 * r^4) * exp(-r * 3)
    if (derivative == 1) out <- -((81 * r^4 + 162 * r^3 + 135 * r^2 + 45 * r) * exp(-3 * r)) / 35
  }

  return(out)
}

#' calculating separable matern kernel
#'
#' @param X vector or matrix of input.
#' @param x vector or matrix of new input. Default is NULL
#' @param theta lengthscale parameter. It should have the length of ncol(X).
#' @param nu numerical value of smoothness hyperparameter. It should be 0.5, 1.5, 2.5, 3.5, or 4.5.
#' @param derivative logical indicating for its first derivative(derivative=1)
#' @noRd
#' @keywords internal
#' @return A covariance matrix of matern kernel.

cor.sep <- function(X, x = NULL, theta, nu, derivative = 0) {
  if (is.null(dim(X))) X <- matrix(X, ncol = 1)
  d <- NCOL(X)
  n <- NROW(X)
  nu <- rep(nu, d)
  if (is.null(x)) {
    K <- matrix(1, n, n)
    for (i in 1:d) {
      R <- sqrt(distance(X[, i] / theta[i]))
      K <- K * matern.kernel(R, nu = nu[i], derivative = derivative)
    }
  } else {
    n.new <- NROW(x)
    K <- matrix(1, n, n.new)
    for (i in 1:d) {
      R <- sqrt(distance(X[, i] / theta[i], x[, i] / theta[i]))
      K <- K * matern.kernel(R, nu = nu[i], derivative = derivative)
    }
  }
  return(K)
}

#' fitting the model with Matern kernel.
#'
#' @details The choice of bounds for the optimization follows the approach used in \pkg{hetGP}.
#' For more details, see the reference below.
#'
#' @references
#' M. Binois and R. B. Gramacy (2021). hetGP: Heteroskedastic Gaussian Process Modeling and Sequential Design in R.
#' \emph{Journal of Statistical Software}, 98(13), 1-44;
#' \doi{doi: 10.18637/jss.v098.i13}
#'
#' @param X vector or matrix of input locations.
#' @param y vector of response values.
#' @param g nugget parameter. Default is 1.490116e-08.
#' @param nu numerical value of smoothness hyperparameter. It should be 0.5, 1.5, 2.5, 3.5, or 4.5.
#' @param constant logical indicating for constant mean (constant=TRUE) or zero mean (constant=FALSE). Default is FALSE.
#' @param p quantile on distances. Default is 0.1.
#' @param min_cor minimal correlation between two design points at the defined quantile distance. Default is 0.01.
#' @param max_cor maximal correlation between two design points at the defined (1-p) quantile distance. Default is 0.5.
#' @param init initial value of optimization. Default is \code{NULL}.
#' @param lower lower bound of optimization. Default is \code{NULL}.
#' @param upper upper bound of optimization. Default is \code{NULL}.
#'
#' @return A list containing hyperparameters, covariance inverse matrix, X, y and logical inputs:
#' \itemize{
#'   \item \code{K}: matrix of covariance.
#'   \item \code{Ki}: matrix of covariance inverse.
#'   \item \code{X}: copy of X.
#'   \item \code{y}: copy of y.
#'   \item \code{theta}: vector of lengthscale hyperparameter.
#'   \item \code{nu}: copy of nu.
#'   \item \code{g}: copy of g.
#'   \item \code{mu.hat}: optimized constant mean. If constant=FALSE, 0.
#'   \item \code{tau2hat}: estimated scale hyperparameter.
#'   \item \code{constant}: copy of constant.
#' }
#'
#' @importFrom stats optim quantile uniroot
#' @importFrom methods is
#' @noRd
#' @keywords internal
#' @examples
#' \dontrun{
#' library(lhs)
#' ### synthetic function ###
#' f1 <- function(x) {
#'   sin(8 * pi * x)
#' }
#'
#' ### training data ###
#' n1 <- 15
#'
#' X1 <- maximinLHS(n1, 1)
#' y1 <- f1(X1)
#'
#' matGP(X1, y1, nu = 2.5)
#' }
matGP <- function(X, y, nu = 2.5, g = sqrt(.Machine$double.eps), constant = FALSE, p=0.1, min_cor = 0.01, max_cor = 0.5, init=NULL, lower=NULL, upper=NULL) {
  if (!nu %in% c(0.5, 1.5, 2.5, 3.5, 4.5)) {
    stop("matGP: 'nu' must be one of 0.5, 1.5, 2.5, 3.5, or 4.5.")
  }
  if (!is.numeric(p) || length(p) != 1 || p <= 0 || p >= 1) {
    stop("GP: 'p' must be a single numeric value between 0 and 1.")
  }
  if (!is.numeric(min_cor) || length(min_cor) != 1 || min_cor <= 0 || min_cor >= 1) {
    stop("GP: 'min_cor' must be a single numeric value between 0 and 1.")
  }
  if (!is.numeric(max_cor) || length(max_cor) != 1 || max_cor <= 0 || max_cor >= 1) {
    stop("GP: 'max_cor' must be a single numeric value between 0 and 1.")
  }
  if (min_cor >= max_cor) {
    stop("GP: 'min_cor' must be strictly less than 'max_cor'.")
  }
  if (!is.null(init) && !is.numeric(init)) stop("GP: 'init' must be numeric.")
  if (!is.null(lower) && !is.numeric(lower)) stop("GP: 'lower' must be numeric.")
  if (!is.null(upper) && !is.numeric(upper)) stop("GP: 'upper' must be numeric.")

  if (is.null(dim(X))) X <- matrix(X, ncol = 1)
  y <- as.matrix(y)
  n <- nrow(X)

  Xscaled <- crossprod(t(X - matrix(apply(X, 2, range)[1,], nrow = nrow(X), ncol = ncol(X), byrow = TRUE)),
                       diag(1/(apply(X, 2, range)[2,] - apply(X, 2, range)[1,]), ncol(X)))
  tmpfun <- function(theta, repr_dist, value){
    cor.sep(matrix(sqrt(repr_dist/ncol(X)), ncol = ncol(X)), matrix(0, ncol = ncol(X)), theta = rep(theta,ncol(X)), nu = nu) - value
  }
  theta_min <- try(uniroot(tmpfun, interval = c(g, 100), value = min_cor,
                           repr_dist = quantile(distance(Xscaled)[lower.tri(distance(Xscaled))], p), tol = g)$root)
  if(is(theta_min, "try-error")){
    warning("The automatic selection of lengthscales bounds was not successful. Perhaps provide lower and upper values.")
    theta_min <- 1e-2
  }
  theta_max <- try(uniroot(tmpfun, interval = c(g, 100), value = max_cor,
                           repr_dist = quantile(distance(Xscaled)[lower.tri(distance(Xscaled))], 1-p), tol = g)$root, silent = TRUE)
  if(is(theta_max, "try-error")){
    theta_max <- 5
  }
  if(is.null(lower)) lower <- theta_min * (apply(X, 2, range)[2,] - apply(X, 2, range)[1,])
  if(is.null(upper)) upper <- max(1, theta_max) * (apply(X, 2, range)[2,] - apply(X, 2, range)[1,])
  if(is.null(init)) init <- sqrt(lower * upper)

  if (constant) {
    one.vec <- matrix(1, ncol = 1, nrow = n)

    nlsep <- function(par, X, Y) {
      theta <- par
      K <- cor.sep(X, theta = theta, nu = nu)
      chol_K <- chol(K)
      ldetK <- 2 * sum(log(diag(chol_K)))

      KinvY <- backsolve(chol_K, forwardsolve(t(chol_K), Y))
      Kinv1 <- backsolve(chol_K, forwardsolve(t(chol_K), one.vec))
      mu.hat <- drop(crossprod(one.vec, KinvY) / crossprod(one.vec, Kinv1))
      tau2hat <- drop(crossprod(Y - mu.hat, KinvY - mu.hat * Kinv1) / n)

      ll <- -(n / 2) * log(tau2hat) - (1 / 2) * ldetK
      return(drop(-ll))
    }

    gradnlsep <- function(par, X, Y) {
      theta <- par
      K <- cor.sep(X, theta = theta, nu = nu)
      chol_K <- chol(K)
      Ki <- chol2inv(chol_K)
      mu.hat <- drop(crossprod(one.vec, crossprod(Ki, Y)) / crossprod(one.vec, crossprod(Ki, one.vec)))

      Ki_resid <- crossprod(Ki, Y - mu.hat)

      dlltheta <- rep(NA, length(theta))
      for (k in 1:length(dlltheta)) {
        if(nu==1.5){
          u_k <- sqrt(distance(X[, k] / theta[k]))
          dotK <- K * (3 * u_k^2 / theta[k]) / (1 + sqrt(3) * u_k)
        }else{
          u_k <- sqrt(distance(X[, k] / theta[k]))
          dotK <- K * (5/3 * u_k^2 / theta[k]) * (1+sqrt(5)*u_k) / (1 + sqrt(5) * u_k + 5/3*u_k^2)
        }
        dlltheta[k] <- (n / 2) * drop(crossprod(Ki_resid, crossprod(dotK, Ki_resid))) / drop(crossprod(Y - mu.hat, Ki_resid)) - (1 / 2) * sum(Ki * dotK)
      }
      return(-c(dlltheta))
    }

    outg <- optim(init, nlsep, gradnlsep, method = "L-BFGS-B", lower = lower, upper = upper, X = X, Y = y)

    theta <- outg$par
    K <- cor.sep(X, theta = theta, nu = nu)
    Ki <- chol2inv(chol(K))
    mu.hat <- drop(crossprod(one.vec, crossprod(Ki, y)) / crossprod(one.vec, crossprod(Ki, one.vec)))
    tau2hat <- drop(crossprod((y - mu.hat), crossprod(Ki, (y - mu.hat))) / n)
    names(theta) <- NULL

    return(list(K = K, Ki = Ki, X = X, y = y, theta = theta, nu = nu, g = g, mu.hat = mu.hat, tau2hat = tau2hat, constant = constant))
  } else {

    nlsep <- function(par, X, Y) {
      theta <- par # lengthscale
      K <- cor.sep(X, theta = theta, nu = nu)
      chol_K <- chol(K)
      ldetK <- 2 * sum(log(diag(chol_K)))

      KinvY <- backsolve(chol_K, forwardsolve(t(chol_K), Y))

      ll <- -(n / 2) * log(drop(crossprod(Y, KinvY))) - (1 / 2) * ldetK
      return(drop(-ll))
    }

    gradnlsep <- function(par, X, Y) {
      theta <- par
      K <- cor.sep(X, theta = theta, nu = nu)
      chol_K <- chol(K)
      Ki <- chol2inv(chol_K)
      KiY <- crossprod(Ki, Y)

      dlltheta <- rep(NA, length(theta))
      for (k in 1:length(dlltheta)) {
        if(nu==1.5){
          u_k <- sqrt(distance(X[, k] / theta[k]))
          dotK <- K * (3 * u_k^2 / theta[k]) / (1 + sqrt(3) * u_k)
        }else{
          u_k <- sqrt(distance(X[, k] / theta[k]))
          dotK <- K * (5/3 * u_k^2 / theta[k]) * (1+sqrt(5)*u_k) / (1 + sqrt(5) * u_k + 5/3*u_k^2)
        }
        dlltheta[k] <- (n / 2) * drop(crossprod(KiY, crossprod(dotK, KiY))) / drop(crossprod(Y, KiY)) - (1 / 2) * sum(Ki * dotK)
      }
      return(-c(dlltheta))
    }

    outg <- optim(init, nlsep, gradnlsep, method = "L-BFGS-B", lower = lower, upper = upper, X = X, Y = y)

    theta <- outg$par
    K <- cor.sep(X, theta = outg$par, nu = nu)
    Ki <- chol2inv(chol(K))
    mu.hat <- 0
    tau2hat <- drop(crossprod(y, crossprod(Ki, y)) / n)
    names(theta) <- NULL

    return(list(K = K, Ki = Ki, X = X, y = y, theta = theta, nu = nu, g = g, mu.hat = mu.hat, tau2hat = tau2hat, constant = constant))
  }
}

#' predictive posterior mean and variance with matern kernel.
#'
#' @param fit an object of class matGP.
#' @param xnew vector or matrix of new input locations to predict.
#' @param cov.out logical indicating for returning covariance matrix (cov.out=TRUE) or not (cov.out=FALSE). Default is FALSE.
#'
#' @return A list predictive posterior mean and variance:
#' \itemize{
#'   \item \code{mu}: vector of predictive posterior mean.
#'   \item \code{cov}: matrix of predictive posterior covariance.
#'   \item \code{sig2}: vector of predictive posterior variance.
#' }
#'
#' @noRd
#' @keywords internal
#' @examples
#' \dontrun{
#' library(lhs)
#' ### synthetic function ###
#' f1 <- function(x) {
#'   sin(8 * pi * x)
#' }
#'
#' ### training data ###
#' n1 <- 15
#'
#' X1 <- maximinLHS(n1, 1)
#' y1 <- f1(X1)
#'
#' fit1 <- matGP(X1, y1, nu = 2.5)
#'
#' ### test data ###
#' x <- seq(0, 1, 0.01)
#' pred.matGP(fit1, x)
#' }
pred.matGP <- function(fit, xnew, cov.out=FALSE) {
  if(is.null(fit$Ki) || is.null(fit$theta)) stop("Invalid fit object passed to pred.matGP")
  xnew <- as.matrix(xnew)

  Ki <- fit$Ki
  theta <- fit$theta
  nu <- fit$nu
  g <- fit$g
  X <- fit$X
  y <- fit$y
  tau2hat <- fit$tau2hat
  mu.hat <- fit$mu.hat

  KXX <- cor.sep(xnew, theta = theta, nu = nu)
  KX <- t(cor.sep(X, xnew, theta = theta, nu = nu))

  mup2 <- mu.hat + crossprod(t(KX), crossprod(Ki, y - mu.hat))
  Sigmap2 <- tau2hat * (KXX - crossprod(t(KX), tcrossprod(Ki, KX)))
  Sigmap2 <- (t(Sigmap2) + Sigmap2) / 2
  if(cov.out){
    return(list(mu = mup2, cov = Sigmap2, sig2 = pmax(0, diag(Sigmap2))))
  }else{
    return(list(mu = mup2, sig2 = pmax(0, diag(Sigmap2))))
  }
}

#' object to optimize the point by ALM criterion updating at level 1 with two levels of fidelity
#'
#' @param Xcand candidate data point to be optimized.
#' @param fit an object of class RNAmf.
#' @return A negative predictive posterior variance at Xcand.
#' @noRd
#'

obj.ALM_level_1 <- function(fit, Xcand) {
  newx <- matrix(Xcand, nrow = 1)
  fit1 <- fit$fits[[1]]
  kernel <- fit$kernel
  
  ### calculate the posterior predictive variance ###
  if (kernel == "sqex") {
    predsig2 <- pred.GP(fit1, newx)$sig2
  } else if (kernel == "matern1.5") {
    predsig2 <- pred.matGP(fit1, newx)$sig2
  } else if (kernel == "matern2.5") {
    predsig2 <- pred.matGP(fit1, newx)$sig2
  }
  
  -predsig2 # to maximize the current variance.
}


#' object to optimize the point by ALM criterion updating at level 2 with two levels of fidelity
#'
#' @param Xcand candidate data point to be optimized.
#' @param fit an object of class RNAmf.
#' @return A negative predictive posterior variance at Xcand.
#' @noRd
#'

obj.ALM_level_2 <- function(fit, Xcand) {
  newx <- matrix(Xcand, nrow = 1)
  -predict(fit, newx)$sig2[[2]] # to maximize the current variance.
}


#' object to optimize the point by ALM criterion updating at level 3 with three levels of fidelity
#'
#' @param Xcand candidate data point to be optimized.
#' @param fit an object of class RNAmf.
#' @return A predictive posterior variance at Xcand.
#' @noRd
#'

obj.ALM_level_3 <- function(fit, Xcand) {
  newx <- matrix(Xcand, nrow = 1)
  -predict(fit, newx)$sig2[[3]] # to maximize the current variance.
}


#' object to optimize the point by ALD criterion updating V1 with two levels of fidelity
#'
#' @param Xcand candidate data point to be optimized.
#' @param fit an object of class RNAmf.
#' @return A negative V1 at Xcand.
#' @noRd
#'

obj.ALD_V1_2level <- function(fit, Xcand) { # low
  newx <- matrix(Xcand, nrow = 1)
  
  kernel <- fit$kernel
  constant <- fit$constant
  fit1 <- fit$fits[[1]]
  fit2 <- fit$fits[[2]]
  
  if (kernel == "sqex") {
    if (constant) {
      d <- ncol(fit1$X)
      newx <- matrix(newx, ncol = d)
      pred.fit <- pred.GP(fit1, newx)
      x.mu <- pred.fit$mu # mean of f1(u)
      sig2 <- pred.fit$sig2 #* 0
      
      ### calculate the closed form ###
      X2 <- matrix(fit2$X[, -(d + 1)], ncol = d)
      w1.x2 <- fit2$X[, d + 1]
      y2 <- fit2$y
      n <- length(y2)
      theta <- fit2$theta
      tau2hat <- fit2$tau2hat
      mu2 <- fit2$mu.hat
      
      Ci <- fit2$Ki
      a <- crossprod(Ci, y2 - mu2)
      
      # mean
      predy <- mu2 + crossprod(t(exp(-distance(t(t(newx) / sqrt(theta[-(d + 1)])), t(t(X2) / sqrt(theta[-(d + 1)])))) *
                                   1 / sqrt(1 + 2 * sig2 / theta[d + 1]) *
                                   exp(-(drop(outer(x.mu, w1.x2, FUN = "-")))^2 / (theta[d + 1] + 2 * sig2))), a)
      
      # var
      mat <- drop(crossprod(exp(-distance(t(t(newx) / sqrt(theta[-(d + 1)])), t(t(X2) / sqrt(theta[-(d + 1)])))),
                            exp(-distance(t(t(newx) / sqrt(theta[-(d + 1)])), t(t(X2) / sqrt(theta[-(d + 1)])))))) * # common components
        1 / sqrt(1 + 4 * sig2 / theta[d + 1]) *
        exp(-(outer(w1.x2, w1.x2, FUN = "+") / 2 - matrix(x.mu, n, n))^2 / (theta[d + 1] / 2 + 2 * sig2)) *
        exp(-(outer(w1.x2, w1.x2, FUN = "-"))^2 / (2 * theta[d + 1]))
      
      VE <- - (predy - mu2)^2 + drop(crossprod(a, crossprod(mat, a)))
    } else {
      d <- ncol(fit1$X)
      newx <- matrix(newx, ncol = d)
      pred.fit <- pred.GP(fit1, newx)
      x.mu <- pred.fit$mu # mean of f1(u)
      sig2 <- pred.fit$sig2
      
      ### calculate the closed form ###
      X2 <- matrix(fit2$X[, -(d + 1)], ncol = d)
      w1.x2 <- fit2$X[, d + 1]
      y2 <- fit2$y
      n <- length(y2)
      theta <- fit2$theta
      tau2hat <- fit2$tau2hat
      
      Ci <- fit2$Ki
      a <- crossprod(Ci, y2)
      
      # mean
      predy <- crossprod(t(exp(-distance(t(t(newx) / sqrt(theta[-(d + 1)])), t(t(X2) / sqrt(theta[-(d + 1)])))) *
                             1 / sqrt(1 + 2 * sig2 / theta[d + 1]) *
                             exp(-(drop(outer(x.mu, w1.x2, FUN = "-")))^2 / (theta[d + 1] + 2 * sig2))), a)
      
      # var
      mat <- drop(crossprod(exp(-distance(t(t(newx) / sqrt(theta[-(d + 1)])), t(t(X2) / sqrt(theta[-(d + 1)])))),
                            exp(-distance(t(t(newx) / sqrt(theta[-(d + 1)])), t(t(X2) / sqrt(theta[-(d + 1)])))))) * # common components
        1 / sqrt(1 + 4 * sig2 / theta[d + 1]) *
        exp(-(outer(w1.x2, w1.x2, FUN = "+") / 2 - matrix(x.mu, n, n))^2 / (theta[d + 1] / 2 + 2 * sig2)) *
        exp(-(outer(w1.x2, w1.x2, FUN = "-"))^2 / (2 * theta[d + 1]))
      
      VE <- - predy^2 + drop(crossprod(a, crossprod(mat, a)))
    }
  } else if (kernel == "matern1.5") {
    if (constant) {
      d <- ncol(fit1$X)
      newx <- matrix(newx, ncol = d)
      pred.fit <- pred.matGP(fit1, newx)
      x.mu <- drop(pred.fit$mu) # mean of f1(u)
      sig2 <- pred.fit$sig2 #* 0
      
      ### calculate the closed form ###
      X2 <- matrix(fit2$X[, -(d + 1)], ncol = d)
      w1.x2 <- fit2$X[, d + 1]
      y2 <- fit2$y
      n <- length(y2)
      theta <- fit2$theta
      tau2hat <- fit2$tau2hat
      mu2 <- fit2$mu.hat
      
      Ci <- fit2$Ki
      a <- crossprod(Ci, y2 - mu2)
      
      # mean
      mua <- x.mu - sqrt(3) * sig2 / theta[d + 1]
      mub <- x.mu + sqrt(3) * sig2 / theta[d + 1]
      
      lambda11 <- c(1, mua)
      lambda12 <- c(0, 1)
      lambda21 <- c(1, -mub)
      
      e1 <- cbind(matrix(1 - sqrt(3) * w1.x2 / theta[d + 1]), sqrt(3) / theta[d + 1])
      e2 <- cbind(matrix(1 + sqrt(3) * w1.x2 / theta[d + 1]), sqrt(3) / theta[d + 1])
      
      predy <- mu2 + drop(crossprod(t(cor.sep(t(newx), X2, theta[-(d + 1)], nu = 1.5)) * # common but depends on kernel
                                      (exp((3 * sig2 + 2 * sqrt(3) * theta[d + 1] * (w1.x2 - x.mu)) / (2 * theta[d + 1]^2)) *
                                         (crossprod(t(e1), lambda11) * pnorm((mua - w1.x2) / sqrt(sig2)) +
                                            crossprod(t(e1), lambda12) * sqrt(sig2) / sqrt(2 * pi) * exp(-(w1.x2 - mua)^2 / (2 * sig2))) +
                                         exp((3 * sig2 - 2 * sqrt(3) * theta[d + 1] * (w1.x2 - x.mu)) / (2 * theta[d + 1]^2)) *
                                         (crossprod(t(e2), lambda21) * pnorm((-mub + w1.x2) / sqrt(sig2)) +
                                            crossprod(t(e2), lambda12) * sqrt(sig2) / sqrt(2 * pi) * exp(-(w1.x2 - mub)^2 / (2 * sig2)))), a))
      
      # var
      zeta <- function(x, y) {
        zetafun(w1 = x, w2 = y, m = x.mu, s = sig2, nu = 1.5, theta = theta[d + 1])
      }
      mat <- drop(crossprod(cor.sep(t(newx), X2, theta[-(d + 1)], nu = 1.5), cor.sep(t(newx), X2, theta[-(d + 1)], nu = 1.5))) * # constant depends on kernel
        outer(w1.x2, w1.x2, FUN = Vectorize(zeta))
      
      VE <- - (predy - mu2)^2 + drop(crossprod(a, crossprod(mat, a)))
    } else {
      d <- ncol(fit1$X)
      newx <- matrix(newx, ncol = d)
      pred.fit <- pred.matGP(fit1, newx)
      x.mu <- drop(pred.fit$mu) # mean of f1(u)
      sig2 <- pred.fit$sig2
      
      ### calculate the closed form ###
      X2 <- matrix(fit2$X[, -(d + 1)], ncol = d)
      w1.x2 <- fit2$X[, d + 1]
      y2 <- fit2$y
      n <- length(y2)
      theta <- fit2$theta
      tau2hat <- fit2$tau2hat
      
      Ci <- fit2$Ki
      a <- crossprod(Ci, y2)
      
      # mean
      mua <- x.mu - sqrt(3) * sig2 / theta[d + 1]
      mub <- x.mu + sqrt(3) * sig2 / theta[d + 1]
      
      lambda11 <- c(1, mua)
      lambda12 <- c(0, 1)
      lambda21 <- c(1, -mub)
      
      e1 <- cbind(matrix(1 - sqrt(3) * w1.x2 / theta[d + 1]), sqrt(3) / theta[d + 1])
      e2 <- cbind(matrix(1 + sqrt(3) * w1.x2 / theta[d + 1]), sqrt(3) / theta[d + 1])
      
      predy <- drop(crossprod(t(cor.sep(t(newx), X2, theta[-(d + 1)], nu = 1.5)) * # common but depends on kernel
                                (exp((3 * sig2 + 2 * sqrt(3) * theta[d + 1] * (w1.x2 - x.mu)) / (2 * theta[d + 1]^2)) *
                                   (crossprod(t(e1), lambda11) * pnorm((mua - w1.x2) / sqrt(sig2)) +
                                      crossprod(t(e1), lambda12) * sqrt(sig2) / sqrt(2 * pi) * exp(-(w1.x2 - mua)^2 / (2 * sig2))) +
                                   exp((3 * sig2 - 2 * sqrt(3) * theta[d + 1] * (w1.x2 - x.mu)) / (2 * theta[d + 1]^2)) *
                                   (crossprod(t(e2), lambda21) * pnorm((-mub + w1.x2) / sqrt(sig2)) +
                                      crossprod(t(e2), lambda12) * sqrt(sig2) / sqrt(2 * pi) * exp(-(w1.x2 - mub)^2 / (2 * sig2)))), a))
      # var
      zeta <- function(x, y) {
        zetafun(w1 = x, w2 = y, m = x.mu, s = sig2, nu = 1.5, theta = theta[d + 1])
      }
      mat <- drop(crossprod(cor.sep(t(newx), X2, theta[-(d + 1)], nu = 1.5), cor.sep(t(newx), X2, theta[-(d + 1)], nu = 1.5))) * # constant depends on kernel
        outer(w1.x2, w1.x2, FUN = Vectorize(zeta))
      
      VE <- - predy^2 + drop(crossprod(a, crossprod(mat, a)))
    }
  } else if (kernel == "matern2.5") {
    if (constant) {
      d <- ncol(fit1$X)
      newx <- matrix(newx, ncol = d)
      pred.fit <- pred.matGP(fit1, newx)
      x.mu <- drop(pred.fit$mu) # mean of f1(u)
      sig2 <- pred.fit$sig2 #* 0
      
      ### calculate the closed form ###
      X2 <- matrix(fit2$X[, -(d + 1)], ncol = d)
      w1.x2 <- fit2$X[, d + 1]
      y2 <- fit2$y
      n <- length(y2)
      theta <- fit2$theta
      tau2hat <- fit2$tau2hat
      mu2 <- fit2$mu.hat
      
      Ci <- fit2$Ki
      a <- crossprod(Ci, y2 - mu2)
      
      # mean
      mua <- x.mu - sqrt(5) * sig2 / theta[d + 1]
      mub <- x.mu + sqrt(5) * sig2 / theta[d + 1]
      
      lambda11 <- c(1, mua, mua^2 + sig2)
      lambda12 <- cbind(0, 1, matrix(mua + w1.x2))
      lambda21 <- c(1, -mub, mub^2 + sig2)
      lambda22 <- cbind(0, 1, matrix(-mub - w1.x2))
      
      e1 <- cbind(
        matrix(1 - sqrt(5) * w1.x2 / theta[d + 1] + 5 * w1.x2^2 / (3 * theta[d + 1]^2)),
        matrix(sqrt(5) / theta[d + 1] - 10 * w1.x2 / (3 * theta[d + 1]^2)),
        5 / (3 * theta[d + 1]^2)
      )
      e2 <- cbind(
        matrix(1 + sqrt(5) * w1.x2 / theta[d + 1] + 5 * w1.x2^2 / (3 * theta[d + 1]^2)),
        matrix(sqrt(5) / theta[d + 1] + 10 * w1.x2 / (3 * theta[d + 1]^2)),
        5 / (3 * theta[d + 1]^2)
      )
      
      predy <- mu2 + drop(crossprod(t(cor.sep(t(newx), X2, theta[-(d + 1)], nu = 2.5)) *
                                      (exp((5 * sig2 + 2 * sqrt(5) * theta[d + 1] * (w1.x2 - x.mu)) / (2 * theta[d + 1]^2)) *
                                         (crossprod(t(e1), lambda11) * pnorm((mua - w1.x2) / sqrt(sig2)) +
                                            rowSums(e1 * lambda12) * sqrt(sig2) / sqrt(2 * pi) * exp(-(w1.x2 - mua)^2 / (2 * sig2))) +
                                         exp((5 * sig2 - 2 * sqrt(5) * theta[d + 1] * (w1.x2 - x.mu)) / (2 * theta[d + 1]^2)) *
                                         (crossprod(t(e2), lambda21) * pnorm((-mub + w1.x2) / sqrt(sig2)) +
                                            rowSums(e2 * lambda22) * sqrt(sig2) / sqrt(2 * pi) * exp(-(w1.x2 - mub)^2 / (2 * sig2)))), a))
      
      # var
      zeta <- function(x, y) {
        zetafun(w1 = x, w2 = y, m = x.mu, s = sig2, nu = 2.5, theta = theta[d + 1])
      }
      mat <- drop(crossprod(cor.sep(t(newx), X2, theta[-(d + 1)], nu = 2.5), cor.sep(t(newx), X2, theta[-(d + 1)], nu = 2.5))) * # constant depends on kernel
        outer(w1.x2, w1.x2, FUN = Vectorize(zeta))
      
      VE <- - (predy - mu2)^2 + drop(crossprod(a, crossprod(mat, a)))
    } else {
      d <- ncol(fit1$X)
      newx <- matrix(newx, ncol = d)
      pred.fit <- pred.matGP(fit1, newx)
      x.mu <- drop(pred.fit$mu) # mean of f1(u)
      sig2 <- pred.fit$sig2
      
      ### calculate the closed form ###
      X2 <- matrix(fit2$X[, -(d + 1)], ncol = d)
      w1.x2 <- fit2$X[, d + 1]
      y2 <- fit2$y
      n <- length(y2)
      theta <- fit2$theta
      tau2hat <- fit2$tau2hat
      
      Ci <- fit2$Ki
      a <- crossprod(Ci, y2)
      
      # mean
      mua <- x.mu - sqrt(5) * sig2 / theta[d + 1]
      mub <- x.mu + sqrt(5) * sig2 / theta[d + 1]
      
      lambda11 <- c(1, mua, mua^2 + sig2)
      lambda12 <- cbind(0, 1, matrix(mua + w1.x2))
      lambda21 <- c(1, -mub, mub^2 + sig2)
      lambda22 <- cbind(0, 1, matrix(-mub - w1.x2))
      
      e1 <- cbind(
        matrix(1 - sqrt(5) * w1.x2 / theta[d + 1] + 5 * w1.x2^2 / (3 * theta[d + 1]^2)),
        matrix(sqrt(5) / theta[d + 1] - 10 * w1.x2 / (3 * theta[d + 1]^2)),
        5 / (3 * theta[d + 1]^2)
      )
      e2 <- cbind(
        matrix(1 + sqrt(5) * w1.x2 / theta[d + 1] + 5 * w1.x2^2 / (3 * theta[d + 1]^2)),
        matrix(sqrt(5) / theta[d + 1] + 10 * w1.x2 / (3 * theta[d + 1]^2)),
        5 / (3 * theta[d + 1]^2)
      )
      
      predy <- drop(crossprod(t(cor.sep(t(newx), X2, theta[-(d + 1)], nu = 2.5)) *
                                (exp((5 * sig2 + 2 * sqrt(5) * theta[d + 1] * (w1.x2 - x.mu)) / (2 * theta[d + 1]^2)) *
                                   (crossprod(t(e1), lambda11) * pnorm((mua - w1.x2) / sqrt(sig2)) +
                                      rowSums(e1 * lambda12) * sqrt(sig2) / sqrt(2 * pi) * exp(-(w1.x2 - mua)^2 / (2 * sig2))) +
                                   exp((5 * sig2 - 2 * sqrt(5) * theta[d + 1] * (w1.x2 - x.mu)) / (2 * theta[d + 1]^2)) *
                                   (crossprod(t(e2), lambda21) * pnorm((-mub + w1.x2) / sqrt(sig2)) +
                                      rowSums(e2 * lambda22) * sqrt(sig2) / sqrt(2 * pi) * exp(-(w1.x2 - mub)^2 / (2 * sig2)))), a))
      
      # var
      zeta <- function(x, y) {
        zetafun(w1 = x, w2 = y, m = x.mu, s = sig2, nu = 2.5, theta = theta[d + 1])
      }
      mat <- drop(crossprod(cor.sep(t(newx), X2, theta[-(d + 1)], nu = 2.5), cor.sep(t(newx), X2, theta[-(d + 1)], nu = 2.5))) * # constant depends on kernel
        outer(w1.x2, w1.x2, FUN = Vectorize(zeta))
      
      VE <- - predy^2 + drop(crossprod(a, crossprod(mat, a)))
    }
  }
  return(-VE) # to maximize the V1.
}


#' object to optimize the point by ALD criterion updating V2 with two levels of fidelity
#'
#' @param Xcand candidate data point to be optimized.
#' @param fit an object of class RNAmf.
#' @return A negative V2 at Xcand.
#' @noRd
#'

obj.ALD_V2_2level <- function(fit, Xcand) { # high
  newx <- matrix(Xcand, nrow = 1)
  
  kernel <- fit$kernel
  constant <- fit$constant
  fit1 <- fit$fits[[1]]
  fit2 <- fit$fits[[2]]
  
  if (kernel == "sqex") {
    if (constant) {
      d <- ncol(fit1$X)
      newx <- matrix(newx, ncol = d)
      pred.fit <- pred.GP(fit1, newx)
      x.mu <- pred.fit$mu # mean of f1(u)
      sig2 <- pred.fit$sig2 #* 0
      
      ### calculate the closed form ###
      X2 <- matrix(fit2$X[, -(d + 1)], ncol = d)
      w1.x2 <- fit2$X[, d + 1]
      y2 <- fit2$y
      n <- length(y2)
      theta <- fit2$theta
      tau2hat <- fit2$tau2hat
      mu2 <- fit2$mu.hat
      
      Ci <- fit2$Ki
      a <- crossprod(Ci, y2 - mu2)
      
      # mean
      predy <- mu2 + crossprod(t(exp(-distance(t(t(newx) / sqrt(theta[-(d + 1)])), t(t(X2) / sqrt(theta[-(d + 1)])))) *
                                   1 / sqrt(1 + 2 * sig2 / theta[d + 1]) *
                                   exp(-(drop(outer(x.mu, w1.x2, FUN = "-")))^2 / (theta[d + 1] + 2 * sig2))), a)
      
      # var
      mat <- drop(crossprod(exp(-distance(t(t(newx) / sqrt(theta[-(d + 1)])), t(t(X2) / sqrt(theta[-(d + 1)])))),
                            exp(-distance(t(t(newx) / sqrt(theta[-(d + 1)])), t(t(X2) / sqrt(theta[-(d + 1)])))))) * # common components
        1 / sqrt(1 + 4 * sig2 / theta[d + 1]) *
        exp(-(outer(w1.x2, w1.x2, FUN = "+") / 2 - matrix(x.mu, n, n))^2 / (theta[d + 1] / 2 + 2 * sig2)) *
        exp(-(outer(w1.x2, w1.x2, FUN = "-"))^2 / (2 * theta[d + 1]))
      
      EV <- tau2hat - tau2hat * sum(Ci * mat)
    } else {
      d <- ncol(fit1$X)
      newx <- matrix(newx, ncol = d)
      pred.fit <- pred.GP(fit1, newx)
      x.mu <- pred.fit$mu # mean of f1(u)
      sig2 <- pred.fit$sig2
      
      ### calculate the closed form ###
      X2 <- matrix(fit2$X[, -(d + 1)], ncol = d)
      w1.x2 <- fit2$X[, d + 1]
      y2 <- fit2$y
      n <- length(y2)
      theta <- fit2$theta
      tau2hat <- fit2$tau2hat
      
      Ci <- fit2$Ki
      a <- crossprod(Ci, y2)
      
      # mean
      predy <- crossprod(t(exp(-distance(t(t(newx) / sqrt(theta[-(d + 1)])), t(t(X2) / sqrt(theta[-(d + 1)])))) *
                             1 / sqrt(1 + 2 * sig2 / theta[d + 1]) *
                             exp(-(drop(outer(x.mu, w1.x2, FUN = "-")))^2 / (theta[d + 1] + 2 * sig2))), a)
      
      # var
      mat <- drop(crossprod(exp(-distance(t(t(newx) / sqrt(theta[-(d + 1)])), t(t(X2) / sqrt(theta[-(d + 1)])))),
                            exp(-distance(t(t(newx) / sqrt(theta[-(d + 1)])), t(t(X2) / sqrt(theta[-(d + 1)])))))) * # common components
        1 / sqrt(1 + 4 * sig2 / theta[d + 1]) *
        exp(-(outer(w1.x2, w1.x2, FUN = "+") / 2 - matrix(x.mu, n, n))^2 / (theta[d + 1] / 2 + 2 * sig2)) *
        exp(-(outer(w1.x2, w1.x2, FUN = "-"))^2 / (2 * theta[d + 1]))
      
      EV <- tau2hat - tau2hat * sum(Ci * mat)
    }
  } else if (kernel == "matern1.5") {
    if (constant) {
      d <- ncol(fit1$X)
      newx <- matrix(newx, ncol = d)
      pred.fit <- pred.matGP(fit1, newx)
      x.mu <- drop(pred.fit$mu) # mean of f1(u)
      sig2 <- pred.fit$sig2 #* 0
      
      ### calculate the closed form ###
      X2 <- matrix(fit2$X[, -(d + 1)], ncol = d)
      w1.x2 <- fit2$X[, d + 1]
      y2 <- fit2$y
      n <- length(y2)
      theta <- fit2$theta
      tau2hat <- fit2$tau2hat
      mu2 <- fit2$mu.hat
      
      Ci <- fit2$Ki
      a <- crossprod(Ci, y2 - mu2)
      
      # mean
      mua <- x.mu - sqrt(3) * sig2 / theta[d + 1]
      mub <- x.mu + sqrt(3) * sig2 / theta[d + 1]
      
      lambda11 <- c(1, mua)
      lambda12 <- c(0, 1)
      lambda21 <- c(1, -mub)
      
      e1 <- cbind(matrix(1 - sqrt(3) * w1.x2 / theta[d + 1]), sqrt(3) / theta[d + 1])
      e2 <- cbind(matrix(1 + sqrt(3) * w1.x2 / theta[d + 1]), sqrt(3) / theta[d + 1])
      
      predy <- mu2 + drop(crossprod(t(cor.sep(t(newx), X2, theta[-(d + 1)], nu = 1.5)) * # common but depends on kernel
                                      (exp((3 * sig2 + 2 * sqrt(3) * theta[d + 1] * (w1.x2 - x.mu)) / (2 * theta[d + 1]^2)) *
                                         (crossprod(t(e1), lambda11) * pnorm((mua - w1.x2) / sqrt(sig2)) +
                                            crossprod(t(e1), lambda12) * sqrt(sig2) / sqrt(2 * pi) * exp(-(w1.x2 - mua)^2 / (2 * sig2))) +
                                         exp((3 * sig2 - 2 * sqrt(3) * theta[d + 1] * (w1.x2 - x.mu)) / (2 * theta[d + 1]^2)) *
                                         (crossprod(t(e2), lambda21) * pnorm((-mub + w1.x2) / sqrt(sig2)) +
                                            crossprod(t(e2), lambda12) * sqrt(sig2) / sqrt(2 * pi) * exp(-(w1.x2 - mub)^2 / (2 * sig2)))), a))
      
      # var
      zeta <- function(x, y) {
        zetafun(w1 = x, w2 = y, m = x.mu, s = sig2, nu = 1.5, theta = theta[d + 1])
      }
      mat <- drop(crossprod(cor.sep(t(newx), X2, theta[-(d + 1)], nu = 1.5), cor.sep(t(newx), X2, theta[-(d + 1)], nu = 1.5))) * # constant depends on kernel
        outer(w1.x2, w1.x2, FUN = Vectorize(zeta))
      
      EV <- tau2hat - tau2hat * sum(Ci * mat)
    } else {
      d <- ncol(fit1$X)
      newx <- matrix(newx, ncol = d)
      pred.fit <- pred.matGP(fit1, newx)
      x.mu <- drop(pred.fit$mu) # mean of f1(u)
      sig2 <- pred.fit$sig2
      
      ### calculate the closed form ###
      X2 <- matrix(fit2$X[, -(d + 1)], ncol = d)
      w1.x2 <- fit2$X[, d + 1]
      y2 <- fit2$y
      n <- length(y2)
      theta <- fit2$theta
      tau2hat <- fit2$tau2hat
      
      Ci <- fit2$Ki
      a <- crossprod(Ci, y2)
      
      # mean
      mua <- x.mu - sqrt(3) * sig2 / theta[d + 1]
      mub <- x.mu + sqrt(3) * sig2 / theta[d + 1]
      
      lambda11 <- c(1, mua)
      lambda12 <- c(0, 1)
      lambda21 <- c(1, -mub)
      
      e1 <- cbind(matrix(1 - sqrt(3) * w1.x2 / theta[d + 1]), sqrt(3) / theta[d + 1])
      e2 <- cbind(matrix(1 + sqrt(3) * w1.x2 / theta[d + 1]), sqrt(3) / theta[d + 1])
      
      predy <- drop(crossprod(t(cor.sep(t(newx), X2, theta[-(d + 1)], nu = 1.5)) * # common but depends on kernel
                                (exp((3 * sig2 + 2 * sqrt(3) * theta[d + 1] * (w1.x2 - x.mu)) / (2 * theta[d + 1]^2)) *
                                   (crossprod(t(e1), lambda11) * pnorm((mua - w1.x2) / sqrt(sig2)) +
                                      crossprod(t(e1), lambda12) * sqrt(sig2) / sqrt(2 * pi) * exp(-(w1.x2 - mua)^2 / (2 * sig2))) +
                                   exp((3 * sig2 - 2 * sqrt(3) * theta[d + 1] * (w1.x2 - x.mu)) / (2 * theta[d + 1]^2)) *
                                   (crossprod(t(e2), lambda21) * pnorm((-mub + w1.x2) / sqrt(sig2)) +
                                      crossprod(t(e2), lambda12) * sqrt(sig2) / sqrt(2 * pi) * exp(-(w1.x2 - mub)^2 / (2 * sig2)))), a))
      
      # var
      zeta <- function(x, y) {
        zetafun(w1 = x, w2 = y, m = x.mu, s = sig2, nu = 1.5, theta = theta[d + 1])
      }
      mat <- drop(crossprod(cor.sep(t(newx), X2, theta[-(d + 1)], nu = 1.5), cor.sep(t(newx), X2, theta[-(d + 1)], nu = 1.5))) * # constant depends on kernel
        outer(w1.x2, w1.x2, FUN = Vectorize(zeta))
      
      EV <- tau2hat - tau2hat * sum(Ci * mat)
    }
  } else if (kernel == "matern2.5") {
    if (constant) {
      d <- ncol(fit1$X)
      newx <- matrix(newx, ncol = d)
      pred.fit <- pred.matGP(fit1, newx)
      x.mu <- drop(pred.fit$mu) # mean of f1(u)
      sig2 <- pred.fit$sig2 #* 0
      
      ### calculate the closed form ###
      X2 <- matrix(fit2$X[, -(d + 1)], ncol = d)
      w1.x2 <- fit2$X[, d + 1]
      y2 <- fit2$y
      n <- length(y2)
      theta <- fit2$theta
      tau2hat <- fit2$tau2hat
      mu2 <- fit2$mu.hat
      
      Ci <- fit2$Ki
      a <- crossprod(Ci, y2 - mu2)
      
      # mean
      mua <- x.mu - sqrt(5) * sig2 / theta[d + 1]
      mub <- x.mu + sqrt(5) * sig2 / theta[d + 1]
      
      lambda11 <- c(1, mua, mua^2 + sig2)
      lambda12 <- cbind(0, 1, matrix(mua + w1.x2))
      lambda21 <- c(1, -mub, mub^2 + sig2)
      lambda22 <- cbind(0, 1, matrix(-mub - w1.x2))
      
      e1 <- cbind(
        matrix(1 - sqrt(5) * w1.x2 / theta[d + 1] + 5 * w1.x2^2 / (3 * theta[d + 1]^2)),
        matrix(sqrt(5) / theta[d + 1] - 10 * w1.x2 / (3 * theta[d + 1]^2)),
        5 / (3 * theta[d + 1]^2)
      )
      e2 <- cbind(
        matrix(1 + sqrt(5) * w1.x2 / theta[d + 1] + 5 * w1.x2^2 / (3 * theta[d + 1]^2)),
        matrix(sqrt(5) / theta[d + 1] + 10 * w1.x2 / (3 * theta[d + 1]^2)),
        5 / (3 * theta[d + 1]^2)
      )
      
      predy <- mu2 + drop(crossprod(t(cor.sep(t(newx), X2, theta[-(d + 1)], nu = 2.5)) *
                                      (exp((5 * sig2 + 2 * sqrt(5) * theta[d + 1] * (w1.x2 - x.mu)) / (2 * theta[d + 1]^2)) *
                                         (crossprod(t(e1), lambda11) * pnorm((mua - w1.x2) / sqrt(sig2)) +
                                            rowSums(e1 * lambda12) * sqrt(sig2) / sqrt(2 * pi) * exp(-(w1.x2 - mua)^2 / (2 * sig2))) +
                                         exp((5 * sig2 - 2 * sqrt(5) * theta[d + 1] * (w1.x2 - x.mu)) / (2 * theta[d + 1]^2)) *
                                         (crossprod(t(e2), lambda21) * pnorm((-mub + w1.x2) / sqrt(sig2)) +
                                            rowSums(e2 * lambda22) * sqrt(sig2) / sqrt(2 * pi) * exp(-(w1.x2 - mub)^2 / (2 * sig2)))), a))
      
      # var
      zeta <- function(x, y) {
        zetafun(w1 = x, w2 = y, m = x.mu, s = sig2, nu = 2.5, theta = theta[d + 1])
      }
      mat <- drop(crossprod(cor.sep(t(newx), X2, theta[-(d + 1)], nu = 2.5), cor.sep(t(newx), X2, theta[-(d + 1)], nu = 2.5))) * # constant depends on kernel
        outer(w1.x2, w1.x2, FUN = Vectorize(zeta))
      
      EV <- tau2hat - tau2hat * sum(Ci * mat)
    } else {
      d <- ncol(fit1$X)
      newx <- matrix(newx, ncol = d)
      pred.fit <- pred.matGP(fit1, newx)
      x.mu <- drop(pred.fit$mu) # mean of f1(u)
      sig2 <- pred.fit$sig2
      
      ### calculate the closed form ###
      X2 <- matrix(fit2$X[, -(d + 1)], ncol = d)
      w1.x2 <- fit2$X[, d + 1]
      y2 <- fit2$y
      n <- length(y2)
      theta <- fit2$theta
      tau2hat <- fit2$tau2hat
      
      Ci <- fit2$Ki
      a <- crossprod(Ci, y2)
      
      # mean
      mua <- x.mu - sqrt(5) * sig2 / theta[d + 1]
      mub <- x.mu + sqrt(5) * sig2 / theta[d + 1]
      
      lambda11 <- c(1, mua, mua^2 + sig2)
      lambda12 <- cbind(0, 1, matrix(mua + w1.x2))
      lambda21 <- c(1, -mub, mub^2 + sig2)
      lambda22 <- cbind(0, 1, matrix(-mub - w1.x2))
      
      e1 <- cbind(
        matrix(1 - sqrt(5) * w1.x2 / theta[d + 1] + 5 * w1.x2^2 / (3 * theta[d + 1]^2)),
        matrix(sqrt(5) / theta[d + 1] - 10 * w1.x2 / (3 * theta[d + 1]^2)),
        5 / (3 * theta[d + 1]^2)
      )
      e2 <- cbind(
        matrix(1 + sqrt(5) * w1.x2 / theta[d + 1] + 5 * w1.x2^2 / (3 * theta[d + 1]^2)),
        matrix(sqrt(5) / theta[d + 1] + 10 * w1.x2 / (3 * theta[d + 1]^2)),
        5 / (3 * theta[d + 1]^2)
      )
      
      predy <- drop(crossprod(t(cor.sep(t(newx), X2, theta[-(d + 1)], nu = 2.5)) *
                                (exp((5 * sig2 + 2 * sqrt(5) * theta[d + 1] * (w1.x2 - x.mu)) / (2 * theta[d + 1]^2)) *
                                   (crossprod(t(e1), lambda11) * pnorm((mua - w1.x2) / sqrt(sig2)) +
                                      rowSums(e1 * lambda12) * sqrt(sig2) / sqrt(2 * pi) * exp(-(w1.x2 - mua)^2 / (2 * sig2))) +
                                   exp((5 * sig2 - 2 * sqrt(5) * theta[d + 1] * (w1.x2 - x.mu)) / (2 * theta[d + 1]^2)) *
                                   (crossprod(t(e2), lambda21) * pnorm((-mub + w1.x2) / sqrt(sig2)) +
                                      rowSums(e2 * lambda22) * sqrt(sig2) / sqrt(2 * pi) * exp(-(w1.x2 - mub)^2 / (2 * sig2)))), a))
      
      # var
      zeta <- function(x, y) {
        zetafun(w1 = x, w2 = y, m = x.mu, s = sig2, nu = 2.5, theta = theta[d + 1])
      }
      mat <- drop(crossprod(cor.sep(t(newx), X2, theta[-(d + 1)], nu = 2.5), cor.sep(t(newx), X2, theta[-(d + 1)], nu = 2.5))) * # constant depends on kernel
        outer(w1.x2, w1.x2, FUN = Vectorize(zeta))
      
      EV <- tau2hat - tau2hat * sum(Ci * mat)
    }
  }
  return(-EV) # to maximize the V2.
}


#' object to optimize the point by ALD criterion updating V1 with three levels of fidelity
#'
#' @param Xcand candidate data point to be optimized.
#' @param fit an object of class RNAmf.
#' @param mc.sample a number of mc samples generated for this approach. Default is 100.
#' @param parallel logical indicating whether to run parallel or not. Default is FALSE.
#' @param ncore the number of core for parallel. Default is 1.
#' @return A negative V1 at Xcand.
#' @importFrom stats var
#' @noRd
#'

obj.ALD_V1_3level <- function(fit, Xcand, mc.sample, parallel = FALSE, ncore = 1) { # low
  
  kernel <- fit$kernel
  constant <- fit$constant
  fit1 <- fit$fits[[1]]
  fit2 <- fit$fits[[2]]
  fit3 <- fit$fits[[3]]
  
  d <- ncol(fit1$X)
  newx <- matrix(Xcand, nrow = 1)
  
  if (kernel == "sqex") {
    y1.sample <- rnorm(mc.sample, mean = pred.GP(fit1, newx)$mu, sd = sqrt(pred.GP(fit1, newx)$sig2))
  } else if (kernel == "matern1.5") {
    y1.sample <- rnorm(mc.sample, mean = pred.matGP(fit1, newx)$mu, sd = sqrt(pred.matGP(fit1, newx)$sig2))
  } else if (kernel == "matern2.5") {
    y1.sample <- rnorm(mc.sample, mean = pred.matGP(fit1, newx)$mu, sd = sqrt(pred.matGP(fit1, newx)$sig2))
  }
  
  ### V1 MC approximation ###
  if (parallel) {
    VEE.out <- foreach(i = 1:mc.sample, .combine = c) %dorng% {
      if (kernel == "sqex") {
        pred2 <- pred.GP(fit2, cbind(newx, y1.sample[i]))
      } else if (kernel == "matern1.5") {
        pred2 <- pred.matGP(fit2, cbind(newx, y1.sample[i]))
      } else if (kernel == "matern2.5") {
        pred2 <- pred.matGP(fit2, cbind(newx, y1.sample[i]))
      }
      x.mu <- pred2$mu
      sig2 <- pred2$sig2
      
      if (kernel == "sqex") {
        ### calculate the closed form ###
        X3 <- matrix(fit3$X[, -(d + 1)], ncol = d)
        w2.x3 <- fit3$X[, d + 1]
        y3 <- fit3$y
        n <- length(y3)
        theta <- fit3$theta
        tau2hat <- fit3$tau2hat
        Ci <- fit3$Ki
        
        if (constant) {
          mu3 <- fit3$mu.hat
        } else {
          mu3 <- 0
        }
        a <- crossprod(Ci, y3 - mu3)
        # mean
        VEE <- crossprod(t(exp(-distance(t(t(newx) / sqrt(theta[-(d + 1)])), t(t(X3) / sqrt(theta[-(d + 1)])))) *
                             1 / sqrt(1 + 2 * sig2 / theta[d + 1]) *
                             exp(-(drop(outer(x.mu, w2.x3, FUN = "-")))^2 / (theta[d + 1] + 2 * sig2))), a)
      } else if (kernel == "matern1.5") {
        ### calculate the closed form ###
        X3 <- matrix(fit3$X[, -(d + 1)], ncol = d)
        w2.x3 <- fit3$X[, d + 1]
        y3 <- fit3$y
        n <- length(y3)
        theta <- fit3$theta
        tau2hat <- fit3$tau2hat
        Ci <- fit3$Ki
        
        if (constant) {
          mu3 <- fit3$mu.hat
        } else {
          mu3 <- 0
        }
        a <- crossprod(Ci, y3 - mu3)
        # mean
        mua <- x.mu - sqrt(3) * sig2 / theta[d + 1]
        mub <- x.mu + sqrt(3) * sig2 / theta[d + 1]
        
        lambda11 <- c(1, mua)
        lambda12 <- c(0, 1)
        lambda21 <- c(1, -mub)
        
        e1 <- cbind(matrix(1 - sqrt(3) * w2.x3 / theta[d + 1]), sqrt(3) / theta[d + 1])
        e2 <- cbind(matrix(1 + sqrt(3) * w2.x3 / theta[d + 1]), sqrt(3) / theta[d + 1])
        
        VEE <- drop(crossprod(t(cor.sep(t(newx), X3, theta[-(d + 1)], nu = 1.5)) * # common but depends on kernel
                                (exp((3 * sig2 + 2 * sqrt(3) * theta[d + 1] * (w2.x3 - x.mu)) / (2 * theta[d + 1]^2)) *
                                   (crossprod(t(e1), lambda11) * pnorm((mua - w2.x3) / sqrt(sig2)) +
                                      crossprod(t(e1), lambda12) * sqrt(sig2) / sqrt(2 * pi) * exp(-(w2.x3 - mua)^2 / (2 * sig2))) +
                                   exp((3 * sig2 - 2 * sqrt(3) * theta[d + 1] * (w2.x3 - x.mu)) / (2 * theta[d + 1]^2)) *
                                   (crossprod(t(e2), lambda21) * pnorm((-mub + w2.x3) / sqrt(sig2)) +
                                      crossprod(t(e2), lambda12) * sqrt(sig2) / sqrt(2 * pi) * exp(-(w2.x3 - mub)^2 / (2 * sig2)))), a))
      } else if (kernel == "matern2.5") {
        ### calculate the closed form ###
        X3 <- matrix(fit3$X[, -(d + 1)], ncol = d)
        w2.x3 <- fit3$X[, d + 1]
        y3 <- fit3$y
        n <- length(y3)
        theta <- fit3$theta
        tau2hat <- fit3$tau2hat
        Ci <- fit3$Ki
        
        if (constant) {
          mu3 <- fit3$mu.hat
        } else {
          mu3 <- 0
        }
        a <- crossprod(Ci, y3 - mu3)
        # mean
        mua <- x.mu - sqrt(5) * sig2 / theta[d + 1]
        mub <- x.mu + sqrt(5) * sig2 / theta[d + 1]
        
        lambda11 <- c(1, mua, mua^2 + sig2)
        lambda12 <- cbind(0, 1, matrix(mua + w2.x3))
        lambda21 <- c(1, -mub, mub^2 + sig2)
        lambda22 <- cbind(0, 1, matrix(-mub - w2.x3))
        
        e1 <- cbind(
          matrix(1 - sqrt(5) * w2.x3 / theta[d + 1] + 5 * w2.x3^2 / (3 * theta[d + 1]^2)),
          matrix(sqrt(5) / theta[d + 1] - 10 * w2.x3 / (3 * theta[d + 1]^2)),
          5 / (3 * theta[d + 1]^2)
        )
        e2 <- cbind(
          matrix(1 + sqrt(5) * w2.x3 / theta[d + 1] + 5 * w2.x3^2 / (3 * theta[d + 1]^2)),
          matrix(sqrt(5) / theta[d + 1] + 10 * w2.x3 / (3 * theta[d + 1]^2)),
          5 / (3 * theta[d + 1]^2)
        )
        
        VEE <- drop(crossprod(t(cor.sep(t(newx), X3, theta[-(d + 1)], nu = 2.5)) *
                                (exp((5 * sig2 + 2 * sqrt(5) * theta[d + 1] * (w2.x3 - x.mu)) / (2 * theta[d + 1]^2)) *
                                   (crossprod(t(e1), lambda11) * pnorm((mua - w2.x3) / sqrt(sig2)) +
                                      rowSums(e1 * lambda12) * sqrt(sig2) / sqrt(2 * pi) * exp(-(w2.x3 - mua)^2 / (2 * sig2))) +
                                   exp((5 * sig2 - 2 * sqrt(5) * theta[d + 1] * (w2.x3 - x.mu)) / (2 * theta[d + 1]^2)) *
                                   (crossprod(t(e2), lambda21) * pnorm((-mub + w2.x3) / sqrt(sig2)) +
                                      rowSums(e2 * lambda22) * sqrt(sig2) / sqrt(2 * pi) * exp(-(w2.x3 - mub)^2 / (2 * sig2)))), a))
      }
      return(VEE) # to maximize the V1.
    }
    attr(VEE.out, "rng") <- NULL
    attr(VEE.out, "doRNG_version") <- NULL
  } else {
    VEE.out <- rep(0, mc.sample)
    for (i in 1:mc.sample) {
      if (kernel == "sqex") {
        pred2 <- pred.GP(fit2, cbind(newx, y1.sample[i]))
      } else if (kernel == "matern1.5") {
        pred2 <- pred.matGP(fit2, cbind(newx, y1.sample[i]))
      } else if (kernel == "matern2.5") {
        pred2 <- pred.matGP(fit2, cbind(newx, y1.sample[i]))
      }
      x.mu <- pred2$mu
      sig2 <- pred2$sig2
      
      if (kernel == "sqex") {
        ### calculate the closed form ###
        X3 <- matrix(fit3$X[, -(d + 1)], ncol = d)
        w2.x3 <- fit3$X[, d + 1]
        y3 <- fit3$y
        n <- length(y3)
        theta <- fit3$theta
        tau2hat <- fit3$tau2hat
        Ci <- fit3$Ki
        
        if (constant) {
          mu3 <- fit3$mu.hat
        } else {
          mu3 <- 0
        }
        a <- crossprod(Ci, y3 - mu3)
        # mean
        VEE <- crossprod(t(exp(-distance(t(t(newx) / sqrt(theta[-(d + 1)])), t(t(X3) / sqrt(theta[-(d + 1)])))) *
                             1 / sqrt(1 + 2 * sig2 / theta[d + 1]) *
                             exp(-(drop(outer(x.mu, w2.x3, FUN = "-")))^2 / (theta[d + 1] + 2 * sig2))), a)
      } else if (kernel == "matern1.5") {
        ### calculate the closed form ###
        X3 <- matrix(fit3$X[, -(d + 1)], ncol = d)
        w2.x3 <- fit3$X[, d + 1]
        y3 <- fit3$y
        n <- length(y3)
        theta <- fit3$theta
        tau2hat <- fit3$tau2hat
        Ci <- fit3$Ki
        
        if (constant) {
          mu3 <- fit3$mu.hat
        } else {
          mu3 <- 0
        }
        a <- crossprod(Ci, y3 - mu3)
        # mean
        mua <- x.mu - sqrt(3) * sig2 / theta[d + 1]
        mub <- x.mu + sqrt(3) * sig2 / theta[d + 1]
        
        lambda11 <- c(1, mua)
        lambda12 <- c(0, 1)
        lambda21 <- c(1, -mub)
        
        e1 <- cbind(matrix(1 - sqrt(3) * w2.x3 / theta[d + 1]), sqrt(3) / theta[d + 1])
        e2 <- cbind(matrix(1 + sqrt(3) * w2.x3 / theta[d + 1]), sqrt(3) / theta[d + 1])
        
        VEE <- drop(crossprod(t(cor.sep(t(newx), X3, theta[-(d + 1)], nu = 1.5)) * # common but depends on kernel
                                (exp((3 * sig2 + 2 * sqrt(3) * theta[d + 1] * (w2.x3 - x.mu)) / (2 * theta[d + 1]^2)) *
                                   (crossprod(t(e1), lambda11) * pnorm((mua - w2.x3) / sqrt(sig2)) +
                                      crossprod(t(e1), lambda12) * sqrt(sig2) / sqrt(2 * pi) * exp(-(w2.x3 - mua)^2 / (2 * sig2))) +
                                   exp((3 * sig2 - 2 * sqrt(3) * theta[d + 1] * (w2.x3 - x.mu)) / (2 * theta[d + 1]^2)) *
                                   (crossprod(t(e2), lambda21) * pnorm((-mub + w2.x3) / sqrt(sig2)) +
                                      crossprod(t(e2), lambda12) * sqrt(sig2) / sqrt(2 * pi) * exp(-(w2.x3 - mub)^2 / (2 * sig2)))), a))
      } else if (kernel == "matern2.5") {
        ### calculate the closed form ###
        X3 <- matrix(fit3$X[, -(d + 1)], ncol = d)
        w2.x3 <- fit3$X[, d + 1]
        y3 <- fit3$y
        n <- length(y3)
        theta <- fit3$theta
        tau2hat <- fit3$tau2hat
        Ci <- fit3$Ki
        
        if (constant) {
          mu3 <- fit3$mu.hat
        } else {
          mu3 <- 0
        }
        a <- crossprod(Ci, y3 - mu3)
        # mean
        mua <- x.mu - sqrt(5) * sig2 / theta[d + 1]
        mub <- x.mu + sqrt(5) * sig2 / theta[d + 1]
        
        lambda11 <- c(1, mua, mua^2 + sig2)
        lambda12 <- cbind(0, 1, matrix(mua + w2.x3))
        lambda21 <- c(1, -mub, mub^2 + sig2)
        lambda22 <- cbind(0, 1, matrix(-mub - w2.x3))
        
        e1 <- cbind(
          matrix(1 - sqrt(5) * w2.x3 / theta[d + 1] + 5 * w2.x3^2 / (3 * theta[d + 1]^2)),
          matrix(sqrt(5) / theta[d + 1] - 10 * w2.x3 / (3 * theta[d + 1]^2)),
          5 / (3 * theta[d + 1]^2)
        )
        e2 <- cbind(
          matrix(1 + sqrt(5) * w2.x3 / theta[d + 1] + 5 * w2.x3^2 / (3 * theta[d + 1]^2)),
          matrix(sqrt(5) / theta[d + 1] + 10 * w2.x3 / (3 * theta[d + 1]^2)),
          5 / (3 * theta[d + 1]^2)
        )
        
        VEE <- drop(crossprod(t(cor.sep(t(newx), X3, theta[-(d + 1)], nu = 2.5)) *
                                (exp((5 * sig2 + 2 * sqrt(5) * theta[d + 1] * (w2.x3 - x.mu)) / (2 * theta[d + 1]^2)) *
                                   (crossprod(t(e1), lambda11) * pnorm((mua - w2.x3) / sqrt(sig2)) +
                                      rowSums(e1 * lambda12) * sqrt(sig2) / sqrt(2 * pi) * exp(-(w2.x3 - mua)^2 / (2 * sig2))) +
                                   exp((5 * sig2 - 2 * sqrt(5) * theta[d + 1] * (w2.x3 - x.mu)) / (2 * theta[d + 1]^2)) *
                                   (crossprod(t(e2), lambda21) * pnorm((-mub + w2.x3) / sqrt(sig2)) +
                                      rowSums(e2 * lambda22) * sqrt(sig2) / sqrt(2 * pi) * exp(-(w2.x3 - mub)^2 / (2 * sig2)))), a))
      }
      VEE.out[i] <- VEE # to maximize the V1.
    }
  }
  return(-var(VEE.out)) # to maximize the V1.
}


#' object to optimize the point by ALD criterion updating V2 with three levels of fidelity
#'
#' @param Xcand candidate data point to be optimized.
#' @param fit an object of class RNAmf.
#' @param mc.sample a number of mc samples generated for this approach. Default is 100.
#' @param parallel logical indicating whether to run parallel or not. Default is FALSE.
#' @param ncore the number of core for parallel. Default is 1.
#' @return A negative V2 at Xcand.
#' @noRd
#'

obj.ALD_V2_3level <- function(fit, Xcand, mc.sample, parallel = FALSE, ncore = 1) { # med
  
  V <- predict(fit, Xcand)$sig2[[3]]
  EVE <- V +
    obj.ALD_V1_3level(fit, Xcand, mc.sample, parallel = FALSE, ncore = 1) +# -V1
    obj.ALD_V3_3level(fit, Xcand) # -V3
  
  return(-EVE) # to maximize the V2.
}


#' object to optimize the point by ALD criterion updating V3 with three levels of fidelity
#'
#' @param Xcand candidate data point to be optimized.
#' @param fit an object of class RNAmf.
#' @return A negative V2 at Xcand.
#' @noRd
#'

obj.ALD_V3_3level <- function(fit, Xcand) { # high
  newx <- matrix(Xcand, nrow = 1)
  
  kernel <- fit$kernel
  constant <- fit$constant
  fit1 <- fit$fits[[1]]
  fit2 <- fit$fits[[2]]
  fit3 <- fit$fits[[3]]
  
  if (kernel == "sqex") {
    pred.RNAmf_two_level <- predict(fit, newx)
    x.mu <- pred.RNAmf_two_level$mu[[2]]
    sig2 <- pred.RNAmf_two_level$sig2[[2]]
    
    d <- ncol(fit1$X)
    newx <- matrix(newx, ncol = d)
    
    ### calculate the closed form ###
    X3 <- matrix(fit3$X[, -(d + 1)], ncol = d)
    w2.x3 <- fit3$X[, d + 1]
    y3 <- fit3$y
    n <- length(y3)
    theta <- fit3$theta
    tau2hat <- fit3$tau2hat
    Ci <- fit3$Ki
    
    if (constant) {
      mu3 <- fit3$mu.hat
    } else {
      mu3 <- 0
    }
    a <- crossprod(Ci, y3 - mu3)
    
    # mean
    predy <- mu3 + crossprod(t(exp(-distance(t(t(newx) / sqrt(theta[-(d + 1)])), t(t(X3) / sqrt(theta[-(d + 1)])))) *
                                 1 / sqrt(1 + 2 * sig2 / theta[d + 1]) *
                                 exp(-(drop(outer(x.mu, w2.x3, FUN = "-")))^2 / (theta[d + 1] + 2 * sig2))), a)
    # var
    mat <- drop(crossprod(exp(-distance(t(t(newx) / sqrt(theta[-(d + 1)])), t(t(X3) / sqrt(theta[-(d + 1)])))),
                          exp(-distance(t(t(newx) / sqrt(theta[-(d + 1)])), t(t(X3) / sqrt(theta[-(d + 1)])))))) * # common components
      1 / sqrt(1 + 4 * sig2 / theta[d + 1]) *
      exp(-(outer(w2.x3, w2.x3, FUN = "+") / 2 - matrix(x.mu, n, n))^2 / (theta[d + 1] / 2 + 2 * sig2)) *
      exp(-(outer(w2.x3, w2.x3, FUN = "-"))^2 / (2 * theta[d + 1]))
    
    EEV <- tau2hat - tau2hat * sum(Ci * mat)
  } else if (kernel == "matern1.5") {
    d <- ncol(fit1$X)
    newx <- matrix(newx, ncol = d)
    pred.RNAmf_two_level <- predict(fit, newx)
    x.mu <- pred.RNAmf_two_level$mu[[2]]
    sig2 <- pred.RNAmf_two_level$sig2[[2]]
    
    ### calculate the closed form ###
    X3 <- matrix(fit3$X[, -(d + 1)], ncol = d)
    w2.x3 <- fit3$X[, d + 1]
    y3 <- fit3$y
    n <- length(y3)
    theta <- fit3$theta
    tau2hat <- fit3$tau2hat
    Ci <- fit3$Ki
    
    if (constant) {
      mu3 <- fit3$mu.hat
    } else {
      mu3 <- 0
    }
    a <- crossprod(Ci, y3 - mu3)
    
    # mean
    mua <- x.mu - sqrt(3) * sig2 / theta[d + 1]
    mub <- x.mu + sqrt(3) * sig2 / theta[d + 1]
    
    lambda11 <- c(1, mua)
    lambda12 <- c(0, 1)
    lambda21 <- c(1, -mub)
    
    e1 <- cbind(matrix(1 - sqrt(3) * w2.x3 / theta[d + 1]), sqrt(3) / theta[d + 1])
    e2 <- cbind(matrix(1 + sqrt(3) * w2.x3 / theta[d + 1]), sqrt(3) / theta[d + 1])
    
    predy <- mu3 + drop(crossprod(t(cor.sep(t(newx), X3, theta[-(d + 1)], nu = 1.5)) * # common but depends on kernel
                                    (exp((3 * sig2 + 2 * sqrt(3) * theta[d + 1] * (w2.x3 - x.mu)) / (2 * theta[d + 1]^2)) *
                                       (crossprod(t(e1), lambda11) * pnorm((mua - w2.x3) / sqrt(sig2)) +
                                          crossprod(t(e1), lambda12) * sqrt(sig2) / sqrt(2 * pi) * exp(-(w2.x3 - mua)^2 / (2 * sig2))) +
                                       exp((3 * sig2 - 2 * sqrt(3) * theta[d + 1] * (w2.x3 - x.mu)) / (2 * theta[d + 1]^2)) *
                                       (crossprod(t(e2), lambda21) * pnorm((-mub + w2.x3) / sqrt(sig2)) +
                                          crossprod(t(e2), lambda12) * sqrt(sig2) / sqrt(2 * pi) * exp(-(w2.x3 - mub)^2 / (2 * sig2)))), a))
    # var
    zeta <- function(x, y) {
      zetafun(w1 = x, w2 = y, m = x.mu, s = sig2, nu = 1.5, theta = theta[d + 1])
    }
    mat <- drop(crossprod(cor.sep(t(newx), X3, theta[-(d + 1)], nu = 1.5), cor.sep(t(newx), X3, theta[-(d + 1)], nu = 1.5)))* # constant depends on kernel
      outer(w2.x3, w2.x3, FUN = Vectorize(zeta))
    
    EEV <- tau2hat - tau2hat * sum(Ci * mat)
  } else if (kernel == "matern2.5") {
    d <- ncol(fit1$X)
    newx <- matrix(newx, ncol = d)
    pred.RNAmf_two_level <- predict(fit, newx)
    x.mu <- pred.RNAmf_two_level$mu[[2]]
    sig2 <- pred.RNAmf_two_level$sig2[[2]]
    
    ### calculate the closed form ###
    X3 <- matrix(fit3$X[, -(d + 1)], ncol = d)
    w2.x3 <- fit3$X[, d + 1]
    y3 <- fit3$y
    n <- length(y3)
    theta <- fit3$theta
    tau2hat <- fit3$tau2hat
    Ci <- fit3$Ki
    
    if (constant) {
      mu3 <- fit3$mu.hat
    } else {
      mu3 <- 0
    }
    a <- crossprod(Ci, y3 - mu3)
    
    # mean
    mua <- x.mu - sqrt(5) * sig2 / theta[d + 1]
    mub <- x.mu + sqrt(5) * sig2 / theta[d + 1]
    
    lambda11 <- c(1, mua, mua^2 + sig2)
    lambda12 <- cbind(0, 1, matrix(mua + w2.x3))
    lambda21 <- c(1, -mub, mub^2 + sig2)
    lambda22 <- cbind(0, 1, matrix(-mub - w2.x3))
    
    e1 <- cbind(
      matrix(1 - sqrt(5) * w2.x3 / theta[d + 1] + 5 * w2.x3^2 / (3 * theta[d + 1]^2)),
      matrix(sqrt(5) / theta[d + 1] - 10 * w2.x3 / (3 * theta[d + 1]^2)),
      5 / (3 * theta[d + 1]^2)
    )
    e2 <- cbind(
      matrix(1 + sqrt(5) * w2.x3 / theta[d + 1] + 5 * w2.x3^2 / (3 * theta[d + 1]^2)),
      matrix(sqrt(5) / theta[d + 1] + 10 * w2.x3 / (3 * theta[d + 1]^2)),
      5 / (3 * theta[d + 1]^2)
    )
    
    predy <- mu3 + drop(crossprod(t(cor.sep(t(newx), X3, theta[-(d + 1)], nu = 2.5)) *
                                    (exp((5 * sig2 + 2 * sqrt(5) * theta[d + 1] * (w2.x3 - x.mu)) / (2 * theta[d + 1]^2)) *
                                       (crossprod(t(e1), lambda11) * pnorm((mua - w2.x3) / sqrt(sig2)) +
                                          rowSums(e1 * lambda12) * sqrt(sig2) / sqrt(2 * pi) * exp(-(w2.x3 - mua)^2 / (2 * sig2))) +
                                       exp((5 * sig2 - 2 * sqrt(5) * theta[d + 1] * (w2.x3 - x.mu)) / (2 * theta[d + 1]^2)) *
                                       (crossprod(t(e2), lambda21) * pnorm((-mub + w2.x3) / sqrt(sig2)) +
                                          rowSums(e2 * lambda22) * sqrt(sig2) / sqrt(2 * pi) * exp(-(w2.x3 - mub)^2 / (2 * sig2)))), a))
    
    # var
    zeta <- function(x, y) {
      zetafun(w1 = x, w2 = y, m = x.mu, s = sig2, nu = 2.5, theta = theta[d + 1])
    }
    mat <- drop(crossprod(cor.sep(t(newx), X3, theta[-(d + 1)], nu = 2.5), cor.sep(t(newx), X3, theta[-(d + 1)], nu = 2.5))) * # constant depends on kernel
      outer(w2.x3, w2.x3, FUN = Vectorize(zeta))
    
    EEV <- tau2hat - tau2hat * sum(Ci * mat)
  }
  return(-EEV) # to maximize the V3.
}


#' object to optimize the next point by ALC criterion updating at level 1 with two levels of fidelity
#'
#' @param Xcand candidate data point to be optimized.
#' @param Xref vector or matrix of reference data.
#' @param fit an object of class RNAmf.
#' @param mc.sample a number of mc samples generated for this approach. Default is 100.
#' @param parallel logical indicating whether to run parallel or not. Default is FALSE.
#' @param ncore the number of core for parallel. Default is 1.
#' @return A mean of the deduced variance at Xref.
#' @importFrom plgp covar.sep
#' @importFrom stats rnorm
#' @noRd
#'

obj.ALC_level_1 <- function(Xcand, Xref, fit, MC = FALSE, mc.sample, parallel = FALSE, ncore = 1) {
  fit1 <- f1 <- fit$fits[[1]]
  fit2 <- f2 <- fit$fits[[2]]
  constant <- fit$constant
  kernel <- fit$kernel
  g <- fit1$g
  
  Xcand <- matrix(Xcand, nrow = 1)
  if(MC){
    if (kernel == "sqex") {
      y1.sample <- rnorm(mc.sample, mean = pred.GP(f1, Xcand)$mu, sd = sqrt(pred.GP(f1, Xcand)$sig2))
    } else if (kernel == "matern1.5") {
      y1.sample <- rnorm(mc.sample, mean = pred.matGP(f1, Xcand)$mu, sd = sqrt(pred.matGP(f1, Xcand)$sig2))
    } else if (kernel == "matern2.5") {
      y1.sample <- rnorm(mc.sample, mean = pred.matGP(f1, Xcand)$mu, sd = sqrt(pred.matGP(f1, Xcand)$sig2))
    }
  }else{
    if (kernel == "sqex") {
      y1.sample <- pred.GP(f1, Xcand)$mu
    } else if (kernel == "matern1.5") {
      y1.sample <- pred.matGP(f1, Xcand)$mu
    } else if (kernel == "matern2.5") {
      y1.sample <- pred.matGP(f1, Xcand)$mu
    }
  }
  Xcand1 <- Xcand
  
  ### Choose level 1 ###
  ### update Ki1
  if (kernel == "sqex") {
    cov.newx1 <- covar.sep(X1 = Xcand1, d = f1$theta, g = g)
    cov.Xnewx1 <- covar.sep(X1 = f1$X, X2 = Xcand1, d = f1$theta, g = 0)
  } else if (kernel == "matern1.5") {
    cov.newx1 <- cor.sep(X = Xcand1, theta = f1$theta, nu = 1.5)
    cov.Xnewx1 <- cor.sep(X = f1$X, x = Xcand1, theta = f1$theta, nu = 1.5)
  } else if (kernel == "matern2.5") {
    cov.newx1 <- cor.sep(X = Xcand1, theta = f1$theta, nu = 2.5)
    cov.Xnewx1 <- cor.sep(X = f1$X, x = Xcand1, theta = f1$theta, nu = 2.5)
  }
  v.next1 <- drop(cov.newx1 - crossprod(t(crossprod(cov.Xnewx1, f1$Ki)), cov.Xnewx1))
  g.next1 <- -1 / drop(v.next1) * crossprod(f1$Ki, cov.Xnewx1)
  
  
  fit1$Ki <- rbind(
    cbind(f1$Ki + crossprod(t(g.next1), t(g.next1) * v.next1), g.next1),
    cbind(t(g.next1), 1 / drop(v.next1))
  )
  
  fit1$X <- rbind(f1$X, Xcand1)
  
  if(MC){
    ALC.out <- rep(0, mc.sample)
    if (parallel) {
      ALC.out <- foreach(i = 1:mc.sample, .combine = c) %dorng% {
        fit.tmp <- fit
        fit1$y <- c(f1$y, y1.sample[i])
        fit1$tau2hat <- drop(crossprod(t(crossprod(fit1$y - fit1$mu.hat, fit1$Ki)), (fit1$y - fit1$mu.hat)) / length(fit1$y))
        
        fit.tmp$fits[[1]] <- fit1
        
        return(mean(predict(fit.tmp, Xref)$sig2[[2]])) # to minimize the deduced variance. To maximize, -mean
      }
      attr(ALC.out, "rng") <- NULL
      attr(ALC.out, "doRNG_version") <- NULL
    } else {
      for (i in 1:mc.sample) {
        fit.tmp <- fit
        fit1$y <- c(f1$y, y1.sample[i])
        fit1$tau2hat <- drop(crossprod(t(crossprod(fit1$y - fit1$mu.hat, fit1$Ki)), (fit1$y - fit1$mu.hat)) / length(fit1$y))
        
        fit.tmp$fits[[1]] <- fit1
        
        ALC.out[i] <- mean(predict(fit.tmp, Xref)$sig2[[2]]) # to minimize the deduced variance. To maximize, -mean
      }
    }
  }else{
    fit.tmp <- fit
    fit1$y <- c(f1$y, y1.sample)
    fit1$tau2hat <- drop(crossprod(t(crossprod(fit1$y - fit1$mu.hat, fit1$Ki)), (fit1$y - fit1$mu.hat)) / length(fit1$y))
    
    fit.tmp$fits[[1]] <- fit1
    
    ALC.out <- mean(predict(fit.tmp, Xref)$sig2[[2]])
  }
  return(mean(ALC.out))
}


#' object to optimize the next point by ALC criterion updating at level 2 with two levels of fidelity
#'
#' @param Xcand candidate data point to be optimized.
#' @param Xref vector or matrix of reference data.
#' @param fit an object of class RNAmf.
#' @param mc.sample a number of mc samples generated for this approach. Default is 100.
#' @param parallel logical indicating whether to run parallel or not. Default is FALSE.
#' @param ncore the number of core for parallel. Default is 1.
#' @return A mean of the deduced variance at Xref.
#' @importFrom plgp covar.sep
#' @importFrom stats rnorm
#' @noRd
#'

obj.ALC_level_2 <- function(Xcand, Xref, fit, MC = FALSE, mc.sample, parallel = FALSE, ncore = 1) {
  fit1 <- f1 <- fit$fits[[1]]
  fit2 <- f2 <- fit$fits[[2]]
  constant <- fit$constant
  kernel <- fit$kernel
  g <- fit1$g
  
  Xcand <- matrix(Xcand, nrow = 1)
  if(MC){
    if (kernel == "sqex") {
      y1.sample <- rnorm(mc.sample, mean = pred.GP(f1, Xcand)$mu, sd = sqrt(pred.GP(f1, Xcand)$sig2))
    } else if (kernel == "matern1.5") {
      y1.sample <- rnorm(mc.sample, mean = pred.matGP(f1, Xcand)$mu, sd = sqrt(pred.matGP(f1, Xcand)$sig2))
    } else if (kernel == "matern2.5") {
      y1.sample <- rnorm(mc.sample, mean = pred.matGP(f1, Xcand)$mu, sd = sqrt(pred.matGP(f1, Xcand)$sig2))
    }
  }else{
    if (kernel == "sqex") {
      y1.sample <- pred.GP(f1, Xcand)$mu
    } else if (kernel == "matern1.5") {
      y1.sample <- pred.matGP(f1, Xcand)$mu
    } else if (kernel == "matern2.5") {
      y1.sample <- pred.matGP(f1, Xcand)$mu
    }
  }
  Xcand1 <- Xcand
  
  ### Choose level 1 ###
  ### update Ki1
  if (kernel == "sqex") {
    cov.newx1 <- covar.sep(X1 = Xcand1, d = f1$theta, g = g)
    cov.Xnewx1 <- covar.sep(X1 = f1$X, X2 = Xcand1, d = f1$theta, g = 0)
  } else if (kernel == "matern1.5") {
    cov.newx1 <- cor.sep(X = Xcand1, theta = f1$theta, nu = 1.5)
    cov.Xnewx1 <- cor.sep(X = f1$X, x = Xcand1, theta = f1$theta, nu = 1.5)
  } else if (kernel == "matern2.5") {
    cov.newx1 <- cor.sep(X = Xcand1, theta = f1$theta, nu = 2.5)
    cov.Xnewx1 <- cor.sep(X = f1$X, x = Xcand1, theta = f1$theta, nu = 2.5)
  }
  v.next1 <- drop(cov.newx1 - crossprod(t(crossprod(cov.Xnewx1, f1$Ki)), cov.Xnewx1))
  g.next1 <- -1 / drop(v.next1) * crossprod(f1$Ki, cov.Xnewx1)
  
  
  fit1$Ki <- rbind(
    cbind(f1$Ki + crossprod(t(g.next1), t(g.next1) * v.next1), g.next1),
    cbind(t(g.next1), 1 / drop(v.next1))
  )
  
  fit1$X <- rbind(f1$X, Xcand1)
  
  if(MC){
    if (parallel) {
      ALC.out <- foreach(i = 1:mc.sample, .combine = c) %dorng% {
        fit.tmp <- fit
        fit1$y <- c(f1$y, y1.sample[i])
        fit1$tau2hat <- drop(crossprod(t(crossprod(fit1$y - fit1$mu.hat, fit1$Ki)), (fit1$y - fit1$mu.hat)) / length(fit1$y))
        
        fit.tmp$fits[[1]] <- fit1
        
        ### Choose level 2 ###
        if (kernel == "sqex") {
          pred2 <- pred.GP(fit2, cbind(Xcand, y1.sample[i]))
          y2.sample <- rnorm(1, pred2$mu, sqrt(pred2$sig2))
        } else if (kernel == "matern1.5") {
          pred2 <- pred.matGP(fit2, cbind(Xcand, y1.sample[i]))
          y2.sample <- rnorm(1, pred2$mu, sqrt(pred2$sig2))
        } else if (kernel == "matern2.5") {
          pred2 <- pred.matGP(fit2, cbind(Xcand, y1.sample[i]))
          y2.sample <- rnorm(1, pred2$mu, sqrt(pred2$sig2))
        }
        
        ### update Ki2
        newx2 <- cbind(Xcand, y1.sample[i])
        
        if (kernel == "sqex") {
          cov.newx2 <- covar.sep(X1 = newx2, d = f2$theta, g = g)
          cov.Xnewx2 <- covar.sep(X1 = f2$X, X2 = newx2, d = f2$theta, g = 0)
        } else if (kernel == "matern1.5") {
          cov.newx2 <- cor.sep(X = newx2, theta = f2$theta, nu = 1.5)
          cov.Xnewx2 <- cor.sep(X = f2$X, x = newx2, theta = f2$theta, nu = 1.5)
        } else if (kernel == "matern2.5") {
          cov.newx2 <- cor.sep(X = newx2, theta = f2$theta, nu = 2.5)
          cov.Xnewx2 <- cor.sep(X = f2$X, x = newx2, theta = f2$theta, nu = 2.5)
        }
        v.next2 <- drop(cov.newx2 - crossprod(t(crossprod(cov.Xnewx2, f2$Ki)), cov.Xnewx2))
        g.next2 <- -1 / drop(v.next2) * crossprod(f2$Ki, cov.Xnewx2)
        
        fit2$Ki <- rbind(
          cbind(f2$Ki + crossprod(t(g.next2), t(g.next2) * v.next2), g.next2),
          cbind(t(g.next2), 1 / drop(v.next2))
        )
        
        fit2$X <- rbind(f2$X, newx2)
        fit2$y <- c(f2$y, y2.sample)
        fit2$tau2hat <- drop(crossprod(t(crossprod(fit2$y - fit2$mu.hat, fit2$Ki)), (fit2$y - fit2$mu.hat)) / length(fit2$y))
        
        fit.tmp$fits[[2]] <- fit2
        
        return(mean(predict(fit.tmp, Xref)$sig2[[2]])) # to minimize the deduced variance. To maximize, -mean
      }
      attr(ALC.out, "rng") <- NULL
      attr(ALC.out, "doRNG_version") <- NULL
    } else {
      ALC.out <- rep(0, mc.sample)
      for (i in 1:mc.sample) {
        fit.tmp <- fit
        fit1$y <- c(f1$y, y1.sample[i])
        fit1$tau2hat <- drop(crossprod(t(crossprod(fit1$y - fit1$mu.hat, fit1$Ki)), (fit1$y - fit1$mu.hat)) / length(fit1$y))
        
        fit.tmp$fits[[1]] <- fit1
        
        ### Choose level 2 ###
        if (kernel == "sqex") {
          pred2 <- pred.GP(fit2, cbind(Xcand, y1.sample[i]))
          y2.sample <- rnorm(1, pred2$mu, sqrt(pred2$sig2))
        } else if (kernel == "matern1.5") {
          pred2 <- pred.matGP(fit2, cbind(Xcand, y1.sample[i]))
          y2.sample <- rnorm(1, pred2$mu, sqrt(pred2$sig2))
        } else if (kernel == "matern2.5") {
          pred2 <- pred.matGP(fit2, cbind(Xcand, y1.sample[i]))
          y2.sample <- rnorm(1, pred2$mu, sqrt(pred2$sig2))
        }
        
        ### update Ki2
        newx2 <- cbind(Xcand, y1.sample[i])
        
        if (kernel == "sqex") {
          cov.newx2 <- covar.sep(X1 = newx2, d = f2$theta, g = g)
          cov.Xnewx2 <- covar.sep(X1 = f2$X, X2 = newx2, d = f2$theta, g = 0)
        } else if (kernel == "matern1.5") {
          cov.newx2 <- cor.sep(X = newx2, theta = f2$theta, nu = 1.5)
          cov.Xnewx2 <- cor.sep(X = f2$X, x = newx2, theta = f2$theta, nu = 1.5)
        } else if (kernel == "matern2.5") {
          cov.newx2 <- cor.sep(X = newx2, theta = f2$theta, nu = 2.5)
          cov.Xnewx2 <- cor.sep(X = f2$X, x = newx2, theta = f2$theta, nu = 2.5)
        }
        v.next2 <- drop(cov.newx2 - crossprod(t(crossprod(cov.Xnewx2, f2$Ki)), cov.Xnewx2))
        g.next2 <- -1 / drop(v.next2) * crossprod(f2$Ki, cov.Xnewx2)
        
        fit2$Ki <- rbind(
          cbind(f2$Ki + crossprod(t(g.next2), t(g.next2) * v.next2), g.next2),
          cbind(t(g.next2), 1 / drop(v.next2))
        )
        
        fit2$X <- rbind(f2$X, newx2)
        fit2$y <- c(f2$y, y2.sample)
        fit2$tau2hat <- drop(crossprod(t(crossprod(fit2$y - fit2$mu.hat, fit2$Ki)), (fit2$y - fit2$mu.hat)) / length(fit2$y))
        
        fit.tmp$fits[[2]] <- fit2
        
        ALC.out[i] <- mean(predict(fit.tmp, Xref)$sig2[[2]]) # to minimize the deduced variance. To maximize, -mean
      }
    }
  }else{
    fit.tmp <- fit
    fit1$y <- c(f1$y, y1.sample)
    fit1$tau2hat <- drop(crossprod(t(crossprod(fit1$y - fit1$mu.hat, fit1$Ki)), (fit1$y - fit1$mu.hat)) / length(fit1$y))
    
    fit.tmp$fits[[1]] <- fit1
    
    ### Choose level 2 ###
    if (kernel == "sqex") {
      pred2 <- pred.GP(fit2, cbind(Xcand, y1.sample))
    } else if (kernel == "matern1.5") {
      pred2 <- pred.matGP(fit2, cbind(Xcand, y1.sample))
    } else if (kernel == "matern2.5") {
      pred2 <- pred.matGP(fit2, cbind(Xcand, y1.sample))
    }
    y2.sample <- pred2$mu
    
    ### update Ki2
    newx2 <- cbind(Xcand, y1.sample)
    
    if (kernel == "sqex") {
      cov.newx2 <- covar.sep(X1 = newx2, d = f2$theta, g = g)
      cov.Xnewx2 <- covar.sep(X1 = f2$X, X2 = newx2, d = f2$theta, g = 0)
    } else if (kernel == "matern1.5") {
      cov.newx2 <- cor.sep(X = newx2, theta = f2$theta, nu = 1.5)
      cov.Xnewx2 <- cor.sep(X = f2$X, x = newx2, theta = f2$theta, nu = 1.5)
    } else if (kernel == "matern2.5") {
      cov.newx2 <- cor.sep(X = newx2, theta = f2$theta, nu = 2.5)
      cov.Xnewx2 <- cor.sep(X = f2$X, x = newx2, theta = f2$theta, nu = 2.5)
    }
    v.next2 <- drop(cov.newx2 - crossprod(t(crossprod(cov.Xnewx2, f2$Ki)), cov.Xnewx2))
    g.next2 <- -1 / drop(v.next2) * crossprod(f2$Ki, cov.Xnewx2)
    
    fit2$Ki <- rbind(
      cbind(f2$Ki + crossprod(t(g.next2), t(g.next2) * v.next2), g.next2),
      cbind(t(g.next2), 1 / drop(v.next2))
    )
    
    fit2$X <- rbind(f2$X, newx2)
    fit2$y <- c(f2$y, y2.sample)
    fit2$tau2hat <- drop(crossprod(t(crossprod(fit2$y - fit2$mu.hat, fit2$Ki)), (fit2$y - fit2$mu.hat)) / length(fit2$y))
    
    fit.tmp$fits[[2]] <- fit2
    
    ALC.out <- mean(predict(fit.tmp, Xref)$sig2[[2]])
  }
  return(mean(ALC.out))
}


#' object to optimize the next point by ALC criterion updating at level 1 with three levels of fidelity
#'
#' @param Xcand candidate data point to be optimized.
#' @param Xref vector or matrix of reference data.
#' @param fit an object of class RNAmf.
#' @param mc.sample a number of mc samples generated for this approach. Default is 100.
#' @param parallel logical indicating whether to run parallel or not. Default is FALSE.
#' @param ncore the number of core for parallel. Default is 1.
#' @return A mean of the deduced variance at Xref.
#' @importFrom plgp covar.sep
#' @importFrom stats rnorm
#' @noRd
#'

obj.ALC_level_3_1 <- function(Xcand, Xref, fit, MC=FALSE, mc.sample, parallel = FALSE, ncore = 1) {
  fit1 <- f1 <- fit$fits[[1]]
  fit2 <- f2 <- fit$fits[[2]]
  fit3 <- f3 <- fit$fits[[3]]
  constant <- fit$constant
  kernel <- fit$kernel
  g <- fit1$g
  
  Xcand <- matrix(Xcand, nrow = 1)
  if(MC){
    if (kernel == "sqex") {
      y1.sample <- rnorm(mc.sample, mean = pred.GP(f1, Xcand)$mu, sd = sqrt(pred.GP(f1, Xcand)$sig2))
    } else if (kernel == "matern1.5") {
      y1.sample <- rnorm(mc.sample, mean = pred.matGP(f1, Xcand)$mu, sd = sqrt(pred.matGP(f1, Xcand)$sig2))
    } else if (kernel == "matern2.5") {
      y1.sample <- rnorm(mc.sample, mean = pred.matGP(f1, Xcand)$mu, sd = sqrt(pred.matGP(f1, Xcand)$sig2))
    }
  }else{
    if (kernel == "sqex") {
      y1.sample <- pred.GP(f1, Xcand)$mu
    } else if (kernel == "matern1.5") {
      y1.sample <- pred.matGP(f1, Xcand)$mu
    } else if (kernel == "matern2.5") {
      y1.sample <- pred.matGP(f1, Xcand)$mu
    }
  }
  Xcand1 <- Xcand
  
  ### Choose level 1 ###
  ### update Ki1
  if (kernel == "sqex") {
    cov.newx1 <- covar.sep(X1 = Xcand1, d = f1$theta, g = g)
    cov.Xnewx1 <- covar.sep(X1 = f1$X, X2 = Xcand1, d = f1$theta, g = 0)
  } else if (kernel == "matern1.5") {
    cov.newx1 <- cor.sep(X = Xcand1, theta = f1$theta, nu = 1.5)
    cov.Xnewx1 <- cor.sep(X = f1$X, x = Xcand1, theta = f1$theta, nu = 1.5)
  } else if (kernel == "matern2.5") {
    cov.newx1 <- cor.sep(X = Xcand1, theta = f1$theta, nu = 2.5)
    cov.Xnewx1 <- cor.sep(X = f1$X, x = Xcand1, theta = f1$theta, nu = 2.5)
  }
  v.next1 <- drop(cov.newx1 - crossprod(t(crossprod(cov.Xnewx1, f1$Ki)), cov.Xnewx1))
  g.next1 <- -1 / drop(v.next1) * crossprod(f1$Ki, cov.Xnewx1)
  
  
  fit1$Ki <- rbind(
    cbind(f1$Ki + crossprod(t(g.next1), t(g.next1) * v.next1), g.next1),
    cbind(t(g.next1), 1 / drop(v.next1))
  )
  
  fit1$X <- rbind(f1$X, Xcand1)
  
  if(MC){
    if (parallel) {
      ALC.out <- foreach(i = 1:mc.sample, .combine = c) %dorng% {
        fit.tmp <- fit
        fit1$y <- c(f1$y, y1.sample[i])
        fit1$tau2hat <- drop(crossprod(t(crossprod(fit1$y - fit1$mu.hat, fit1$Ki)), (fit1$y - fit1$mu.hat)) / length(fit1$y))
        
        fit.tmp$fits[[1]] <- fit1
        
        return(mean(predict(fit.tmp, Xref)$sig2[[3]])) # to minimize the deduced variance. To maximize, -mean
      }
      attr(ALC.out, "rng") <- NULL
      attr(ALC.out, "doRNG_version") <- NULL
    } else {
      ALC.out <- rep(0, mc.sample)
      for (i in 1:mc.sample) {
        fit.tmp <- fit
        fit1$y <- c(f1$y, y1.sample[i])
        fit1$tau2hat <- drop(crossprod(t(crossprod(fit1$y - fit1$mu.hat, fit1$Ki)), (fit1$y - fit1$mu.hat)) / length(fit1$y))
        
        fit.tmp$fits[[1]] <- fit1
        
        ALC.out[i] <- mean(predict(fit.tmp, Xref)$sig2[[3]]) # to minimize the deduced variance. To maximize, -mean
      }
    }
  }else{
    fit.tmp <- fit
    fit1$y <- c(f1$y, y1.sample)
    fit1$tau2hat <- drop(crossprod(t(crossprod(fit1$y - fit1$mu.hat, fit1$Ki)), (fit1$y - fit1$mu.hat)) / length(fit1$y))
    
    fit.tmp$fits[[1]] <- fit1
    
    ALC.out <- mean(predict(fit.tmp, Xref)$sig2[[3]])
  }
  return(mean(ALC.out))
}


#' object to optimize the next point by ALC criterion updating at level 2 with three levels of fidelity
#'
#' @param Xcand candidate data point to be optimized.
#' @param Xref vector or matrix of reference data.
#' @param fit an object of class RNAmf.
#' @param mc.sample a number of mc samples generated for this approach. Default is 100.
#' @param parallel logical indicating whether to run parallel or not. Default is FALSE.
#' @param ncore the number of core for parallel. Default is 1.
#' @return A mean of the deduced variance at Xref.
#' @importFrom plgp covar.sep
#' @importFrom stats rnorm
#' @noRd
#'

obj.ALC_level_3_2 <- function(Xcand, Xref, fit, MC=FALSE, mc.sample, parallel = FALSE, ncore = 1) {
  fit1 <- f1 <- fit$fits[[1]]
  fit2 <- f2 <- fit$fits[[2]]
  fit3 <- f3 <- fit$fits[[3]]
  constant <- fit$constant
  kernel <- fit$kernel
  g <- fit1$g
  
  Xcand <- matrix(Xcand, nrow = 1)
  if(MC){
    if (kernel == "sqex") {
      y1.sample <- rnorm(mc.sample, mean = pred.GP(f1, Xcand)$mu, sd = sqrt(pred.GP(f1, Xcand)$sig2))
    } else if (kernel == "matern1.5") {
      y1.sample <- rnorm(mc.sample, mean = pred.matGP(f1, Xcand)$mu, sd = sqrt(pred.matGP(f1, Xcand)$sig2))
    } else if (kernel == "matern2.5") {
      y1.sample <- rnorm(mc.sample, mean = pred.matGP(f1, Xcand)$mu, sd = sqrt(pred.matGP(f1, Xcand)$sig2))
    }
  }else{
    if (kernel == "sqex") {
      y1.sample <- pred.GP(f1, Xcand)$mu
    } else if (kernel == "matern1.5") {
      y1.sample <- pred.matGP(f1, Xcand)$mu
    } else if (kernel == "matern2.5") {
      y1.sample <- pred.matGP(f1, Xcand)$mu
    }
  }
  Xcand1 <- Xcand
  
  ### Choose level 1 ###
  ### update Ki1
  if (kernel == "sqex") {
    cov.newx1 <- covar.sep(X1 = Xcand1, d = f1$theta, g = g)
    cov.Xnewx1 <- covar.sep(X1 = f1$X, X2 = Xcand1, d = f1$theta, g = 0)
  } else if (kernel == "matern1.5") {
    cov.newx1 <- cor.sep(X = Xcand1, theta = f1$theta, nu = 1.5)
    cov.Xnewx1 <- cor.sep(X = f1$X, x = Xcand1, theta = f1$theta, nu = 1.5)
  } else if (kernel == "matern2.5") {
    cov.newx1 <- cor.sep(X = Xcand1, theta = f1$theta, nu = 2.5)
    cov.Xnewx1 <- cor.sep(X = f1$X, x = Xcand1, theta = f1$theta, nu = 2.5)
  }
  v.next1 <- drop(cov.newx1 - crossprod(t(crossprod(cov.Xnewx1, f1$Ki)), cov.Xnewx1))
  g.next1 <- -1 / drop(v.next1) * crossprod(f1$Ki, cov.Xnewx1)
  
  
  fit1$Ki <- rbind(
    cbind(f1$Ki + crossprod(t(g.next1), t(g.next1) * v.next1), g.next1),
    cbind(t(g.next1), 1 / drop(v.next1))
  )
  
  fit1$X <- rbind(f1$X, Xcand1)
  
  if(MC){
    if (parallel) {
      ALC.out <- foreach(i = 1:mc.sample, .combine = c) %dorng% {
        fit.tmp <- fit
        fit1$y <- c(f1$y, y1.sample[i])
        fit1$tau2hat <- drop(crossprod(t(crossprod(fit1$y - fit1$mu.hat, fit1$Ki)), (fit1$y - fit1$mu.hat)) / length(fit1$y))
        
        fit.tmp$fits[[1]] <- fit1
        
        ### Choose level 2 ###
        if (kernel == "sqex") {
          pred2 <- pred.GP(fit2, cbind(Xcand, y1.sample[i]))
          y2.sample <- rnorm(1, pred2$mu, sqrt(pred2$sig2))
        } else if (kernel == "matern1.5") {
          pred2 <- pred.matGP(fit2, cbind(Xcand, y1.sample[i]))
          y2.sample <- rnorm(1, pred2$mu, sqrt(pred2$sig2))
        } else if (kernel == "matern2.5") {
          pred2 <- pred.matGP(fit2, cbind(Xcand, y1.sample[i]))
          y2.sample <- rnorm(1, pred2$mu, sqrt(pred2$sig2))
        }
        
        ### update Ki2
        newx2 <- cbind(Xcand, y1.sample[i])
        
        if (kernel == "sqex") {
          cov.newx2 <- covar.sep(X1 = newx2, d = f2$theta, g = g)
          cov.Xnewx2 <- covar.sep(X1 = f2$X, X2 = newx2, d = f2$theta, g = 0)
        } else if (kernel == "matern1.5") {
          cov.newx2 <- cor.sep(X = newx2, theta = f2$theta, nu = 1.5)
          cov.Xnewx2 <- cor.sep(X = f2$X, x = newx2, theta = f2$theta, nu = 1.5)
        } else if (kernel == "matern2.5") {
          cov.newx2 <- cor.sep(X = newx2, theta = f2$theta, nu = 2.5)
          cov.Xnewx2 <- cor.sep(X = f2$X, x = newx2, theta = f2$theta, nu = 2.5)
        }
        v.next2 <- drop(cov.newx2 - crossprod(t(crossprod(cov.Xnewx2, f2$Ki)), cov.Xnewx2))
        g.next2 <- -1 / drop(v.next2) * crossprod(f2$Ki, cov.Xnewx2)
        
        fit2$Ki <- rbind(
          cbind(f2$Ki + crossprod(t(g.next2), t(g.next2) * v.next2), g.next2),
          cbind(t(g.next2), 1 / drop(v.next2))
        )
        
        fit2$X <- rbind(f2$X, newx2)
        fit2$y <- c(f2$y, y2.sample)
        fit2$tau2hat <- drop(crossprod(t(crossprod(fit2$y - fit2$mu.hat, fit2$Ki)), (fit2$y - fit2$mu.hat)) / length(fit2$y))
        
        fit.tmp$fits[[2]] <- fit2
        
        return(mean(predict(fit.tmp, Xref)$sig2[[3]])) # to minimize the deduced variance. To maximize, -mean
      }
      attr(ALC.out, "rng") <- NULL
      attr(ALC.out, "doRNG_version") <- NULL
    } else {
      ALC.out <- rep(0, mc.sample)
      for (i in 1:mc.sample) {
        fit.tmp <- fit
        fit1$y <- c(f1$y, y1.sample[i])
        fit1$tau2hat <- drop(crossprod(t(crossprod(fit1$y - fit1$mu.hat, fit1$Ki)), (fit1$y - fit1$mu.hat)) / length(fit1$y))
        
        fit.tmp$fits[[1]] <- fit1
        
        ### Choose level 2 ###
        if (kernel == "sqex") {
          pred2 <- pred.GP(fit2, cbind(Xcand, y1.sample[i]))
          y2.sample <- rnorm(1, pred2$mu, sqrt(pred2$sig2))
        } else if (kernel == "matern1.5") {
          pred2 <- pred.matGP(fit2, cbind(Xcand, y1.sample[i]))
          y2.sample <- rnorm(1, pred2$mu, sqrt(pred2$sig2))
        } else if (kernel == "matern2.5") {
          pred2 <- pred.matGP(fit2, cbind(Xcand, y1.sample[i]))
          y2.sample <- rnorm(1, pred2$mu, sqrt(pred2$sig2))
        }
        
        ### update Ki2
        newx2 <- cbind(Xcand, y1.sample[i])
        
        if (kernel == "sqex") {
          cov.newx2 <- covar.sep(X1 = newx2, d = f2$theta, g = g)
          cov.Xnewx2 <- covar.sep(X1 = f2$X, X2 = newx2, d = f2$theta, g = 0)
        } else if (kernel == "matern1.5") {
          cov.newx2 <- cor.sep(X = newx2, theta = f2$theta, nu = 1.5)
          cov.Xnewx2 <- cor.sep(X = f2$X, x = newx2, theta = f2$theta, nu = 1.5)
        } else if (kernel == "matern2.5") {
          cov.newx2 <- cor.sep(X = newx2, theta = f2$theta, nu = 2.5)
          cov.Xnewx2 <- cor.sep(X = f2$X, x = newx2, theta = f2$theta, nu = 2.5)
        }
        v.next2 <- drop(cov.newx2 - crossprod(t(crossprod(cov.Xnewx2, f2$Ki)), cov.Xnewx2))
        g.next2 <- -1 / drop(v.next2) * crossprod(f2$Ki, cov.Xnewx2)
        
        fit2$Ki <- rbind(
          cbind(f2$Ki + crossprod(t(g.next2), t(g.next2) * v.next2), g.next2),
          cbind(t(g.next2), 1 / drop(v.next2))
        )
        
        fit2$X <- rbind(f2$X, newx2)
        fit2$y <- c(f2$y, y2.sample)
        fit2$tau2hat <- drop(crossprod(t(crossprod(fit2$y - fit2$mu.hat, fit2$Ki)), (fit2$y - fit2$mu.hat)) / length(fit2$y))
        
        fit.tmp$fits[[2]] <- fit2
        
        ALC.out[i] <- mean(predict(fit.tmp, Xref)$sig2[[3]]) # to minimize the deduced variance. To maximize, -mean
      }
    }
  }else{
    fit.tmp <- fit
    fit1$y <- c(f1$y, y1.sample)
    fit1$tau2hat <- drop(crossprod(t(crossprod(fit1$y - fit1$mu.hat, fit1$Ki)), (fit1$y - fit1$mu.hat)) / length(fit1$y))
    
    fit.tmp$fits[[1]] <- fit1
    
    ### Choose level 2 ###
    if (kernel == "sqex") {
      pred2 <- pred.GP(fit2, cbind(Xcand, y1.sample))
    } else if (kernel == "matern1.5") {
      pred2 <- pred.matGP(fit2, cbind(Xcand, y1.sample))
    } else if (kernel == "matern2.5") {
      pred2 <- pred.matGP(fit2, cbind(Xcand, y1.sample))
    }
    y2.sample <- pred2$mu
    
    ### update Ki2
    newx2 <- cbind(Xcand, y1.sample)
    
    if (kernel == "sqex") {
      cov.newx2 <- covar.sep(X1 = newx2, d = f2$theta, g = g)
      cov.Xnewx2 <- covar.sep(X1 = f2$X, X2 = newx2, d = f2$theta, g = 0)
    } else if (kernel == "matern1.5") {
      cov.newx2 <- cor.sep(X = newx2, theta = f2$theta, nu = 1.5)
      cov.Xnewx2 <- cor.sep(X = f2$X, x = newx2, theta = f2$theta, nu = 1.5)
    } else if (kernel == "matern2.5") {
      cov.newx2 <- cor.sep(X = newx2, theta = f2$theta, nu = 2.5)
      cov.Xnewx2 <- cor.sep(X = f2$X, x = newx2, theta = f2$theta, nu = 2.5)
    }
    v.next2 <- drop(cov.newx2 - crossprod(t(crossprod(cov.Xnewx2, f2$Ki)), cov.Xnewx2))
    g.next2 <- -1 / drop(v.next2) * crossprod(f2$Ki, cov.Xnewx2)
    
    fit2$Ki <- rbind(
      cbind(f2$Ki + crossprod(t(g.next2), t(g.next2) * v.next2), g.next2),
      cbind(t(g.next2), 1 / drop(v.next2))
    )
    
    fit2$X <- rbind(f2$X, newx2)
    fit2$y <- c(f2$y, y2.sample)
    fit2$tau2hat <- drop(crossprod(t(crossprod(fit2$y - fit2$mu.hat, fit2$Ki)), (fit2$y - fit2$mu.hat)) / length(fit2$y))
    
    fit.tmp$fits[[2]] <- fit2
    
    ALC.out <- mean(predict(fit.tmp, Xref)$sig2[[3]])
  }
  return(mean(ALC.out))
}


#' object to optimize the next point by ALC criterion updating at level 3 with three levels of fidelity
#'
#' @param Xcand candidate data point to be optimized.
#' @param Xref vector or matrix of reference data.
#' @param fit an object of class RNAmf.
#' @param mc.sample a number of mc samples generated for this approach. Default is 100.
#' @param parallel logical indicating whether to run parallel or not. Default is FALSE.
#' @param ncore the number of core for parallel. Default is 1.
#' @return A mean of the deduced variance at Xref.
#' @importFrom plgp covar.sep
#' @importFrom stats rnorm
#' @noRd
#'

obj.ALC_level_3_3 <- function(Xcand, Xref, fit, MC=FALSE, mc.sample, parallel = FALSE, ncore = 1) {
  fit1 <- f1 <- fit$fits[[1]]
  fit2 <- f2 <- fit$fits[[2]]
  fit3 <- f3 <- fit$fits[[3]]
  constant <- fit$constant
  kernel <- fit$kernel
  g <- fit1$g
  
  Xcand <- matrix(Xcand, nrow = 1)
  if(MC){
    if (kernel == "sqex") {
      y1.sample <- rnorm(mc.sample, mean = pred.GP(f1, Xcand)$mu, sd = sqrt(pred.GP(f1, Xcand)$sig2))
    } else if (kernel == "matern1.5") {
      y1.sample <- rnorm(mc.sample, mean = pred.matGP(f1, Xcand)$mu, sd = sqrt(pred.matGP(f1, Xcand)$sig2))
    } else if (kernel == "matern2.5") {
      y1.sample <- rnorm(mc.sample, mean = pred.matGP(f1, Xcand)$mu, sd = sqrt(pred.matGP(f1, Xcand)$sig2))
    }
  }else{
    if (kernel == "sqex") {
      y1.sample <- pred.GP(f1, Xcand)$mu
    } else if (kernel == "matern1.5") {
      y1.sample <- pred.matGP(f1, Xcand)$mu
    } else if (kernel == "matern2.5") {
      y1.sample <- pred.matGP(f1, Xcand)$mu
    }
  }
  Xcand1 <- Xcand
  
  ### Choose level 1 ###
  ### update Ki1
  if (kernel == "sqex") {
    cov.newx1 <- covar.sep(X1 = Xcand1, d = f1$theta, g = g)
    cov.Xnewx1 <- covar.sep(X1 = f1$X, X2 = Xcand1, d = f1$theta, g = 0)
  } else if (kernel == "matern1.5") {
    cov.newx1 <- cor.sep(X = Xcand1, theta = f1$theta, nu = 1.5)
    cov.Xnewx1 <- cor.sep(X = f1$X, x = Xcand1, theta = f1$theta, nu = 1.5)
  } else if (kernel == "matern2.5") {
    cov.newx1 <- cor.sep(X = Xcand1, theta = f1$theta, nu = 2.5)
    cov.Xnewx1 <- cor.sep(X = f1$X, x = Xcand1, theta = f1$theta, nu = 2.5)
  }
  v.next1 <- drop(cov.newx1 - crossprod(t(crossprod(cov.Xnewx1, f1$Ki)), cov.Xnewx1))
  g.next1 <- -1 / drop(v.next1) * crossprod(f1$Ki, cov.Xnewx1)
  
  
  fit1$Ki <- rbind(
    cbind(f1$Ki + crossprod(t(g.next1), t(g.next1) * v.next1), g.next1),
    cbind(t(g.next1), 1 / drop(v.next1))
  )
  
  fit1$X <- rbind(f1$X, Xcand1)
  
  if(MC){
    if (parallel) {
      ALC.out <- foreach(i = 1:mc.sample, .combine = c) %dorng% {
        fit.tmp <- fit
        fit1$y <- c(f1$y, y1.sample[i])
        fit1$tau2hat <- drop(crossprod(t(crossprod(fit1$y - fit1$mu.hat, fit1$Ki)), (fit1$y - fit1$mu.hat)) / length(fit1$y))
        
        fit.tmp$fits[[1]] <- fit1
        
        ### Choose level 2 ###
        if (kernel == "sqex") {
          pred2 <- pred.GP(fit2, cbind(Xcand, y1.sample[i]))
          y2.sample <- rnorm(1, pred2$mu, sqrt(pred2$sig2))
        } else if (kernel == "matern1.5") {
          pred2 <- pred.matGP(fit2, cbind(Xcand, y1.sample[i]))
          y2.sample <- rnorm(1, pred2$mu, sqrt(pred2$sig2))
        } else if (kernel == "matern2.5") {
          pred2 <- pred.matGP(fit2, cbind(Xcand, y1.sample[i]))
          y2.sample <- rnorm(1, pred2$mu, sqrt(pred2$sig2))
        }
        
        ### update Ki2
        newx2 <- cbind(Xcand, y1.sample[i])
        
        if (kernel == "sqex") {
          cov.newx2 <- covar.sep(X1 = newx2, d = f2$theta, g = g)
          cov.Xnewx2 <- covar.sep(X1 = f2$X, X2 = newx2, d = f2$theta, g = 0)
        } else if (kernel == "matern1.5") {
          cov.newx2 <- cor.sep(X = newx2, theta = f2$theta, nu = 1.5)
          cov.Xnewx2 <- cor.sep(X = f2$X, x = newx2, theta = f2$theta, nu = 1.5)
        } else if (kernel == "matern2.5") {
          cov.newx2 <- cor.sep(X = newx2, theta = f2$theta, nu = 2.5)
          cov.Xnewx2 <- cor.sep(X = f2$X, x = newx2, theta = f2$theta, nu = 2.5)
        }
        v.next2 <- drop(cov.newx2 - crossprod(t(crossprod(cov.Xnewx2, f2$Ki)), cov.Xnewx2))
        g.next2 <- -1 / drop(v.next2) * crossprod(f2$Ki, cov.Xnewx2)
        
        fit2$Ki <- rbind(
          cbind(f2$Ki + crossprod(t(g.next2), t(g.next2) * v.next2), g.next2),
          cbind(t(g.next2), 1 / drop(v.next2))
        )
        
        fit2$X <- rbind(f2$X, newx2)
        fit2$y <- c(f2$y, y2.sample)
        fit2$tau2hat <- drop(crossprod(t(crossprod(fit2$y - fit2$mu.hat, fit2$Ki)), (fit2$y - fit2$mu.hat)) / length(fit2$y))
        
        fit.tmp$fits[[2]] <- fit2
        
        ### Choose level 3 ###
        if (kernel == "sqex") {
          pred3 <- pred.GP(fit2, cbind(Xcand, y2.sample))
          y3.sample <- rnorm(1, pred3$mu, sqrt(pred3$sig2))
        } else if (kernel == "matern1.5") {
          pred3 <- pred.matGP(fit2, cbind(Xcand, y2.sample))
          y3.sample <- rnorm(1, pred3$mu, sqrt(pred3$sig2))
        } else if (kernel == "matern2.5") {
          pred3 <- pred.matGP(fit2, cbind(Xcand, y2.sample))
          y3.sample <- rnorm(1, pred3$mu, sqrt(pred3$sig2))
        }
        
        ### update Ki2
        newx3 <- cbind(Xcand, y2.sample)
        
        if (kernel == "sqex") {
          cov.newx3 <- covar.sep(X1 = newx3, d = f3$theta, g = g)
          cov.Xnewx3 <- covar.sep(X1 = f3$X, X2 = newx3, d = f3$theta, g = 0)
        } else if (kernel == "matern1.5") {
          cov.newx3 <- cor.sep(X = newx3, theta = f3$theta, nu = 1.5)
          cov.Xnewx3 <- cor.sep(X = f3$X, x = newx3, theta = f3$theta, nu = 1.5)
        } else if (kernel == "matern2.5") {
          cov.newx3 <- cor.sep(X = newx3, theta = f3$theta, nu = 2.5)
          cov.Xnewx3 <- cor.sep(X = f3$X, x = newx3, theta = f3$theta, nu = 2.5)
        }
        v.next3 <- drop(cov.newx3 - crossprod(t(crossprod(cov.Xnewx3, f3$Ki)), cov.Xnewx3))
        g.next3 <- -1 / drop(v.next3) * crossprod(f3$Ki, cov.Xnewx3)
        
        fit3$Ki <- rbind(
          cbind(f3$Ki + crossprod(t(g.next3), t(g.next3) * v.next3), g.next3),
          cbind(t(g.next3), 1 / drop(v.next3))
        )
        
        fit3$X <- rbind(f3$X, newx3)
        fit3$y <- c(f3$y, y3.sample)
        fit3$tau2hat <- drop(crossprod(t(crossprod(fit3$y - fit3$mu.hat, fit3$Ki)), (fit3$y - fit3$mu.hat)) / length(fit3$y))
        
        fit.tmp$fits[[3]] <- fit3
        
        return(mean(predict(fit.tmp, Xref)$sig2[[3]])) # to minimize the deduced variance. To maximize, -mean
      }
      attr(ALC.out, "rng") <- NULL
      attr(ALC.out, "doRNG_version") <- NULL
    } else {
      ALC.out <- rep(0, mc.sample)
      for (i in 1:mc.sample) {
        fit.tmp <- fit
        fit1$y <- c(f1$y, y1.sample[i])
        fit1$tau2hat <- drop(crossprod(t(crossprod(fit1$y - fit1$mu.hat, fit1$Ki)), (fit1$y - fit1$mu.hat)) / length(fit1$y))
        
        fit.tmp$fits[[1]] <- fit1
        
        ### Choose level 2 ###
        if (kernel == "sqex") {
          pred2 <- pred.GP(fit2, cbind(Xcand, y1.sample[i]))
          y2.sample <- rnorm(1, pred2$mu, sqrt(pred2$sig2))
        } else if (kernel == "matern1.5") {
          pred2 <- pred.matGP(fit2, cbind(Xcand, y1.sample[i]))
          y2.sample <- rnorm(1, pred2$mu, sqrt(pred2$sig2))
        } else if (kernel == "matern2.5") {
          pred2 <- pred.matGP(fit2, cbind(Xcand, y1.sample[i]))
          y2.sample <- rnorm(1, pred2$mu, sqrt(pred2$sig2))
        }
        
        ### update Ki2
        newx2 <- cbind(Xcand, y1.sample[i])
        
        if (kernel == "sqex") {
          cov.newx2 <- covar.sep(X1 = newx2, d = f2$theta, g = g)
          cov.Xnewx2 <- covar.sep(X1 = f2$X, X2 = newx2, d = f2$theta, g = 0)
        } else if (kernel == "matern1.5") {
          cov.newx2 <- cor.sep(X = newx2, theta = f2$theta, nu = 1.5)
          cov.Xnewx2 <- cor.sep(X = f2$X, x = newx2, theta = f2$theta, nu = 1.5)
        } else if (kernel == "matern2.5") {
          cov.newx2 <- cor.sep(X = newx2, theta = f2$theta, nu = 2.5)
          cov.Xnewx2 <- cor.sep(X = f2$X, x = newx2, theta = f2$theta, nu = 2.5)
        }
        v.next2 <- drop(cov.newx2 - crossprod(t(crossprod(cov.Xnewx2, f2$Ki)), cov.Xnewx2))
        g.next2 <- -1 / drop(v.next2) * crossprod(f2$Ki, cov.Xnewx2)
        
        fit2$Ki <- rbind(
          cbind(f2$Ki + crossprod(t(g.next2), t(g.next2) * v.next2), g.next2),
          cbind(t(g.next2), 1 / drop(v.next2))
        )
        
        fit2$X <- rbind(f2$X, newx2)
        fit2$y <- c(f2$y, y2.sample)
        fit2$tau2hat <- drop(crossprod(t(crossprod(fit2$y - fit2$mu.hat, fit2$Ki)), (fit2$y - fit2$mu.hat)) / length(fit2$y))
        
        fit.tmp$fits[[2]] <- fit2
        
        ### Choose level 3 ###
        if (kernel == "sqex") {
          pred3 <- pred.GP(fit2, cbind(Xcand, y2.sample))
          y3.sample <- rnorm(1, pred3$mu, sqrt(pred3$sig2))
        } else if (kernel == "matern1.5") {
          pred3 <- pred.matGP(fit2, cbind(Xcand, y2.sample))
          y3.sample <- rnorm(1, pred3$mu, sqrt(pred3$sig2))
        } else if (kernel == "matern2.5") {
          pred3 <- pred.matGP(fit2, cbind(Xcand, y2.sample))
          y3.sample <- rnorm(1, pred3$mu, sqrt(pred3$sig2))
        }
        
        ### update Ki2
        newx3 <- cbind(Xcand, y2.sample)
        
        if (kernel == "sqex") {
          cov.newx3 <- covar.sep(X1 = newx3, d = f3$theta, g = g)
          cov.Xnewx3 <- covar.sep(X1 = f3$X, X2 = newx3, d = f3$theta, g = 0)
        } else if (kernel == "matern1.5") {
          cov.newx3 <- cor.sep(X = newx3, theta = f3$theta, nu = 1.5)
          cov.Xnewx3 <- cor.sep(X = f3$X, x = newx3, theta = f3$theta, nu = 1.5)
        } else if (kernel == "matern2.5") {
          cov.newx3 <- cor.sep(X = newx3, theta = f3$theta, nu = 2.5)
          cov.Xnewx3 <- cor.sep(X = f3$X, x = newx3, theta = f3$theta, nu = 2.5)
        }
        v.next3 <- drop(cov.newx3 - crossprod(t(crossprod(cov.Xnewx3, f3$Ki)), cov.Xnewx3))
        g.next3 <- -1 / drop(v.next3) * crossprod(f3$Ki, cov.Xnewx3)
        
        fit3$Ki <- rbind(
          cbind(f3$Ki + crossprod(t(g.next3), t(g.next3) * v.next3), g.next3),
          cbind(t(g.next3), 1 / drop(v.next3))
        )
        
        fit3$X <- rbind(f3$X, newx3)
        fit3$y <- c(f3$y, y3.sample)
        fit3$tau2hat <- drop(crossprod(t(crossprod(fit3$y - fit3$mu.hat, fit3$Ki)), (fit3$y - fit3$mu.hat)) / length(fit3$y))
        
        fit.tmp$fits[[3]] <- fit3
        
        ALC.out[i] <- mean(predict(fit.tmp, Xref)$sig2[[3]]) # to minimize the deduced variance. To maximize, -mean
      }
    }
  }else{
    fit.tmp <- fit
    fit1$y <- c(f1$y, y1.sample)
    fit1$tau2hat <- drop(crossprod(t(crossprod(fit1$y - fit1$mu.hat, fit1$Ki)), (fit1$y - fit1$mu.hat)) / length(fit1$y))
    
    fit.tmp$fits[[1]] <- fit1
    
    ### Choose level 2 ###
    if (kernel == "sqex") {
      pred2 <- pred.GP(fit2, cbind(Xcand, y1.sample))
    } else if (kernel == "matern1.5") {
      pred2 <- pred.matGP(fit2, cbind(Xcand, y1.sample))
    } else if (kernel == "matern2.5") {
      pred2 <- pred.matGP(fit2, cbind(Xcand, y1.sample))
    }
    y2.sample <- pred2$mu
    
    ### update Ki2
    newx2 <- cbind(Xcand, y1.sample)
    
    if (kernel == "sqex") {
      cov.newx2 <- covar.sep(X1 = newx2, d = f2$theta, g = g)
      cov.Xnewx2 <- covar.sep(X1 = f2$X, X2 = newx2, d = f2$theta, g = 0)
    } else if (kernel == "matern1.5") {
      cov.newx2 <- cor.sep(X = newx2, theta = f2$theta, nu = 1.5)
      cov.Xnewx2 <- cor.sep(X = f2$X, x = newx2, theta = f2$theta, nu = 1.5)
    } else if (kernel == "matern2.5") {
      cov.newx2 <- cor.sep(X = newx2, theta = f2$theta, nu = 2.5)
      cov.Xnewx2 <- cor.sep(X = f2$X, x = newx2, theta = f2$theta, nu = 2.5)
    }
    v.next2 <- drop(cov.newx2 - crossprod(t(crossprod(cov.Xnewx2, f2$Ki)), cov.Xnewx2))
    g.next2 <- -1 / drop(v.next2) * crossprod(f2$Ki, cov.Xnewx2)
    
    fit2$Ki <- rbind(
      cbind(f2$Ki + crossprod(t(g.next2), t(g.next2) * v.next2), g.next2),
      cbind(t(g.next2), 1 / drop(v.next2))
    )
    
    fit2$X <- rbind(f2$X, newx2)
    fit2$y <- c(f2$y, y2.sample)
    fit2$tau2hat <- drop(crossprod(t(crossprod(fit2$y - fit2$mu.hat, fit2$Ki)), (fit2$y - fit2$mu.hat)) / length(fit2$y))
    
    fit.tmp$fits[[2]] <- fit2
    
    ### Choose level 3 ###
    if (kernel == "sqex") {
      pred3 <- pred.GP(fit2, cbind(Xcand, y2.sample))
    } else if (kernel == "matern1.5") {
      pred3 <- pred.matGP(fit2, cbind(Xcand, y2.sample))
    } else if (kernel == "matern2.5") {
      pred3 <- pred.matGP(fit2, cbind(Xcand, y2.sample))
    }
    y3.sample <- pred3$mu
    
    ### update Ki2
    newx3 <- cbind(Xcand, y2.sample)
    
    if (kernel == "sqex") {
      cov.newx3 <- covar.sep(X1 = newx3, d = f3$theta, g = g)
      cov.Xnewx3 <- covar.sep(X1 = f3$X, X2 = newx3, d = f3$theta, g = 0)
    } else if (kernel == "matern1.5") {
      cov.newx3 <- cor.sep(X = newx3, theta = f3$theta, nu = 1.5)
      cov.Xnewx3 <- cor.sep(X = f3$X, x = newx3, theta = f3$theta, nu = 1.5)
    } else if (kernel == "matern2.5") {
      cov.newx3 <- cor.sep(X = newx3, theta = f3$theta, nu = 2.5)
      cov.Xnewx3 <- cor.sep(X = f3$X, x = newx3, theta = f3$theta, nu = 2.5)
    }
    v.next3 <- drop(cov.newx3 - crossprod(t(crossprod(cov.Xnewx3, f3$Ki)), cov.Xnewx3))
    g.next3 <- -1 / drop(v.next3) * crossprod(f3$Ki, cov.Xnewx3)
    
    fit3$Ki <- rbind(
      cbind(f3$Ki + crossprod(t(g.next3), t(g.next3) * v.next3), g.next3),
      cbind(t(g.next3), 1 / drop(v.next3))
    )
    
    fit3$X <- rbind(f3$X, newx3)
    fit3$y <- c(f3$y, y3.sample)
    fit3$tau2hat <- drop(crossprod(t(crossprod(fit3$y - fit3$mu.hat, fit3$Ki)), (fit3$y - fit3$mu.hat)) / length(fit3$y))
    
    fit.tmp$fits[[3]] <- fit3
    
    ALC.out <- mean(predict(fit.tmp, Xref)$sig2[[3]])
  }
  return(mean(ALC.out))
}


#' @title Active Learning for Recursive Non-Additive Emulator
#'
#' @description The function acquires the new point and fidelity level by maximizing one of the
#' four active learning criteria: ALM, ALC, ALD, or ALMC.
#'
#' \itemize{
#'   \item \strong{ALM} (Active Learning MacKay): It calculates the ALM criterion \eqn{\frac{\sigma^{*2}_l(\bm{x})}{\sum^l_{j=1}C_j}},
#'   where \eqn{\sigma^{*2}_l(\bm{x})} is the posterior predictive variance
#'   at each fidelity level \eqn{l} and \eqn{C_j} is the simulation cost at level \eqn{j}.
#'
#'   \item \strong{ALD} (Active Learning Decomposition): It calculates the ALD criterion \eqn{\frac{V_l(\bm{x})}{\sum^l_{j=1}C_j}},
#'   where \eqn{V_l(\bm{x})} is the variance contribution of GP emulator
#'   at each fidelity level \eqn{l} and \eqn{C_j} is the simulation cost at level \eqn{j}.
#'
#'   \item \strong{ALC} (Active Learning Cohn): It calculates the ALC criterion
#'   \eqn{\frac{\Delta \sigma_L^{2}(l,\bm{x})}{\sum^l_{j=1}C_j} =
#'   \frac{\int_{\Omega} \sigma_L^{*2}(\bm{\xi})-\tilde{\sigma}_L^{*2}(\bm{\xi};l,\bm{x}){\rm{d}}\bm{\xi}}{\sum^l_{j=1}C_j}},
#'   where \eqn{f_L} is the highest-fidelity simulation code,
#'   \eqn{\sigma_L^{*2}(\bm{\xi})} is the posterior variance of \eqn{f_L(\bm{\xi})},
#'   \eqn{C_j} is the simulation cost at fidelity level \eqn{j},
#'   and \eqn{\tilde{\sigma}_L^{*2}(\bm{\xi};l,\bm{x})} is the posterior variance
#'   based on the augmented design combining the current design and a new input location \eqn{\bm{x}}
#'   at each fidelity level lower than or equal to \eqn{l}.
#'   The integration is approximated by MC integration using uniform reference samples.
#'
#'   \item \strong{ALMC} (Active Learning MacKay-Cohn): A hybrid approach.
#'   It finds the optimal input location \eqn{\bm{x}^*}
#'   by maximizing \eqn{\sigma^{*2}_L(\bm{x})},
#'   the posterior predictive variance at the highest-fidelity level \eqn{L}.
#'   After selecting \eqn{\bm{x}^*},
#'   it finds the optimal fidelity level by maximizing ALC criterion at \eqn{\bm{x}^*},
#'   \eqn{\text{argmax}_{l\in\{1,\ldots,L\}} \frac{\Delta \sigma_L^{2}(l,\bm{x}^*)}{\sum^l_{j=1}C_j}},
#'   where \eqn{C_j} is the simulation cost at level \eqn{j}.
#' }
#'
#' A new point is acquired on \code{Xcand}. If \code{Xcand=NULL} and \code{Xref=NULL}, a new point is acquired on unit hypercube \eqn{[0,1]^d}.
#'
#' For details, see Heo and Sung (2025, <\doi{https://doi.org/10.1080/00401706.2024.2376173}>).
#'
#' @details For \code{"ALC"}, or \code{"ALMC"}, \code{Xref} plays a role of \eqn{\bm{\xi}} to approximate the integration.
#' To impute the posterior variance based on the augmented design \eqn{\tilde{\sigma}_L^{*2}(\bm{\xi};l,\bm{x})},
#' MC approximation is used.
#' Due to the nested assumption, imputing \eqn{y^{[s]}_{n_s+1}} for each \eqn{1\leq s\leq l} by drawing samples
#' from the posterior distribution of \eqn{f_s(\bm{x}^{[s]}_{n_s+1})}
#' based on the current design allows to compute \eqn{\tilde{\sigma}_L^{*2}(\bm{\xi};l,\bm{x})}.
#' Inverse of covariance matrix is computed by the Sherman-Morrison formula.
#'
#' To search for the next acquisition \eqn{\bm{x}^*} by maximizing AL criterion,
#' the gradient-based optimization can be used by \code{optim=TRUE}.
#' Firstly, \eqn{\tilde{\sigma}_L^{*2}(\bm{\xi};l,\bm{x})} is computed at
#' \eqn{5 \times d} number of points.
#' After that, the point minimizing \eqn{\tilde{\sigma}_L^{*2}(\bm{\xi};l,\bm{x})}
#' serves as a starting point of optimization by \code{L-BFGS-B} method.
#' Otherwise, when \code{optim=FALSE}, AL criterion is optimized only on \code{Xcand}.
#'
#' The point is selected by maximizing the ALC criterion:
#' \eqn{\text{argmax}_{l\in\{1,\ldots,L\}; \bm{x} \in \Omega}
#' \frac{\Delta \sigma_L^{2}(l,\bm{x})}{\sum^l_{j=1}C_j}}.
#'
#' @param criterion character string specifying the active learning criterion to use. Must be one of \code{"ALM"}, \code{"ALD"}, \code{"ALC"}, or \code{"ALMC"}. Default is \code{"ALM"}.
#' @param fit object of class \code{RNAmf}.
#' @param Xref vector or matrix of reference locations to approximate the integral of ALC. If \code{Xref=NULL}, \eqn{100 \times d} points from 0 to 1 are generated by Latin hypercube design. Only used when \code{criterion="ALC"} or \code{"ALMC"}. Default is \code{NULL}.
#' @param Xcand vector or matrix of the candidate set for grid-based search. If \code{use_optim=FALSE}, the criterion is evaluated and optimized only on this set. If \code{Xcand=NULL}, \eqn{100 \times d} points from 0 to 1 generated by Latin hypercube design (or \code{Xref} for ALC and ALMC) are used. Default is \code{NULL}.
#' @param MC logical indicating whether to use Monte Carlo approximation to impute the posterior variance (for ALC/ALMC). If \code{FALSE}, posterior means are used. Default is \code{FALSE}.
#' @param mc.sample a number of MC samples generated for the imputation through MC approximation. Default is \code{100}.
#' @param cost vector of the costs for each level of fidelity. If \code{cost=NULL}, total costs at all levels would be 1. \code{cost} is encouraged to have an ascending order of positive values. Default is \code{NULL}.
#' @param use_optim logical indicating whether to optimize the criterion using \code{optim}'s gradient-based \code{L-BFGS-B} method. If \code{TRUE}, \eqn{5 \times d} starting points are generated by Latin hypercube design for optimization. If \code{FALSE}, the point is selected from \code{Xcand}. Default is \code{TRUE}.
#' @param parallel logical indicating whether to use parallel computation. Default is \code{FALSE}.
#' @param ncore integer specifying the number of cores for parallel computation. Used only if \code{parallel=TRUE}. Default is 1.
#' @param trace logical indicating whether to print the computational progress and time. Default is \code{TRUE}.
#'
#' @return A list containing:
#' \itemize{
#'   \item \code{AL}: The values of the selected criterion at the candidate points (if \code{use_optim=FALSE}) or the optimized value (if \code{use_optim=TRUE}). For \code{ALMC}, it returns the ALC scores for each level at the chosen point.
#'   \item \code{cost}: A copy of the \code{cost} argument.
#'   \item \code{Xcand}: A copy of the \code{Xcand} argument used.
#'   \item \code{chosen}: A list containing the chosen fidelity \code{level} and the new input location \code{Xnext}.
#'   \item \code{time}: The computation time in seconds.
#' }
#'
#' @importFrom plgp covar.sep
#' @importFrom stats rnorm optim
#' @importFrom lhs randomLHS
#' @importFrom foreach foreach
#' @importFrom doRNG %dorng%
#' @importFrom doParallel registerDoParallel stopImplicitCluster
#'
#' @usage AL_RNAmf(criterion = c("ALM", "ALC", "ALD", "ALMC"), fit,
#' Xref = NULL, Xcand = NULL, MC = FALSE, mc.sample = 100, cost = NULL,
#' use_optim = TRUE, parallel = FALSE, ncore = 1, trace = TRUE)
#' @export
#' @examples
#' \donttest{
#' ### simulation costs ###
#' cost <- c(1, 3)
#'
#' ### 1-d Perdikaris function in Perdikaris, et al. (2017) ###
#' # low-fidelity function
#' f1 <- function(x) {
#'   sin(8 * pi * x)
#' }
#'
#' # high-fidelity function
#' f2 <- function(x) {
#'   (x - sqrt(2)) * (sin(8 * pi * x))^2
#' }
#'
#' ### training data ###
#' n1 <- 13
#' n2 <- 8
#'
#' ### fix seed to reproduce the result ###
#' set.seed(1)
#'
#' ### generate initial nested design ###
#' X <- NestedX(c(n1, n2), 1)
#' X1 <- X[[1]]
#' X2 <- X[[2]]
#'
#' y1 <- f1(X1)
#' y2 <- f2(X2)
#'
#' ### n=100 uniform test data ###
#' x <- seq(0, 1, length.out = 100)
#'
#' ### fit an RNAmf ###
#' fit.RNAmf <- RNAmf(list(X1, X2), list(y1, y2), kernel = "sqex", constant=TRUE)
#'
#' ### 1. ALM Criterion ###
#' alm.RNAmf <- AL_RNAmf(criterion="ALM",
#'                       Xcand = x, fit=fit.RNAmf, cost = cost,
#'                       use_optim = FALSE, parallel = TRUE, ncore = 2)
#' print(alm.RNAmf$chosen)
#'
#' ### visualize ALM ###
#' oldpar <- par(mfrow = c(1, 2))
#' plot(x, alm.RNAmf$AL$ALM1,
#'      type = "l", lty = 2,
#'      xlab = "x", ylab = "ALM criterion at the low-fidelity level",
#'      ylim = c(min(c(alm.RNAmf$AL$ALM1, alm.RNAmf$AL$ALM2)),
#'               max(c(alm.RNAmf$AL$ALM1, alm.RNAmf$AL$ALM2))))
#' points(alm.RNAmf$chosen$Xnext,
#'        alm.RNAmf$AL$ALM1[which(x == drop(alm.RNAmf$chosen$Xnext))],
#'        pch = 16, cex = 1, col = "red")
#' plot(x, alm.RNAmf$AL$ALM2,
#'      type = "l", lty = 2,
#'      xlab = "x", ylab = "ALM criterion at the high-fidelity level",
#'      ylim = c(min(c(alm.RNAmf$AL$ALM1, alm.RNAmf$AL$ALM2)),
#'               max(c(alm.RNAmf$AL$ALM1, alm.RNAmf$AL$ALM2))))
#' par(oldpar)
#'
#'
#' ### 2. ALD Criterion ###
#' ald.RNAmf <- AL_RNAmf(criterion="ALD",
#'                       Xcand = x, fit=fit.RNAmf, cost = cost,
#'                       use_optim = FALSE, parallel = TRUE, ncore = 2)
#' print(ald.RNAmf$chosen)
#'
#' ### visualize ALD ###
#' oldpar <- par(mfrow = c(1, 2))
#' plot(x, ald.RNAmf$AL$ALD1,
#'      type = "l", lty = 2,
#'      xlab = "x", ylab = "ALD criterion at the low-fidelity level",
#'      ylim = c(min(c(ald.RNAmf$AL$ALD1, ald.RNAmf$AL$ALD2)),
#'               max(c(ald.RNAmf$AL$ALD1, ald.RNAmf$AL$ALD2))))
#' points(ald.RNAmf$chosen$Xnext,
#'        ald.RNAmf$AL$ALD1[which(x == drop(ald.RNAmf$chosen$Xnext))],
#'        pch = 16, cex = 1, col = "red")
#' plot(x, ald.RNAmf$AL$ALD2,
#'      type = "l", lty = 2,
#'      xlab = "x", ylab = "ALD criterion at the high-fidelity level",
#'      ylim = c(min(c(ald.RNAmf$AL$ALD1, ald.RNAmf$AL$ALD2)),
#'               max(c(ald.RNAmf$AL$ALD1, ald.RNAmf$AL$ALD2))))
#' par(oldpar)
#'
#'
#' ### 3. ALC Criterion ###
#' alc.RNAmf <- AL_RNAmf(criterion="ALC",
#'                       Xref = x, Xcand = x, fit=fit.RNAmf, cost = cost,
#'                       use_optim = FALSE, parallel = TRUE, ncore = 2)
#' print(alc.RNAmf$chosen)
#'
#' ### visualize ALC ###
#' oldpar <- par(mfrow = c(1, 2))
#' plot(x, alc.RNAmf$AL$ALC1,
#'      type = "l", lty = 2,
#'      xlab = "x", ylab = "ALC criterion augmented at the low-fidelity level",
#'      ylim = c(min(c(alc.RNAmf$AL$ALC1, alc.RNAmf$AL$ALC2)),
#'               max(c(alc.RNAmf$AL$ALC1, alc.RNAmf$AL$ALC2))))
#' points(alc.RNAmf$chosen$Xnext,
#'        alc.RNAmf$AL$ALC1[which(x == drop(alc.RNAmf$chosen$Xnext))],
#'        pch = 16, cex = 1, col = "red")
#' plot(x, alc.RNAmf$AL$ALC2,
#'      type = "l", lty = 2,
#'      xlab = "x", ylab = "ALC criterion augmented at the high-fidelity level",
#'      ylim = c(min(c(alc.RNAmf$AL$ALC1, alc.RNAmf$AL$ALC2)),
#'               max(c(alc.RNAmf$AL$ALC1, alc.RNAmf$AL$ALC2))))
#' par(oldpar)
#'
#'
#' ### 4. ALMC Criterion ###
#' almc.RNAmf <- AL_RNAmf(criterion="ALMC",
#'                        Xref = x, Xcand = x, fit=fit.RNAmf, cost = cost,
#'                        use_optim = FALSE, parallel = TRUE, ncore = 2)
#' print(almc.RNAmf$chosen)
#' }

AL_RNAmf <- function(criterion = c("ALM", "ALC", "ALD", "ALMC"), fit, Xref = NULL, Xcand = NULL, MC = FALSE, mc.sample = 100, cost = NULL, use_optim = TRUE, parallel = FALSE, ncore = 1, trace = TRUE) {
  t0 <- proc.time()
  ### check the object ###
  if (!inherits(fit, "RNAmf")) stop("The object is not of class \"RNAmf\" \n")
  criterion <- match.arg(criterion)
  L <- fit$level
  d <- ncol(fit$fits[[1]]$X)
  
  ### check the cost ###
  if (is.null(cost)) {
    cost <- c(1, rep(0, L - 1))
  } else {
    if (!is.numeric(cost)) stop("'cost' must be a numeric vector.")
    if (length(cost) != L) stop(paste0("The length of 'cost' must equal the number of fidelity levels (", L, ")."))
    if (any(cost < 0)) stop("'cost' must contain non-negative values.")
    for(l in 1:(L-1)) {
      if (cost[l] >= cost[l+1]) warning(paste("Cost for level", l, "is >= level", l+1, ". Usually, higher fidelity is more expensive."))
    }
  }
  
  ### Xref check ###
  if (!is.null(Xref)) {
    if (is.vector(Xref)) Xref <- matrix(Xref, ncol = 1)
    if (!is.matrix(Xref)) stop("'Xref' must be a matrix or vector.")
    if (ncol(Xref) != d) stop(paste0("'Xref' must have ", d, " columns (matching the model input dimension)."))
  }
  ### Xcand check ###
  if (!is.null(Xcand)) {
    if (is.vector(Xcand)) Xcand <- matrix(Xcand, ncol = 1)
    if (!is.matrix(Xcand)) stop("'Xcand' must be a matrix or vector.")
    if (ncol(Xcand) != d) stop(paste0("'Xcand' must have ", d, " columns (matching the model input dimension)."))
  }
  
  if (!is.logical(MC)) stop("'MC' must be TRUE or FALSE.")
  if (!is.numeric(mc.sample) || length(mc.sample) != 1 || mc.sample < 1) stop("'mc.sample' must be a positive integer.")
  if (!is.logical(use_optim)) stop("'use_optim' must be TRUE or FALSE.")
  if (!is.logical(parallel)) stop("'parallel' must be TRUE or FALSE.")
  if (!is.numeric(ncore) || length(ncore) != 1 || ncore < 1) stop("'ncore' must be a positive integer.")
  if (!is.logical(trace)) stop("'trace' must be TRUE or FALSE.")
  
  if ((criterion %in% c("ALC", "ALMC")) && is.null(Xref)) {
    Xref <- randomLHS(100 * d, d)
  }
  
  ### parallel setup ###
  if (parallel) {
    registerDoParallel(cores = ncore)
    on.exit(stopImplicitCluster(), add = TRUE)
  }
  
  ### Xcand setup ###
  if (use_optim) {
    Xcand <- randomLHS(5 * d, d)
  } else {
    if (is.null(Xcand)) {
      Xcand <- randomLHS(100 * d, d)
    } else if (is.null(dim(Xcand))) {
      Xcand <- matrix(Xcand, ncol = 1)
    }
  }
  
  ### Helper: Get Function Name ###
  get_obj_func_name <- function(crit, l, total_L) {
    if (crit == "ALM") return(paste0("obj.ALM_level_", l))
    if (crit == "ALC") {
      if (total_L == 3) return(paste0("obj.ALC_level_3_", l))
      return(paste0("obj.ALC_level_", l))
    }
    if (crit == "ALD") {
      suffix <- if (total_L == 2) "2level" else "3level"
      return(paste0("obj.ALD_V", l, "_", suffix))
    }
    return(NULL)
  }
  
  ### Helper: Current Variance (for ALC/ALMC) ###
  Icurrent <- 0
  if (criterion %in% c("ALC", "ALMC")) {
    pred_sigs <- predict(fit, Xref)$sig2
    Icurrent <- mean(pred_sigs[[L]])
  }
  
  if (criterion == "ALMC") { ### ALMC ###
    ### 1. Find Xnext using ALM at the highest level (L) ###
    obj_fun <- get(get_obj_func_name("ALM", L, L))
    
    # A. Grid Search on Xcand
    if (parallel) {
      vals <- foreach(i = 1:nrow(Xcand), .combine = c, .export = c("pred.GP", "covar.sep", "distance"), .packages = "RNAmf") %dorng% {
        -obj_fun(matrix(Xcand[i, ], nrow = 1), fit = fit)
      }
      attr(vals, "rng") <- NULL
      attr(vals, "doRNG_version") <- NULL
    } else {
      vals <- numeric(nrow(Xcand))
      for(i in 1:nrow(Xcand)) vals[i] <- -obj_fun(matrix(Xcand[i, ], nrow = 1), fit = fit)
    }
    
    # B. Optimization on Xcand
    best_idx <- which.max(vals)
    X.start <- matrix(Xcand[best_idx, ], nrow = 1)
    if (use_optim) {
      o <- optim(X.start, obj_fun, method = "L-BFGS-B", lower = 0, upper = 1, fit = fit)
      Xnext <- matrix(o$par, nrow = 1)
    } else {
      Xnext <- X.start
    }
    
    if(trace) cat("Finding the next point (ALMC):", (proc.time() - t0)[3], "seconds\n")
    
    ### 2. Calculate ALC scores for all levels at the chosen point Xnext ###
    ALMC_scores <- numeric(L)
    ALMC_vals <- numeric(L)
    
    for (l in 1:L) {
      alc_func <- get(get_obj_func_name("ALC", l, L))
      val <- alc_func(Xnext, Xref = Xref, fit = fit, MC = MC, mc.sample = mc.sample, parallel = parallel, ncore = ncore)
      ALMC_vals[l] <- Icurrent - val
      ALMC_scores[l] <- (Icurrent - val) / sum(cost[1:l])
    }
    
    if(trace) cat("Calculating deduced variance:", (proc.time() - t0)[3], "seconds\n")
    
    best_level <- which.max(ALMC_scores)
    final_Xnext <- c(Xnext)
    AL_out <- ALMC_vals / cumsum(cost)
    names(AL_out) <- paste0("ALMC", 1:L)
    
  } else { ### ALM, ALD, ALC ###
    AL_out <- vector("list", L)
    level_scores <- numeric(L)
    best_pars <- list()
    
    for (l in 1:L) {
      obj_fun <- get(get_obj_func_name(criterion, l, L))
      
      # 1. Grid Search on Xcand
      if (criterion == "ALC") {
        wrapper <- function(x) obj_fun(x, Xref = Xref, fit = fit, MC = MC, mc.sample = mc.sample)
      } else if (criterion == "ALD" && L == 3 && l < 3) {
        wrapper <- function(x) obj_fun(x, fit = fit, mc.sample = mc.sample)
      } else {
        wrapper <- function(x) obj_fun(x, fit = fit)
      }
      
      if (parallel) {
        vals <- foreach(i = 1:nrow(Xcand), .combine = c, .export = c("pred.GP", "covar.sep", "distance"), .packages = "RNAmf") %dorng% {
          val <- wrapper(matrix(Xcand[i, ], nrow = 1))
          if (criterion %in% c("ALM", "ALD")) return(-val) else return(val)
        }
        attr(vals, "rng") <- NULL
        attr(vals, "doRNG_version") <- NULL
      } else {
        vals <- numeric(nrow(Xcand))
        for (i in 1:nrow(Xcand)) {
          val <- wrapper(matrix(Xcand[i, ], nrow = 1))
          vals[i] <- if (criterion %in% c("ALM", "ALD")) -val else val
        }
      }
      
      # 2. Optimization on Xcand
      if (criterion %in% c("ALM", "ALD")) best_idx <- which.max(vals) else best_idx <- which.min(vals)
      X.start <- matrix(Xcand[best_idx, ], nrow = 1)
      
      if (use_optim) {
        if (criterion == "ALC") {
          o <- optim(par = X.start, fn = obj_fun, method = "L-BFGS-B", lower = 0, upper = 1, fit = fit, Xref = Xref, MC = MC, mc.sample = mc.sample)
        } else if (criterion == "ALD" && L == 3 && l < 3) { # ALD V1 or V2 when L = 3
          o <- optim(par = X.start, fn = obj_fun, method = "L-BFGS-B", lower = 0, upper = 1, fit = fit, mc.sample = mc.sample)
        } else { # ALM or ALD V3
          o <- optim(par = X.start, fn = obj_fun, method = "L-BFGS-B", lower = 0, upper = 1, fit = fit)
        }
        
        best_val <- if (criterion %in% c("ALM", "ALD")) -o$value else o$value
        best_par <- o$par
      } else {
        best_val <- vals[best_idx]
        best_par <- X.start
      }
      
      # Calculate Score
      if (criterion == "ALC") {
        score <- (Icurrent - best_val) / sum(cost[1:l])
        grid_out <- vals
      } else {
        score <- best_val / sum(cost[1:l])
        grid_out <- vals
      }
      
      level_scores[l] <- score
      best_pars[[l]] <- best_par
      AL_out[[l]] <- grid_out / sum(cost[1:l])
    }
    
    names(AL_out) <- paste0(criterion, 1:L)
    best_level <- which.max(level_scores)
    final_Xnext <- c(matrix(best_pars[[best_level]], nrow=1))
    
    if(trace) cat("Calculation completed in:", (proc.time() - t0)[3], "seconds\n")
  }
  
  return(list(AL = AL_out, cost = cost, Xcand = Xcand, chosen = list(level = best_level, Xnext = final_Xnext), time = unname((proc.time() - t0)[3])))
}

#' subsetting matrix to be nested
#'
#' @param X1 matrix of the design at the lower level
#' @param X2 matrix of the design at the higher level
#' @return A list containing the design and the length of its subset
#' \itemize{
#'   \item \code{X}: matrix of the design at the lower level.
#'   \item \code{le}: length of the indices of subset.
#' }
#' @noRd
#'

subsetX <- function(X1 = NULL, X2 = NULL) {
  d <- dim(as.matrix(X2))[2] # d2
  n2 <- dim(as.matrix(X2))[1] # n2
  n1 <- dim(as.matrix(X1))[1] # n1
  if (n2 > n1) stop("Need n1 >= n2 to nest: lower level must have >= points than higher level.")

  a <- rowSums(X2^2)                # length n2
  b <- rowSums(X1^2)                # length n1
  D <- outer(a, b, "+") - 2 * tcrossprod(X2, X1)  # n2 x n1 squared distances

  taken <- rep(FALSE, n1)
  indice <- integer(n2)

  # process rows in order of their best available distance (helps quality)
  row_order <- order(apply(D, 1, min))
  for (r in row_order) {
    # try nearest, then next nearest, until we find a free column
    cols <- order(D[r, ])
    c <- cols[which(!taken[cols])[1]]
    if (is.na(c)) stop("Greedy matching failed (shouldn't happen if n1 >= n2).")
    taken[c] <- TRUE
    indice[r] <- c
  }

  X1 <- matrix(X1[-indice, ], ncol=d)
  X1 <- rbind(X1, matrix(X2, ncol=d))

  return(list(X = X1, le = length(indice)))
}

#' @title Constructing nested design sets for the RNA model.
#'
#' @description The function constructs nested design sets with multiple fidelity levels
#' \eqn{\mathcal{X}_l \subseteq \cdots \subseteq \mathcal{X}_{1}} for use in \code{\link{RNAmf}}.
#'
#' @details The procedure replace the points of lower level design \eqn{\mathcal{X}_{l-1}}
#' with the closest points from higher level design \eqn{\mathcal{X}_{l}}.
#' For details, see "\href{https://github.com/cran/MuFiCokriging}{\code{NestedDesign}}".
#'
#' @references
#' L. Le Gratiet and J. Garnier (2014). Recursive co-kriging model for design of computer experiments
#' with multiple levels of fidelity. \emph{International Journal for Uncertainty Quantification}, 4(5), 365-386;
#' \doi{doi:10.1615/Int.J.UncertaintyQuantification.2014006914}
#'
#' @param n A vector specifying the number of design points at each fidelity level \eqn{l}. Thus, the vector must have a positive value \eqn{n_1, \ldots, n_l} where \eqn{n_1 > \cdots > n_l}.
#' @param d A positive integer specifying the dimension of the design.
#' @return A list containing the nested design sets at each level, i.e., \eqn{\mathcal{X}_{1}, \ldots, \mathcal{X}_{l}}.
#' @importFrom lhs maximinLHS
#' @export
#' @examples
#' ### number of design points ###
#' n1 <- 30
#' n2 <- 15
#'
#' ### dimension of the design ###
#' d <- 2
#'
#' ### fix seed to reproduce the result ###
#' set.seed(1)
#'
#' ### generate the nested design ###
#' NX <- NestedX(c(n1, n2), d)
#'
#' ### visualize nested design ###
#' plot(NX[[1]], col="red", pch=1, xlab="x1", ylab="x2")
#' points(NX[[2]], col="blue", pch=4)
#'

NestedX <- function(n, d) { # n; vector, d; dim

  # Input validation
  if (is.unsorted(rev(n), strictly = TRUE)) {
    stop("The number of design at each level must be descending order \n")
  }
  if (!all(n > 0)) {
    stop("The number of design at each level must be positive \n")
  }
  if (length(d) != 1) {
    stop("The dimension of design at each level must be same \n")
  }

  level <- length(n)
  if (level < 2) {
    stop("The level of design should be larger than 1 \n")
  }

  # Generating initial designs
  X <- list()
  for (i in 1:level) {
    X[[i]] <- maximinLHS(n[i], d)
  }

  # Subsetting designs
  indices <- list()
  for (i in (level - 1):1) {
    SB <- subsetX(matrix(X[[i]], ncol=d), matrix(X[[i + 1]], ncol=d))
    X[[i]] <- SB$X
    n_rows <- nrow(SB$X)
    indices[[i]] <- seq(n_rows - SB$le + 1, n_rows, by = 1)
  }

  # Constructing nested designs
  X_list <- list()
  current_idx <- 1:nrow(X[[1]])
  for (k in 1:level) {
    if (k > 1) {
      current_idx <- current_idx[indices[[k-1]]]
    }
    X_list[[k]] <- matrix(X[[1]][current_idx, ], ncol = d)
  }

  return(X = X_list)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict.RNAmf.R
\name{predict.RNAmf}
\alias{predict.RNAmf}
\title{prediction of the RNAmf emulator with multiple fidelity levels.}
\usage{
\method{predict}{RNAmf}(object, x = NULL, nimpute = 50, ...)
}
\arguments{
\item{object}{An object of class \code{RNAmf} fitted by \code{\link{RNAmf}}.}

\item{x}{A vector or matrix of new input locations for prediction.}

\item{nimpute}{Number of imputations for non-nested designs. Default is 50.}

\item{...}{Additional arguments for compatibility with generic method \code{predict}.}
}
\value{
\itemize{
  \item \code{mu}: A list of vectors containing the predictive posterior mean at each fidelity level.
  \item \code{sig2}: A list of vectors containing the predictive posterior variance at each fidelity level.
  \item \code{time}: A scalar indicating the computation time.
}
}
\description{
The function computes the posterior mean and variance of RNA models with multiple fidelity levels
by fitted model from \code{\link{RNAmf}}.
}
\details{
The \code{predict.RNAmf} function internally calls \code{\link{closed_form_RNA}}
to recursively compute the closed-form posterior mean and variance at each level.

From the fitted model from \code{\link{RNAmf}},
the posterior mean and variance are calculated based on the closed-form expression derived by a recursive fashion.
The formulas depend on its kernel choices.
For further details, see Heo and Sung (2025, <\doi{https://doi.org/10.1080/00401706.2024.2376173}>).
}
\examples{
\donttest{
### two levels example ###

### Perdikaris function ###
f1 <- function(x) {
  sin(8 * pi * x)
}

f2 <- function(x) {
  (x - sqrt(2)) * (sin(8 * pi * x))^2
}

### training data ###
n1 <- 13
n2 <- 8

### fix seed to reproduce the result ###
set.seed(1)

### generate initial nested design ###
X <- NestedX(c(n1, n2), 1)
X1 <- X[[1]]
X2 <- X[[2]]

y1 <- f1(X1)
y2 <- f2(X2)

### n=100 uniform test data ###
x <- seq(0, 1, length.out = 100)

### fit an RNAmf ###
fit.RNAmf <- RNAmf(list(X1, X2), list(y1, y2), kernel = "sqex", constant=TRUE)

### predict ###
predy <- predict(fit.RNAmf, x)$mu[[2]]
predsig2 <- predict(fit.RNAmf, x)$sig2[[2]]

### RMSE ###
print(sqrt(mean((predy - f2(x))^2)))

### visualize the emulation performance ###
plot(x, predy,
  type = "l", lwd = 2, col = 3, # emulator and confidence interval
  ylim = c(-2, 1)
)
lines(x, predy + 1.96 * sqrt(predsig2 * length(y2) / (length(y2) - 2)), col = 3, lty = 2)
lines(x, predy - 1.96 * sqrt(predsig2 * length(y2) / (length(y2) - 2)), col = 3, lty = 2)

curve(f2(x), add = TRUE, col = 1, lwd = 2, lty = 2) # high fidelity function

points(X1, y1, pch = 1, col = "red") # low-fidelity design
points(X2, y2, pch = 4, col = "blue") # high-fidelity design

### three levels example ###
### Branin function ###
branin <- function(xx, l){
  x1 <- xx[1]
  x2 <- xx[2]
  if(l == 1){
    10*sqrt((-1.275*(1.2*x1+0.4)^2/pi^2+5*(1.2*x1+0.4)/pi+(1.2*x2+0.4)-6)^2 +
    (10-5/(4*pi))*cos((1.2*x1+0.4))+ 10) + 2*(1.2*x1+1.9) - 3*(3*(1.2*x2+2.4)-1) - 1 - 3*x2 + 1
  }else if(l == 2){
    10*sqrt((-1.275*(x1+2)^2/pi^2+5*(x1+2)/pi+(x2+2)-6)^2 +
    (10-5/(4*pi))*cos((x1+2))+ 10) + 2*(x1-0.5) - 3*(3*x2-1) - 1
  }else if(l == 3){
    (-1.275*x1^2/pi^2+5*x1/pi+x2-6)^2 + (10-5/(4*pi))*cos(x1)+ 10
  }
}

output.branin <- function(x, l){
  factor_range <- list("x1" = c(-5, 10), "x2" = c(0, 15))

  for(i in 1:length(factor_range)) x[i] <- factor_range[[i]][1] + x[i] * diff(factor_range[[i]])
  branin(x[1:2], l)
}

### training data ###
n1 <- 20; n2 <- 15; n3 <- 10

### fix seed to reproduce the result ###
set.seed(1)

### generate initial nested design ###
X <- NestedX(c(n1, n2, n3), 2)
X1 <- X[[1]]
X2 <- X[[2]]
X3 <- X[[3]]

y1 <- apply(X1,1,output.branin, l=1)
y2 <- apply(X2,1,output.branin, l=2)
y3 <- apply(X3,1,output.branin, l=3)

### n=10000 grid test data ###
x <- as.matrix(expand.grid(seq(0, 1, length.out = 100),seq(0, 1, length.out = 100)))

### fit an RNAmf ###
fit.RNAmf <- RNAmf(list(X1, X2, X3), list(y1, y2, y3), kernel = "sqex", constant=TRUE)

### predict ###
pred.RNAmf <- predict(fit.RNAmf, x)
predy <- pred.RNAmf$mu[[3]]
predsig2 <- pred.RNAmf$sig2[[3]]

### RMSE ###
print(sqrt(mean((predy - apply(x,1,output.branin, l=3))^2)))

### visualize the emulation performance ###
x1 <- x2 <- seq(0, 1, length.out = 100)
oldpar <- par(mfrow=c(1,2))
image(x1, x2, matrix(apply(x,1,output.branin, l=3), ncol=100),
zlim=c(0,310), main="Branin function")
image(x1, x2, matrix(predy, ncol=100),
zlim=c(0,310), main="RNAmf prediction")
par(oldpar)

### predictive variance ###
print(predsig2)}
}
\seealso{
\code{\link{RNAmf}} for model fitting.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict.RNAmf.R
\name{closed_form_RNA}
\alias{closed_form_RNA}
\title{Closed-form prediction for RNAmf model}
\usage{
closed_form_RNA(fits, x, kernel, XX = NULL, pseudo_yy = NULL)
}
\arguments{
\item{fits}{A fitted GP object from \code{RNAmf}.}

\item{x}{A vector or matrix of new input locations to predict.}

\item{kernel}{A character specifying the kernel type to be used. Choices are \code{"sqex"}(squared exponential kernel), \code{"matern1.5"}(Matern kernel with \eqn{\nu=1.5}), or \code{"matern2.5"}(Matern kernel with \eqn{\nu=2.5}). Default is \code{"sqex"}.}

\item{XX}{A list containing a pseudo-complete inputs \code{X_star}(\eqn{\left\{\mathcal{X}^*_l\right\}_{l=1}^{L}}), an original inputs \code{X_list}(\eqn{\left\{\mathcal{X}_l\right\}_{l=1}^{L}}), and a pseudo inputs \code{X_tilde}(\eqn{\left\{\widetilde{\mathcal{X}}_l\right\}_{l=1}^{L}}) for non-nested design.}

\item{pseudo_yy}{A list containing a pseudo-complete outputs \code{y_star}(\eqn{\left\{\mathbf{y}^*_l\right\}_{l=1}^{L}}), an original outputs \code{y_list}(\eqn{\left\{\mathbf{y}_l\right\}_{l=1}^{L}}), and a pseudo outputs \code{y_tilde}(\eqn{\left\{\widetilde{\mathbf{y}}_l\right\}_{l=1}^{L}}) imputed by \code{\link{imputer_RNA}}.}
}
\value{
A list of predictive posterior mean and variance for each level containing:
\itemize{
  \item \code{mu}: A list of predictive posterior mean at each fidelity level.
  \item \code{sig2}: A list of predictive posterior variance at each fidelity level.
}
}
\description{
The function computes the closed-form posterior mean and variance for the RNAmf model
both at the fidelity levels used in model fitting using the chosen kernel.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RNAmf.R
\name{RNAmf}
\alias{RNAmf}
\title{Fitting the Recursive Non-Additive model with multiple fidelity levels}
\usage{
RNAmf(X_list, y_list, kernel = "sqex", constant = TRUE,
init = NULL, n.iter = 50, burn.ratio = 0.75, trace = TRUE, ...)
}
\arguments{
\item{X_list}{A list of the matrices of input locations for all fidelity levels.}

\item{y_list}{A list of the vectors or matrices of response values for all fidelity levels.}

\item{kernel}{A character specifying the kernel type to be used. Choices are \code{"sqex"}(squared exponential), \code{"matern1.5"}, or \code{"matern2.5"}. Default is \code{"sqex"}.}

\item{constant}{A logical indicating for constant mean of GP (\code{constant=TRUE}) or zero mean (\code{constant=FALSE}). Default is \code{TRUE}.}

\item{init}{A list of the vectors of initial parameter values for optimization. Default is \code{NULL}.}

\item{n.iter}{Number of iterations for the stochastic EM algorithm for non-nested designs. Default is 50.}

\item{burn.ratio}{Fraction of iterations to discard as burn-in. Default is 0.75.}

\item{trace}{A logical indicating to print progress of iterations if \code{TRUE}, or not if \code{FALSE}. Default is \code{TRUE}.}

\item{...}{Additional arguments for compatibility with \code{optim}.}
}
\value{
A list of class \code{RNAmf} with:
\itemize{
  \item \code{fits}: A list of fitted Gaussian process models \eqn{f_l(x)} at each level \eqn{1, \ldots, l}. Each element contains:
  \eqn{\begin{cases} & f_1 \text{ for } (X_1, y_1),\\ & f_l \text{ for } ((X_l, f_{l-1}(X_l)), y_l), \end{cases}}.
  \item \code{level}: The number of fidelity levels \eqn{l}.
  \item \code{kernel}: A copy of \code{kernel}.
  \item \code{constant}: A copy of \code{constant}.
  \item \code{nested}: A logical indicating whether the design is nested.
  \item \code{time}: A scalar indicating the computation time.
}
}
\description{
The function fits RNA models with designs of multiple fidelity levels.
The estimation method is based on MLE.
Available kernel choices include the squared exponential kernel, and the Matern kernel with smoothness parameter 1.5 and 2.5.
The function returns the fitted model at each level \eqn{1, \ldots, l}, the number of fidelity levels \eqn{l}, the kernel choice, whether constant mean or not, and the computation time.
}
\details{
Consider the model
\eqn{\begin{cases}
& f_1(\bm{x}) = W_1(\bm{x}),\\
& f_l(\bm{x}) = W_l(\bm{x}, f_{l-1}(\bm{x})) \quad\text{for}\quad l \geq 2,
\end{cases}}
where \eqn{f_l} is the simulation code at fidelity level \eqn{l}, and
\eqn{W_l(\bm{x}) \sim GP(\alpha_l, \tau_l^2 K_l(\bm{x}, \bm{x}'))} is GP model.
Hyperparameters \eqn{(\alpha_l, \tau_l^2, \bm{\theta_l})} are estimated by
maximizing the log-likelihood via an optimization algorithm "L-BFGS-B".
For \code{constant=FALSE}, \eqn{\alpha_l=0}.

Covariance kernel is defined as:
\eqn{K_l(\bm{x}, \bm{x}')=\prod^d_{j=1}\phi(x_j,x'_j;\theta_{lj})} with
\eqn{\phi(x, x';\theta) = \exp \left( -\frac{ \left( x - x' \right)^2}{\theta}  \right)}
for squared exponential kernel; \code{kernel="sqex"},
\eqn{\phi(x,x';\theta) =\left( 1+\frac{\sqrt{3}|x- x'|}{\theta} \right) \exp \left( -\frac{\sqrt{3}|x- x'|}{\theta} \right)}
for Matern kernel with the smoothness parameter of 1.5; \code{kernel="matern1.5"} and
\eqn{\phi(x, x';\theta) = \left( 1+\frac{\sqrt{5}|x-x'|}{\theta} +\frac{5(x-x')^2}{3\theta^2} \right) \exp \left( -\frac{\sqrt{5}|x-x'|}{\theta} \right)}
for Matern kernel with the smoothness parameter of 2.5; \code{kernel="matern2.5"}.

For further details, see Heo and Sung (2025, <\doi{https://doi.org/10.1080/00401706.2024.2376173}>).
}
\examples{
\donttest{
### two levels example ###

### Perdikaris function ###
f1 <- function(x) {
  sin(8 * pi * x)
}

f2 <- function(x) {
  (x - sqrt(2)) * (sin(8 * pi * x))^2
}

### training data ###
n1 <- 13
n2 <- 8

### fix seed to reproduce the result ###
set.seed(1)

### generate initial nested design ###
X <- NestedX(c(n1, n2), 1)
X1 <- X[[1]]
X2 <- X[[2]]

y1 <- f1(X1)
y2 <- f2(X2)

### fit an RNAmf ###
fit.RNAmf <- RNAmf(list(X1, X2), list(y1, y2), kernel = "sqex", constant=TRUE)


### three levels example ###

### Branin function ###
branin <- function(xx, l){
  x1 <- xx[1]
  x2 <- xx[2]
  if(l == 1){
    10*sqrt((-1.275*(1.2*x1+0.4)^2/pi^2+5*(1.2*x1+0.4)/pi+(1.2*x2+0.4)-6)^2 +
              (10-5/(4*pi))*cos((1.2*x1+0.4))+ 10) +
              2*(1.2*x1+1.9) - 3*(3*(1.2*x2+2.4)-1) - 1 - 3*x2 + 1
  }else if(l == 2){
    10*sqrt((-1.275*(x1+2)^2/pi^2+5*(x1+2)/pi+(x2+2)-6)^2 +
              (10-5/(4*pi))*cos((x1+2))+ 10) + 2*(x1-0.5) - 3*(3*x2-1) - 1
  }else if(l == 3){
    (-1.275*x1^2/pi^2+5*x1/pi+x2-6)^2 + (10-5/(4*pi))*cos(x1)+ 10
  }
}

output.branin <- function(x, l){
  factor_range <- list("x1" = c(-5, 10), "x2" = c(0, 15))

  for(i in 1:length(factor_range)) x[i] <- factor_range[[i]][1] + x[i] * diff(factor_range[[i]])
  branin(x[1:2], l)
}

### training data ###
n1 <- 20; n2 <- 15; n3 <- 10

### fix seed to reproduce the result ###
set.seed(1)

### generate initial nested design ###
X <- NestedX(c(n1, n2, n3), 2)
X1 <- X[[1]]
X2 <- X[[2]]
X3 <- X[[3]]

y1 <- apply(X1,1,output.branin, l=1)
y2 <- apply(X2,1,output.branin, l=2)
y3 <- apply(X3,1,output.branin, l=3)

### fit an RNAmf ###
fit.RNAmf <- RNAmf(list(X1, X2, X3), list(y1, y2, y3), kernel = "sqex", constant=TRUE)
}


}
\seealso{
\code{\link{predict.RNAmf}} for prediction.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AL_RNAmf.R
\name{AL_RNAmf}
\alias{AL_RNAmf}
\title{Active Learning for Recursive Non-Additive Emulator}
\usage{
AL_RNAmf(criterion = c("ALM", "ALC", "ALD", "ALMC"), fit,
Xref = NULL, Xcand = NULL, MC = FALSE, mc.sample = 100, cost = NULL,
use_optim = TRUE, parallel = FALSE, ncore = 1, trace = TRUE)
}
\arguments{
\item{criterion}{character string specifying the active learning criterion to use. Must be one of \code{"ALM"}, \code{"ALD"}, \code{"ALC"}, or \code{"ALMC"}. Default is \code{"ALM"}.}

\item{fit}{object of class \code{RNAmf}.}

\item{Xref}{vector or matrix of reference locations to approximate the integral of ALC. If \code{Xref=NULL}, \eqn{100 \times d} points from 0 to 1 are generated by Latin hypercube design. Only used when \code{criterion="ALC"} or \code{"ALMC"}. Default is \code{NULL}.}

\item{Xcand}{vector or matrix of the candidate set for grid-based search. If \code{use_optim=FALSE}, the criterion is evaluated and optimized only on this set. If \code{Xcand=NULL}, \eqn{100 \times d} points from 0 to 1 generated by Latin hypercube design (or \code{Xref} for ALC and ALMC) are used. Default is \code{NULL}.}

\item{MC}{logical indicating whether to use Monte Carlo approximation to impute the posterior variance (for ALC/ALMC). If \code{FALSE}, posterior means are used. Default is \code{FALSE}.}

\item{mc.sample}{a number of MC samples generated for the imputation through MC approximation. Default is \code{100}.}

\item{cost}{vector of the costs for each level of fidelity. If \code{cost=NULL}, total costs at all levels would be 1. \code{cost} is encouraged to have an ascending order of positive values. Default is \code{NULL}.}

\item{use_optim}{logical indicating whether to optimize the criterion using \code{optim}'s gradient-based \code{L-BFGS-B} method. If \code{TRUE}, \eqn{5 \times d} starting points are generated by Latin hypercube design for optimization. If \code{FALSE}, the point is selected from \code{Xcand}. Default is \code{TRUE}.}

\item{parallel}{logical indicating whether to use parallel computation. Default is \code{FALSE}.}

\item{ncore}{integer specifying the number of cores for parallel computation. Used only if \code{parallel=TRUE}. Default is 1.}

\item{trace}{logical indicating whether to print the computational progress and time. Default is \code{TRUE}.}
}
\value{
A list containing:
\itemize{
  \item \code{AL}: The values of the selected criterion at the candidate points (if \code{use_optim=FALSE}) or the optimized value (if \code{use_optim=TRUE}). For \code{ALMC}, it returns the ALC scores for each level at the chosen point.
  \item \code{cost}: A copy of the \code{cost} argument.
  \item \code{Xcand}: A copy of the \code{Xcand} argument used.
  \item \code{chosen}: A list containing the chosen fidelity \code{level} and the new input location \code{Xnext}.
  \item \code{time}: The computation time in seconds.
}
}
\description{
The function acquires the new point and fidelity level by maximizing one of the
four active learning criteria: ALM, ALC, ALD, or ALMC.

\itemize{
  \item \strong{ALM} (Active Learning MacKay): It calculates the ALM criterion \eqn{\frac{\sigma^{*2}_l(\bm{x})}{\sum^l_{j=1}C_j}},
  where \eqn{\sigma^{*2}_l(\bm{x})} is the posterior predictive variance
  at each fidelity level \eqn{l} and \eqn{C_j} is the simulation cost at level \eqn{j}.

  \item \strong{ALD} (Active Learning Decomposition): It calculates the ALD criterion \eqn{\frac{V_l(\bm{x})}{\sum^l_{j=1}C_j}},
  where \eqn{V_l(\bm{x})} is the variance contribution of GP emulator
  at each fidelity level \eqn{l} and \eqn{C_j} is the simulation cost at level \eqn{j}.

  \item \strong{ALC} (Active Learning Cohn): It calculates the ALC criterion
  \eqn{\frac{\Delta \sigma_L^{2}(l,\bm{x})}{\sum^l_{j=1}C_j} =
  \frac{\int_{\Omega} \sigma_L^{*2}(\bm{\xi})-\tilde{\sigma}_L^{*2}(\bm{\xi};l,\bm{x}){\rm{d}}\bm{\xi}}{\sum^l_{j=1}C_j}},
  where \eqn{f_L} is the highest-fidelity simulation code,
  \eqn{\sigma_L^{*2}(\bm{\xi})} is the posterior variance of \eqn{f_L(\bm{\xi})},
  \eqn{C_j} is the simulation cost at fidelity level \eqn{j},
  and \eqn{\tilde{\sigma}_L^{*2}(\bm{\xi};l,\bm{x})} is the posterior variance
  based on the augmented design combining the current design and a new input location \eqn{\bm{x}}
  at each fidelity level lower than or equal to \eqn{l}.
  The integration is approximated by MC integration using uniform reference samples.

  \item \strong{ALMC} (Active Learning MacKay-Cohn): A hybrid approach.
  It finds the optimal input location \eqn{\bm{x}^*}
  by maximizing \eqn{\sigma^{*2}_L(\bm{x})},
  the posterior predictive variance at the highest-fidelity level \eqn{L}.
  After selecting \eqn{\bm{x}^*},
  it finds the optimal fidelity level by maximizing ALC criterion at \eqn{\bm{x}^*},
  \eqn{\text{argmax}_{l\in\{1,\ldots,L\}} \frac{\Delta \sigma_L^{2}(l,\bm{x}^*)}{\sum^l_{j=1}C_j}},
  where \eqn{C_j} is the simulation cost at level \eqn{j}.
}

A new point is acquired on \code{Xcand}. If \code{Xcand=NULL} and \code{Xref=NULL}, a new point is acquired on unit hypercube \eqn{[0,1]^d}.

For details, see Heo and Sung (2025, <\doi{https://doi.org/10.1080/00401706.2024.2376173}>).
}
\details{
For \code{"ALC"}, or \code{"ALMC"}, \code{Xref} plays a role of \eqn{\bm{\xi}} to approximate the integration.
To impute the posterior variance based on the augmented design \eqn{\tilde{\sigma}_L^{*2}(\bm{\xi};l,\bm{x})},
MC approximation is used.
Due to the nested assumption, imputing \eqn{y^{[s]}_{n_s+1}} for each \eqn{1\leq s\leq l} by drawing samples
from the posterior distribution of \eqn{f_s(\bm{x}^{[s]}_{n_s+1})}
based on the current design allows to compute \eqn{\tilde{\sigma}_L^{*2}(\bm{\xi};l,\bm{x})}.
Inverse of covariance matrix is computed by the Sherman-Morrison formula.

To search for the next acquisition \eqn{\bm{x}^*} by maximizing AL criterion,
the gradient-based optimization can be used by \code{optim=TRUE}.
Firstly, \eqn{\tilde{\sigma}_L^{*2}(\bm{\xi};l,\bm{x})} is computed at
\eqn{5 \times d} number of points.
After that, the point minimizing \eqn{\tilde{\sigma}_L^{*2}(\bm{\xi};l,\bm{x})}
serves as a starting point of optimization by \code{L-BFGS-B} method.
Otherwise, when \code{optim=FALSE}, AL criterion is optimized only on \code{Xcand}.

The point is selected by maximizing the ALC criterion:
\eqn{\text{argmax}_{l\in\{1,\ldots,L\}; \bm{x} \in \Omega}
\frac{\Delta \sigma_L^{2}(l,\bm{x})}{\sum^l_{j=1}C_j}}.
}
\examples{
\donttest{
### simulation costs ###
cost <- c(1, 3)

### 1-d Perdikaris function in Perdikaris, et al. (2017) ###
# low-fidelity function
f1 <- function(x) {
  sin(8 * pi * x)
}

# high-fidelity function
f2 <- function(x) {
  (x - sqrt(2)) * (sin(8 * pi * x))^2
}

### training data ###
n1 <- 13
n2 <- 8

### fix seed to reproduce the result ###
set.seed(1)

### generate initial nested design ###
X <- NestedX(c(n1, n2), 1)
X1 <- X[[1]]
X2 <- X[[2]]

y1 <- f1(X1)
y2 <- f2(X2)

### n=100 uniform test data ###
x <- seq(0, 1, length.out = 100)

### fit an RNAmf ###
fit.RNAmf <- RNAmf(list(X1, X2), list(y1, y2), kernel = "sqex", constant=TRUE)

### 1. ALM Criterion ###
alm.RNAmf <- AL_RNAmf(criterion="ALM",
                      Xcand = x, fit=fit.RNAmf, cost = cost,
                      use_optim = FALSE, parallel = TRUE, ncore = 2)
print(alm.RNAmf$chosen)

### visualize ALM ###
oldpar <- par(mfrow = c(1, 2))
plot(x, alm.RNAmf$AL$ALM1,
     type = "l", lty = 2,
     xlab = "x", ylab = "ALM criterion at the low-fidelity level",
     ylim = c(min(c(alm.RNAmf$AL$ALM1, alm.RNAmf$AL$ALM2)),
              max(c(alm.RNAmf$AL$ALM1, alm.RNAmf$AL$ALM2))))
points(alm.RNAmf$chosen$Xnext,
       alm.RNAmf$AL$ALM1[which(x == drop(alm.RNAmf$chosen$Xnext))],
       pch = 16, cex = 1, col = "red")
plot(x, alm.RNAmf$AL$ALM2,
     type = "l", lty = 2,
     xlab = "x", ylab = "ALM criterion at the high-fidelity level",
     ylim = c(min(c(alm.RNAmf$AL$ALM1, alm.RNAmf$AL$ALM2)),
              max(c(alm.RNAmf$AL$ALM1, alm.RNAmf$AL$ALM2))))
par(oldpar)


### 2. ALD Criterion ###
ald.RNAmf <- AL_RNAmf(criterion="ALD",
                      Xcand = x, fit=fit.RNAmf, cost = cost,
                      use_optim = FALSE, parallel = TRUE, ncore = 2)
print(ald.RNAmf$chosen)

### visualize ALD ###
oldpar <- par(mfrow = c(1, 2))
plot(x, ald.RNAmf$AL$ALD1,
     type = "l", lty = 2,
     xlab = "x", ylab = "ALD criterion at the low-fidelity level",
     ylim = c(min(c(ald.RNAmf$AL$ALD1, ald.RNAmf$AL$ALD2)),
              max(c(ald.RNAmf$AL$ALD1, ald.RNAmf$AL$ALD2))))
points(ald.RNAmf$chosen$Xnext,
       ald.RNAmf$AL$ALD1[which(x == drop(ald.RNAmf$chosen$Xnext))],
       pch = 16, cex = 1, col = "red")
plot(x, ald.RNAmf$AL$ALD2,
     type = "l", lty = 2,
     xlab = "x", ylab = "ALD criterion at the high-fidelity level",
     ylim = c(min(c(ald.RNAmf$AL$ALD1, ald.RNAmf$AL$ALD2)),
              max(c(ald.RNAmf$AL$ALD1, ald.RNAmf$AL$ALD2))))
par(oldpar)


### 3. ALC Criterion ###
alc.RNAmf <- AL_RNAmf(criterion="ALC",
                      Xref = x, Xcand = x, fit=fit.RNAmf, cost = cost,
                      use_optim = FALSE, parallel = TRUE, ncore = 2)
print(alc.RNAmf$chosen)

### visualize ALC ###
oldpar <- par(mfrow = c(1, 2))
plot(x, alc.RNAmf$AL$ALC1,
     type = "l", lty = 2,
     xlab = "x", ylab = "ALC criterion augmented at the low-fidelity level",
     ylim = c(min(c(alc.RNAmf$AL$ALC1, alc.RNAmf$AL$ALC2)),
              max(c(alc.RNAmf$AL$ALC1, alc.RNAmf$AL$ALC2))))
points(alc.RNAmf$chosen$Xnext,
       alc.RNAmf$AL$ALC1[which(x == drop(alc.RNAmf$chosen$Xnext))],
       pch = 16, cex = 1, col = "red")
plot(x, alc.RNAmf$AL$ALC2,
     type = "l", lty = 2,
     xlab = "x", ylab = "ALC criterion augmented at the high-fidelity level",
     ylim = c(min(c(alc.RNAmf$AL$ALC1, alc.RNAmf$AL$ALC2)),
              max(c(alc.RNAmf$AL$ALC1, alc.RNAmf$AL$ALC2))))
par(oldpar)


### 4. ALMC Criterion ###
almc.RNAmf <- AL_RNAmf(criterion="ALMC",
                       Xref = x, Xcand = x, fit=fit.RNAmf, cost = cost,
                       use_optim = FALSE, parallel = TRUE, ncore = 2)
print(almc.RNAmf$chosen)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/imputer_RNA.R
\name{imputer_RNA}
\alias{imputer_RNA}
\title{Imputation step in stochastic EM for the non-nested RNA Model}
\usage{
imputer_RNA(XX, yy, kernel=kernel, pred1, fits)
}
\arguments{
\item{XX}{A list of design sets for all fidelity levels, containing \code{X_star}, \code{X_list}, and \code{X_tilde}.}

\item{yy}{A list of current observed and pseudo-responses, containing \code{y_star}, \code{y_list}, and \code{y_tilde}.}

\item{kernel}{A character specifying the kernel type to be used. Choices are \code{"sqex"}(squared exponential), \code{"matern1.5"}, or \code{"matern2.5"}.}

\item{pred1}{Predictive results for the lowest fidelity level \eqn{f_1}. It should include \code{cov} obtained by setting \code{cov.out=TRUE}.}

\item{fits}{A fitted GP object from \code{RNAmf}.}
}
\value{
An updated \code{yy} list containing:
\itemize{
  \item \code{y_star}: An updated pseudo-complete outputs \eqn{\mathbf{y}^*_l}.
  \item \code{y_list}: An original outputs \eqn{\mathbf{y}_l}.
  \item \code{y_tilde}: A newly imputed pseudo outputs \eqn{\widetilde{\mathbf{y}}_l}.
}
}
\description{
The function performs the imputation step of the stochastic EM algorithm for the RNA model when the design is not nested.
The function generates pseudo outputs \eqn{\widetilde{\mathbf{y}}_l} at pseudo inputs \eqn{\widetilde{\mathcal{X}}_l}.
}
\details{
The \code{imputer_RNA} function then imputes the corresponding pseudo outputs
\eqn{\widetilde{\mathbf{y}}_l = f_l(\widetilde{\mathcal{X}}_l)}
by drawing samples from the conditional normal distribution,
given fixed parameter estimates and previous-level outputs \eqn{Y_{l}^{*(m-1)}} for each \eqn{l},
at the \eqn{m}-th iteration of the EM algorithm.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/NestedX.R
\name{NestedX}
\alias{NestedX}
\title{Constructing nested design sets for the RNA model.}
\usage{
NestedX(n, d)
}
\arguments{
\item{n}{A vector specifying the number of design points at each fidelity level \eqn{l}. Thus, the vector must have a positive value \eqn{n_1, \ldots, n_l} where \eqn{n_1 > \cdots > n_l}.}

\item{d}{A positive integer specifying the dimension of the design.}
}
\value{
A list containing the nested design sets at each level, i.e., \eqn{\mathcal{X}_{1}, \ldots, \mathcal{X}_{l}}.
}
\description{
The function constructs nested design sets with multiple fidelity levels
\eqn{\mathcal{X}_l \subseteq \cdots \subseteq \mathcal{X}_{1}} for use in \code{\link{RNAmf}}.
}
\details{
The procedure replace the points of lower level design \eqn{\mathcal{X}_{l-1}}
with the closest points from higher level design \eqn{\mathcal{X}_{l}}.
For details, see "\href{https://github.com/cran/MuFiCokriging}{\code{NestedDesign}}".
}
\examples{
### number of design points ###
n1 <- 30
n2 <- 15

### dimension of the design ###
d <- 2

### fix seed to reproduce the result ###
set.seed(1)

### generate the nested design ###
NX <- NestedX(c(n1, n2), d)

### visualize nested design ###
plot(NX[[1]], col="red", pch=1, xlab="x1", ylab="x2")
points(NX[[2]], col="blue", pch=4)

}
\references{
L. Le Gratiet and J. Garnier (2014). Recursive co-kriging model for design of computer experiments
with multiple levels of fidelity. \emph{International Journal for Uncertainty Quantification}, 4(5), 365-386;
\doi{doi:10.1615/Int.J.UncertaintyQuantification.2014006914}
}
